/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_DMALLOC_H
 #include <dmalloc.h>
#endif

#include "types.h"
#include "proto.h"

struct macro *macros = NULL;
struct macro *macro_last = NULL;
int num_macros = 0;

/* add a new macro struct */

struct macro *add_macro(char *name) {
	struct macro *new, *scan;

	for(scan = macros; scan; scan = scan->next)
		if(!strcasecmp(scan->name, name))
			do_error("multiple definition of macro '%s'\n", name);
	
	if(!(new = (struct macro *) malloc(sizeof(struct macro))))
		do_oom();
		
	if(!(new->name = strdup(name)))
		do_oom();

	new->line = NULL;
	new->num_lines = 0;

	new->next = NULL;
	
	if(!macros)
		macros = new;
	else
		macro_last->next = new;
	
	macro_last = new;

	num_macros++;

	return new;
	}

struct macro *get_macro(char *name) {
	struct macro *scan;

	for(scan = macros; scan; scan = scan->next)
		if(!strcasecmp(scan->name, name))
			break;
	
	if(scan) return scan;

	return NULL;
	}

inline void init_macro(struct macro *macro) {
	macro->lineno = 0;
	}

char *get_macro_line(struct macro *macro) {
	char *line;
	
	if(macro->lineno < macro->num_lines) {
		line = macro->line[file_top->lineno];
		macro->lineno++;
		}
	else
		line = NULL;
	
	return line;
	}

void add_macro_line(struct macro *macro, char *line) {
	macro->num_lines++;
	if(!(macro->line = (char **) realloc(macro->line, sizeof(char *) * macro->num_lines)))
		do_oom();
	
	if(!(macro->line[macro->num_lines - 1] = strdup(line)))
		do_oom();
	}

void macro_free(void) {
	struct macro *scan, *next;
	int i;

	for(scan = macros; scan; scan = next) {
		next = scan->next;
		free(scan->name);
		for(i = 0; i < scan->num_lines; i++)
			free(scan->line[i]);
		free(scan);
		}
	}
