module top;

integer i;
reg noisy;

`define DBW             8
`define ABW             16

        reg reset;
        reg clk;
        reg nmi;                              // active high
        reg irq;                              // active high
        reg rdy;
        reg so;                               // set overflow
        reg [`DBW-1:0] di;             // data input bus

        wire [`DBW-1:0] do;    // data output bus
        wire rw;
        wire [`ABW-1:0] ma;
        wire rw_nxt;
        wire [`ABW-1:0] ma_nxt;
        wire sync;
        wire [`DBW-1:0] ir;
        wire [`DBW-1:0] a;
        wire [`DBW-1:0] x;
        wire [`DBW-1:0] y;
        wire [`DBW-1:0] sp;
        wire [`ABW-1:0] pc;
        wire nf,vf,bf,df,im,zf,cf;    // SR status register


reg [7:0] mem[65535:0];

cpu6502 cpu(reset, clk, nmi, irq, rdy, so, di, do, rw, ma,
        a, x, y, pc, sp, ir, im, df, zf, bf, cf, vf, nf,
        rw_nxt, ma_nxt, sync);

always @(posedge clk)
	begin
	if(sync)
        if(noisy)$display("(PC: %X)(A: %X)(X: %X)(Y: %X)(SP: %X)", pc, a, x, y, sp);
	end

always @(negedge clk)
	begin
	if(rw) 
		begin
		di <= mem[ma]; 
		if(noisy)$display("R: %h (%h)", ma, mem[ma]);
		end
	else 
		begin
		mem[ma] <= do;
		if(noisy)$display("W: %h (%h)", ma, do);
		end
	end

initial begin 

noisy=0;
for(i=0;i<65536;i=i+1)
	begin	
	mem[i]=0;
	end

mem['hfffc]='h12;
mem['hfffd]='hc0;

mem['hc012]='ha9;
mem['hc013]='h00;

mem['hc014]='hD0;
mem['hc015]='hfe;

mem['hc016]='h18;

mem['hc017]='h8d;
mem['hc018]='h06;
mem['hc019]='h10;

mem['hc01a]='h20;
mem['hc01b]='h04;
mem['hc01c]='h50;

mem['hc01d]='h4c;
mem['hc01e]='h12;
mem['hc01f]='hc0;

mem['h5004]='hE8;
mem['h5005]='hD0;
mem['h5006]='h01;
mem['h5007]='h02;
mem['h5008]='h60;

reset = 1;
clk = 0;
nmi = 0;
irq = 0;
rdy = 0;
so = 0;

#1 clk=0; #1 clk=1;
#1 clk=0; #1 clk=1;

reset = 0;
rdy = 1;

#1 clk=0; #1 clk=1;
for(i=0;ir!='h02;i=i+1)
	begin
	#1 clk=0; #1 clk=1;
	end
$display("simulation halted at cycle %d", i+2);
end

endmodule
