#include "aet.h"
#include "mem.h"

struct vector *vectors=NULL;

int maketraces(char *str);

/*
 * Parse a line of the wave file and act accordingly.. 
 * Returns nonzero if trace(s) added.
 */
int parsewavline(char *w)
{
  int len;
  char prefix[256], suffix[1024];
  char *w2;
  int z;

  if(!(len=strlen(w))) return(0);
  if(*(w+len-1)=='\n')
    {
      *(w+len-1)=0x00; /* strip newline if present */
      len--;
      if(!len) return(0);
    }

  w2=w;
  while(1)
    {
      if(isspace(*w2)) { w2++; continue; }
      if(!(*w2)) return(0);	/* no args */
      break;			/* start grabbing chars from here */
    }

  sscanf(w2,"%s",prefix);
  if(*w2=='+')
    {
      /* handle aliasing */
      struct symbol *s;
      sscanf(w2+strlen(prefix),"%s",suffix);

      if(s=symfind(suffix))
	{
	  AddNode(s->n,prefix+1);
	  return(~0);
	}
      else
	{
	  return(0);
	}
    }
  else
    if(*w2=='#')
      {
	/* handle bitvec */
	int flag=0;
	struct Bits *b;

	w2=w2+strlen(prefix);
	while(1)
	  {
	    if(isspace(*w2)) { w2++; continue; }
	    if(!(*w2)) return(0);	/* no more args */	
	    break;			/* start grabbing chars from here */
	  }

	b=makevec(prefix+1,w2);
	AddVector(b,&flag);
	return(b != NULL);
      }
    else
      {
	return(maketraces(w));
      }
}

/*
 * Delete vector from vectors list.  (Called by DeleteTrace() only)
 */
void delvector(struct Bits *vec)
{
struct vector *prev=NULL, *current;

current=vectors;
while(current)
	{
	if(current->bits!=vec)
		{
		prev=current;
		current=current->next;
		}
		else
		{
		if(!prev)
			{
			vectors=current->next;
			}
			else
			{
			prev->next=current->next;
			}
		Vfree((Pointer) vec->name);
		Vfree((Pointer) vec);
		break;
		}
	}
}


/*
 * Make solitary traces from wildcarded signals...
 */
int maketraces(char *str)
{
char *pnt, *wild;
char ch;
int len;
int i;
int made=0;

while(1)
{
pnt=str;
len=0;

while(1)
	{
	ch=*pnt++;
	if(isspace(ch)||(!ch)) break;
	len++;
	}

if(len)
	{
	wild=(char *)Vcalloc(1,len+1);
	memcpy(wild,str,len);

	for(i=numfacs-1;i>=0;i--)
		{
		if(str_match(wild,facs[i]->name))
			{
			AddNode(facs[i]->n,NULL);
			made=~0;
			}
		}
	}

if(!ch) break;
str=pnt;
}
return(made);
}


/*
 * Create a vector from wildcarded signals...
 */
struct Bits *makevec(char *vec, char *str)
{
char *pnt, *wild;
char ch;
int len, nodepnt=0;
int i;
struct Node *n[256];
struct Bits *b=NULL;
struct vector *v;

while(1)
{
pnt=str;
len=0;

while(1)
	{
	ch=*pnt++;
	if(isspace(ch)||(!ch)) break;
	len++;
	}

if(len)
	{
	wild=(char *)Vcalloc(1,len+1);
	memcpy(wild,str,len);

	for(i=numfacs-1;i>=0;i--)
		{
		if(str_match(wild,facs[i]->name))
			{
			n[nodepnt++]=facs[i]->n;
			if(nodepnt==256) goto ifnode;
			}
		}
	}

if(!ch) break;
str=pnt;
}

ifnode:
if(nodepnt)
	{
	b=(struct Bits *)Vcalloc(1,sizeof(struct Bits)+(nodepnt-1)*
				  sizeof(struct Node *));

	for(i=0;i<nodepnt;i++)
		{
		b->nodes[i]=n[i];
		}

	b->nbits=nodepnt;
	strcpy(b->name=(char *)Valloc(strlen(vec)+1, 0),vec);

	v=(struct vector *)Vcalloc(1,sizeof(struct vector));
	v->bits=b;		/* frontlink into our table of defined vecs */
	v->next=vectors;
	vectors=v;
	}

return(b);
}


/*
 * splits facility name into signal and bitnumber
 */
void facsplit(char *str, int *len, int *number)
{
char *ptr;
char *numptr=NULL;
char ch;

ptr=str;

while(ch=*ptr)
        {
        if((ch>='0')&&(ch<='9')) 
                {
                if(!numptr) numptr=ptr;
                }
                else numptr=NULL;
        ptr++;
        }

if(numptr)
        {
        *number=atoi(numptr);
        *len=numptr-str;
        }
        else
        {
        *number=0;
        *len=ptr-str;
        }
}


/*
 * compares two facilities a la strcmp...
 */
int sigcmp(char *s1, char *s2)
{
int len1, num1, len2, num2;
int index=0;
int ch1, ch2;

facsplit(s1, &len1, &num1);
facsplit(s2, &len2, &num2);

for(;;)
        {
        if((ch1=*s1)<(ch2=*s2)) return(-1);
        if(ch1>ch2) return(1);

        s1++; s2++;
        index++;
        if(index==len1)
                {
                if(len1<len2)
                        {
                        return(-1);
                        }
                else
                        {
                        return(num1-num2);
                        }
                }
        else
        if(index==len2)
                {
                return(1);      /* len1>len2 */
                }
        }
}


/*
 * Quicksort algorithm from p154 of 
 * "Introduction to Algorithms" by Cormen, Leiserson, and Rivest.
 * 05-jul-97bsi
 */

int partition(struct symbol **a, int p, int r)
{
struct symbol *x, *t;
int i,j;

x=a[p];
i=p-1;
j=r+1;

while(1)
	{
	do
		{
		j--;
		} while(sigcmp(a[j]->name,x->name)>0);	

	do	{
		i++;
		} while(sigcmp(a[i]->name,x->name)<0);

	if(i<j)
		{
		t=a[i];
		a[i]=a[j];
		a[j]=t;
		}
		else
		{
		return(j);
		}
	}
}

void quicksort(struct symbol **a, int p, int r)
{
int q;

if(p<r)
	{
	q=partition(a,p,r);
	quicksort(a,p,q);
	quicksort(a,q+1,r);
	}
}


/*
 * Quick and dirty wildcard string matcher.  Returns "1" on match.
 */
int str_match (char *p, char *s)	/* from vsim's nsubrs.c */
{
    for (;;)
    {
	if (*p == '*')
	{
	    /* skip past multiple wildcards */
	    while (*p == '*')
		p++;

	    /* if pattern ends with wild card, automatic match */
	    if (*p == '\0')
		return (1);

	    /*
	     * *p now points to first non-wildcard character, find matching
	     * character in string, then recursively match remaining pattern.
	     * if recursive match fails, assume current '*' matches more...
	     */
	    while (*s)
	    {
		while (*s != *p)
		{
		    /*
		     * return if at end of string and we still have pattern
		     * left
		     */
		    if (*s++ == '\0')
			return (0);
		}
		/* attempt to match the new occurrance  */
		if (str_match (p + 1, ++s))
		    return (1);
	    }

	    /*
	     * couldn't find matching character after '*', no match
	     */
	    return (0);
	}
	else if (*p == '\0')
	    return (*s == '\0');
	else if (*p++ != *s++)
	    return (0);
    }
}


