/* 
 * Copyright (c) Tony Bybell 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/*
 * csim.c 30/08/99ajb
 */
#include "csim_internal.h"
#include "lxt_write.h"
#include "alloca.h"

/*********************************************************************************************************************/

/*
 * token name, token descriptor, isgate (1=yes, 0=no)
 * keep defines atomic with cpp tricks..
 */
#define Enum_Tokens		   \
D3("ADDC",	T_ADDC,		1),\
D3("ALI",	T_ALI,		0),\
D3("AND",       T_AND,  	1),\
D3("BUF",       T_BUF,  	1),\
D3("BUFIF0",    T_BUFIF0,	1),\
D3("BUFIF1",    T_BUFIF1,	1),\
D3("END",       T_END,  	0),\
D3("FAC",       T_FAC,  	0),\
D3("LAT",       T_LAT,  	0),\
D3("LATCH",     T_LATCH,        1),\
D3("LATCH_N",   T_LATCH_N,      1),\
D3("LATCH_SR",  T_LATCH_SR,     1),\
D3("LATCH_SR_N",T_LATCH_SR_N,   1),\
D3("MUX",       T_MUX,  	1),\
D3("NAND",      T_NAND, 	1),\
D3("NEGC",	T_NEGC,		1),\
D3("NL_GATE",   T_NL_GATE,      1),\
D3("NL_OPTIMIZED",T_NL_OPTIMIZED, 1),\
D3("NMOS",      T_NMOS, 	1),\
D3("NOR",       T_NOR,  	1),\
D3("NOT",       T_NOT,  	1),\
D3("NOTIF0",    T_NOTIF0,       1),\
D3("NOTIF1",    T_NOTIF1,       1),\
D3("OR",        T_OR,   	1),\
D3("PMOS",      T_PMOS, 	1),\
D3("PRE",       T_PRE,  	0),\
D3("PST",       T_PST,  	0),\
D3("SUBC",	T_SUBC,		1),\
D3("SUPPLY0",   T_SUPPLY0,      1),\
D3("SUPPLY1",   T_SUPPLY1,      1),\
D3("SYM",       T_SYM,  	0),\
D3("TRI",       T_TRI,  	1),\
D3("TRI0",      T_TRI0, 	1),\
D3("TRI1",      T_TRI1, 	1),\
D3("TRIAND",    T_TRIAND,       1),\
D3("TRIOR",     T_TRIOR,        1),\
D3("WAND",      T_WAND, 	1),\
D3("WIRE",      T_WIRE, 	1),\
D3("WOR",       T_WOR,  	1),\
D3("XNOR",      T_XNOR, 	1),\
D3("XOR",       T_XOR,  	1),\
D3("",		T_OTHER,	0)

#undef D3
#define D3(a,b,c) a
static char *tokens[]={Enum_Tokens};

#undef D3
#define D3(a,b,c) b
enum TokenType{Enum_Tokens};

#undef D3
#define D3(a,b,c) c
static char isgate[]={Enum_Tokens};

#define TOKENS T_OTHER	/* gives max # of tokens */


static FILE *handle;
static int lineno=1, linechar1;
static int  yylen;
static char yytext[1025];
static int numfacs=0, numpre_inst=0, numlat_inst=0, numpost_inst=0;
static struct fac *facs=NULL;
static struct vcd *vcd=NULL;
static struct vcd *vcd_lxt=NULL;
static struct lt_trace *lxt=NULL;
static Opcode *pre_op=NULL, *lat_op=NULL, *post_op=NULL;
static int simerrors=0, simwarnings=0;
static char simerror_reported_gates[TOKENS]={0};   /* close enough */
static unsigned int num_gates=0;
static unsigned int num_gates_each_type[TOKENS]={0};
static char debug_exec=0;
static int sim_cycle=0, additional_cycles=0;
static char model_dirty=0;
static char ignore_clocks=0;
static char getfac_mode=0;
static char put_before_get=0;
static char init_value=0;
static int skip_tmp_sigs=0;
static int skip_vcd_tmp_sigs=0;
static int skip_lxt_tmp_sigs=0;
static int vcd_tail_stale=0;
static int vcd_stale_time=0;
static int vcd_bigendian=0;
static int remove_zeros_on_nonvectors=1;
static int lxt_remove_zeros_on_nonvectors=1;

/*********************************************************************************************************************/

static char gate_bufif0[4][4]=
	{
	{0, 3, 3, 3},
	{1, 3, 3, 3},
	{2, 3, 2, 2},
	{2, 3, 2, 2}
	};

static char gate_bufif1[4][4]=
	{
	{3, 0, 3, 3},
	{3, 1, 3, 3},
	{3, 2, 2, 2},
	{3, 2, 2, 2}
	};

static char gate_notif0[4][4]=
	{
	{1, 3, 3, 3},
	{0, 3, 3, 3},
	{2, 3, 2, 2},
	{2, 3, 2, 2}
	};

static char gate_notif1[4][4]=
	{
	{3, 1, 3, 3},
	{3, 0, 3, 3},
	{3, 2, 2, 2},
	{3, 2, 2, 2}
	};

static char gate_negc[4][4]=	/* (NOTA)B */
	{
	{0, 1, 2, 2},
	{0, 0, 2, 2},
	{2, 2, 2, 2},
	{2, 2, 2, 2}
	};


/*********************************************************************************************************************/


/*
 * wrapped malloc family...
 */

static int mem_allocated;

static void *malloc_2(size_t size)
{
void *ret;
ret=malloc(size);
if(ret)
        {
	mem_allocated+=size;
        return(ret);
        }
        else
        {
        fprintf(stderr, CSIM"FATAL ERROR : Out of memory, sorry.\n");
        exit(255);
        }
}

static void *calloc_2(size_t nmemb, size_t size)
{
void *ret;
ret=calloc(nmemb, size);
if(ret)
        {
	mem_allocated+=nmemb*size;
        return(ret);
        }
        else
        {
        fprintf(stderr, CSIM"FATAL ERROR: Out of memory, sorry.\n");
        exit(1);
        }
}

static void free_2(void *ptr)
{
if(ptr)
        {
        free(ptr);
        }
        else
        {
        fprintf(stderr, CSIM"WARNING: Attempt to free NULL pointer caught.\n");
	simwarnings++;
        }
}


/*********************************************************************************************************************/

/*
 * single char get
 */
static int getch(void)
{
int ch;

ch=fgetc(handle);

/* printf("(%c)",ch); */

if(ch=='\n') lineno++;
return(((ch==EOF)||(errno))?(-1):(ch));
}


/*
 * simple lexer
 */
static int tok_compar(const void *key, const void *ele)
{
return(strcmp(key,*((char **)ele)));
}


static char *decompress_sym_str=NULL;
static int  decompress_sym_len=0;

int decompress_sym(void)
{
int ch;
char *pnt;
int replen, newlen;

top:
do	{
	ch=getch();
	if(ch<0) return(ch);
	} while((ch==' ')||(ch=='\t')||(ch=='\r')||(ch=='\n'));

pnt=yytext;
*(pnt++)=ch;

for(;;)
	{
	ch=getch();
	if((ch<0)||(ch==' ')||(ch=='\t')||(ch=='\r')||(ch=='\n')) break;
	*(pnt++)=ch;
	}
*pnt=0x00;

yylen=(pnt-yytext);

replen=((int)((unsigned char)yytext[0]))-33;
newlen=replen+yylen-1;
pnt=(char *)malloc_2(newlen+1);
if(decompress_sym_str)
	{
	if(replen) strncpy(pnt, decompress_sym_str, replen);
	strcpy(pnt+replen, yytext+1);
	free_2(decompress_sym_str);
	}
	else
	{
	strcpy(pnt, yytext+1);
	}

decompress_sym_str=pnt;
decompress_sym_len=newlen;
return(0);
}

static int lex(void)
{
int ch;
char *pnt;
char **rvar;
char is_number=1;

top:
do	{
	ch=getch();
	if(ch<0) return(ch);
	} while((ch==' ')||(ch=='\t')||(ch=='\r')||(ch=='\n'));

linechar1=lineno;
if(ch=='#')
	{
	do	{
		ch=getch();
		if(ch<0) return(ch);
		} while (ch!='\n');
	goto top;
	}

pnt=yytext;
*(pnt++)=ch;
if((ch<'0')||(ch>'9')) is_number=0;

for(;;)
	{
	ch=getch();
	if((ch<0)||(ch==' ')||(ch=='\t')||(ch=='\r')||(ch=='\n')) break;
	*(pnt++)=ch;
	if((ch<'0')||(ch>'9')) is_number=0;
	}
*pnt=0x00;

yylen=(pnt-yytext);

if(is_number)
	{
	return(T_OTHER);
	}
	else
	{
	rvar=bsearch((const void *)yytext, (const void *)tokens, TOKENS, sizeof(char **), tok_compar); 
	return(rvar?(rvar-tokens):T_OTHER);
	}
}

static void throw_error(void)
{
fprintf(stderr, CSIM"ERROR AT LINE: %d ('%s')\n",linechar1,yytext);
exit(255);
}


void build_model(void)
{
int tok, i;
struct fac *f;

mem_allocated=0;

tok=lex();
if(tok!=T_SYM) { throw_error(); }
lex();
numfacs=atoi(yytext);
if(numfacs<1) throw_error();
facs=calloc_2(numfacs, sizeof(struct fac));

fprintf(stderr, CSIM CSIM_VERSION "\n");

fprintf(stderr, CSIM"Allocating space for %d facilit%s\n",numfacs,(numfacs!=1)?"ies":"y");
for(i=0;i<numfacs;i++)
	{
	char ch, att;
	char *pnt;
	int facnum;

	att=0;
	tok=lex();
	if(tok!=T_FAC) { throw_error(); }
	lex();
	pnt=yytext;
	while((ch=*(pnt++)))
		{
		switch(ch)
			{
			case 'I': att|=SYM_IN;	 break;
			case 'O': att|=SYM_OUT;	 break;
			case 'W': att|=SYM_WIRE; break;
			case 'R': att|=SYM_REG;
			default:		 break;
			}
		}
	(f=facs+i)->attributes=att;

	lex();
	facnum=atoi(yytext);
	if(facnum!=i) { throw_error(); }	

	decompress_sym();
	strcpy(f->name=(char *)malloc_2(decompress_sym_len+1),decompress_sym_str);
	if((decompress_sym_len==2)&&(*decompress_sym_str=='$'))
		{
		char ch;
		f->attributes|=SYM_CONST;

		switch(*(decompress_sym_str+1))
			{
			case '0': ch=0;	break;
			case '1': ch=1; break;
			case 'x':
			case 'X': ch=2; break;
			case 'z': 
			case 'Z': ch=3; break;
			default:  
				ch=2;
				fprintf(stderr,CSIM"Constant '%s' undefined, defaulting to 'X'\n",decompress_sym_str);
				break;
			}		

		f->u.v.real_value=ch;
		}
		else
		{
		if(!(f->attributes&SYM_REG)) f->u.v.real_value=init_value;
		}
	}

tok=lex();
while(tok==T_ALI)
	{
	struct fac *f;
	int num_aliases, i, fakefac;

	lex();
	f=(facs+atoi(yytext));
	lex();
	num_aliases=atoi(yytext);
	for(i=0;i<num_aliases;i++)
		{
		lex();
		fakefac=atoi(yytext);
		facs[fakefac].u.alias=f;
		facs[fakefac].attributes|=SYM_ALIAS;
		}
	
	tok=lex();
	}

if(tok==T_PRE)
	{
	lex();
	numpre_inst=atoi(yytext);
	if(numpre_inst<0) { throw_error(); }
	fprintf(stderr, CSIM"Allocating space for %d prelatch opcodes\n",numpre_inst);	

	if(numpre_inst>0)
		{
		int opcnt=0;
		int iter, s;
		Opcode *op;
		op=pre_op=(Opcode *)calloc_2(numpre_inst, sizeof(Opcode));

		for(;opcnt<numpre_inst;)
			{
			int j;

			tok=lex();
			if(!isgate[tok]) { throw_error(); }
			num_gates_each_type[tok]++;
			(*(op++)).i=tok; opcnt++;

			for(j=0;j<2;j++)	/* 0=inports , 1=outports */
				{
				lex();
				iter=atoi(yytext);
				(*(op++)).i=iter; opcnt++;
				for(i=0;i<iter;i++)			
					{
					lex();
					s=atoi(yytext);
					if((s<0)||(s>=numfacs)) { throw_error(); }
					(*(op++)).pnt=(facs+s);	opcnt++;	/* store addr of fac directly */
					}				
				}
			}		
		}
	tok=lex();		
	}

if(tok==T_LAT)
	{
	lex();
	numlat_inst=atoi(yytext);
	if(numlat_inst<0) { throw_error(); }
	fprintf(stderr, CSIM"Allocating space for %d latch opcodes\n",numlat_inst);	

	if(numlat_inst>0)
		{
		int opcnt=0;
		int iter, s;
		Opcode *op;
		op=lat_op=(Opcode *)calloc_2(numlat_inst, sizeof(Opcode));

		for(;opcnt<numlat_inst;)
			{
			int j;

			tok=lex();
			if((tok<T_LATCH)||(tok>T_LATCH_SR_N)) { throw_error(); }
			num_gates_each_type[tok]++;
			(*(op++)).i=tok; opcnt++;

			for(j=0;j<2;j++)	/* 0=inports , 1=outports */
				{
				lex();
				iter=atoi(yytext);
				(*(op++)).i=iter; opcnt++;
				for(i=0;i<iter;i++)			
					{
					lex();
					s=atoi(yytext);
					if((s<0)||(s>=numfacs)) { throw_error(); }
					(*(op++)).pnt=(facs+s);	opcnt++;	/* store addr of fac directly */
					}				
				}
			}		
		}
	tok=lex();		
	}

if(tok==T_PST)
	{
	lex();
	numpost_inst=atoi(yytext);
	if(numpost_inst<0) { throw_error(); }
	fprintf(stderr, CSIM"Allocating space for %d postlatch opcodes\n",numpost_inst);	

	if(numpost_inst>0)
		{
		int opcnt=0;
		int iter, s;
		Opcode *op;
		op=post_op=(Opcode *)calloc_2(numpost_inst, sizeof(Opcode));

		for(;opcnt<numpost_inst;)
			{
			int j;

			tok=lex();
			if(!isgate[tok]) { throw_error(); }
			num_gates_each_type[tok]++;
			(*(op++)).i=tok; opcnt++;

			for(j=0;j<2;j++)	/* 0=inports , 1=outports */
				{
				lex();
				iter=atoi(yytext);
				(*(op++)).i=iter; opcnt++;
				for(i=0;i<iter;i++)			
					{
					lex();
					s=atoi(yytext);
					if((s<0)||(s>=numfacs)) { throw_error(); }
					(*(op++)).pnt=(facs+s);	opcnt++;	/* store addr of fac directly */
					}				
				}
			}		
		}
	tok=lex();		
	}

if((tok!=T_END)&&(tok>=0)) { throw_error(); }

fprintf(stderr, CSIM"Model occupies %d bytes\n",mem_allocated);
fprintf(stderr, CSIM"\n");

for(num_gates=0, i=0; i<TOKENS; i++)
	{
	num_gates+=num_gates_each_type[i];
	}

if(num_gates)
	{
	fprintf(stderr, CSIM"Operation  | Count      | Percent\n");
	fprintf(stderr, CSIM"---------- + ---------- + -------\n");
	for(i=0;i<TOKENS;i++)
		{
		if(num_gates_each_type[i])
			{
			fprintf(stderr, CSIM"%-10s | %-10d | %.2f%%\n", tokens[i], 
				num_gates_each_type[i], (float)(100*num_gates_each_type[i])/(float)num_gates);
			}
		}
	fprintf(stderr, CSIM"---------- + ---------- + -------\n");
	fprintf(stderr, CSIM"Total      | %-10d | 100.00%%\n", num_gates);
	}
fprintf(stderr, CSIM"\n");

}

static FILE *load_model(char *name)
{
handle=fopen(name,"rb");
if(handle)
	{
	build_model();
	fclose(handle);
	}
return(handle);
}

/*********************************************************************************************************************/

void sim_opcodes(Opcode *op, Opcode *fini, int latchphase)
{
Opcode *opdebug;
struct fac *f;
int i;
int inports, outports;
char val, operand, operand2, operand3;

if(debug_exec)fprintf(stderr, CSIM"------------------------------------------------------------\n");

while(op!=fini)
{
if(debug_exec)
	{
	opdebug=op;
	fprintf(stderr, CSIM"%-10s",tokens[(*(opdebug++)).i]);
	inports=(*(opdebug++)).i;
	for(i=0;i<inports;i++)
		{
		f=(struct fac *)(*(opdebug++)).pnt;
		operand=f->u.v.real_value;
		fprintf(stderr, " %c","01xz"[operand]);
		}
	}

switch((*(op++)).i)
	{
	case T_SUBC:					/* majority function A(NOT B)+AC+(NOT B)C */
			inports=(*(op++)).i;
			if(inports==3)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;
				f=(struct fac *)(*(op++)).pnt;
				operand3=f->u.v.real_value;

				if((operand&(~1))||(operand2&(~1))||(operand3&(~1)))	/* x/z inputs */
					{
					val=2;
					}
					else
					{
					operand2=operand2^1;	/* NOT B */
					val=(operand&operand2)|(operand&operand3)|(operand2&operand3);
					}

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}
			else
			if(inports==2)				/* A + (NOT B) */
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;

				if((operand&(~1))||(operand2&(~1)))			/* x/z inputs */
					{
					val=2;
					}
					else
					{
					operand2=operand2^1;	/* NOT B */
					val=(operand|operand2);
					}

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}
				else
				{
				op+=inports;	/* skip input ports */
				op+=((*op).i+1);/* skip output ports */

				simerrors++;
				}
			break;

	case T_ADDC:					/* majority function AB+AC+BC */
			inports=(*(op++)).i;
			if(inports==3)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;
				f=(struct fac *)(*(op++)).pnt;
				operand3=f->u.v.real_value;

				if((operand&(~1))||(operand2&(~1))||(operand3&(~1)))	/* x/z inputs */
					{
					val=2;
					}
					else
					{
					val=(operand&operand2)|(operand&operand3)|(operand2&operand3);
					}

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}
				else
				{
				op+=inports;	/* skip input ports */
				op+=((*op).i+1);/* skip output ports */

				simerrors++;
				}
			break;

	case T_NEGC:
			inports=(*(op++)).i;
			if(inports==2)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;

				val=gate_negc[operand][operand2];

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}
				else
				{
				op+=inports;	/* skip input ports */
				op+=((*op).i+1);/* skip output ports */

				simerrors++;
				}
			break;

	case T_AND:
			inports=(*(op++)).i;
			val=1;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*op).pnt;
				operand=f->u.v.real_value;
				
				if(!operand)
					{
					val=0; op+=(inports-i); break;
					}
				else
				switch(operand&3)
					{
					case 0:
					case 1: break;
					case 2: 
					case 3:
						if(val!=0) val=2;
						break;
					}
				op++;
				}
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;

	case T_NAND:
			inports=(*(op++)).i;
			val=1;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	val=0; break;
					case 1: break;
					case 2: 
					case 3:
						if(val!=0) val=2;
						break;
					}
				}

			switch(val&3)
				{
				case 0:	val=1; break;
				case 1: val=0; break;
				case 2:
				case 3: break;
				}

			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;

	case T_OR:
			inports=(*(op++)).i;
			val=0;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*op).pnt;
				operand=f->u.v.real_value;

				if(operand==1)
					{
					val=1; op+=(inports-i); break;
					}
					else
					{
					val|=(operand&2);
					}
				op++;
				}
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_NOR:
			inports=(*(op++)).i;
			val=0;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	break;
					case 1: val=1; break;
					case 2: 
					case 3:
						if(val!=1) val=2;
						break;
					}
				}

			switch(val&3)
				{
				case 0:	val=1; break;
				case 1: val=0; break;
				case 2:
				case 3: break;
				}

			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_NOT:
			inports=(*(op++)).i;
			val=0;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	val=1; break;
					case 1: val=0; break;
					case 2: 
					case 3:
						val=2;
						break;
					}
				}
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_BUF:
			inports=(*(op++)).i;
			val=0;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	
					case 1: val=operand; break;
					case 2: 
					case 3:
						val=2;
						break;
					}
				}
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_XOR:
			inports=(*(op++)).i;
			val=0;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	break;
					case 1: if(val!=2) val=1-val; break;
					case 2: 
					case 3:
						val=2;
						break;
					}
				}
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_XNOR:
			inports=(*(op++)).i;
			val=0;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	break;
					case 1: if(val!=2) val=1-val; break;
					case 2: 
					case 3:
						val=2;
						break;
					}
				}

			switch(val&3)
				{
				case 0:	val=1; break;
				case 1: val=0; break;
				case 2:
				case 3: break;
				}

			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_WIRE:
	case T_NL_GATE:
	case T_NL_OPTIMIZED:
			inports=(*(op++)).i;
			val=3;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	if((val==3)||(val==0)) val=0; else val=2; break;
					case 1: if((val==3)||(val==1)) val=1; else val=2; break;
					case 3: break;
					case 2: 
						val=2;
						break;
					}
				}
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_MUX:
			inports=(*(op++)).i;
			if(inports==3)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				switch(operand)
					{
					case 0:
						f=(struct fac *)(*(op+1)).pnt;
						val=f->u.v.real_value;
						break;
					case 1:
						f=(struct fac *)(*(op)).pnt;
						val=f->u.v.real_value;
						break;
					default:
						val=2;
						break;
					}

				op+=2;
				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}
				else
				{
				fprintf(stderr, CSIM"Malformed MUX with %d inputs encountered\n",inports);
				simerrors++;
				op+=inports;	/* skip input ports */
				op+=((*op).i+1);/* skip output ports */
				}

			break;

	case T_LATCH:
			inports=(*(op++)).i;

			if(!latchphase)
				{
				f=(struct fac *)(*(op)).pnt;
				operand=f->u.v.real_value;   /* d pin is 1st pin */

				op+=inports;	/* skip input ports */

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					f->u.v.latch_value=operand;
					}
				}
				else
				{
				if((inports==1)||(ignore_clocks))
					{
					op+=inports;    /* skip input ports */

					outports=(*(op++)).i;
					for(i=0;i<outports;i++)
						{
						f=(struct fac *)(*(op++)).pnt;
						if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=f->u.v.latch_value;				
						}
					}
					else
					{
					f=(struct fac *)(*(op+1)).pnt;
					operand=f->u.v.real_value;   /* clock pin is 2nd pin */
					op+=inports;    /* skip input ports */

					outports=(*(op++)).i;

					if(operand==1)		/* 0->1 transition */
						{
						for(i=0;i<outports;i++)
							{
							f=(struct fac *)(*(op++)).pnt;
							if((f->u.v.prev_clock==0)&&(!(f->attributes&SYM_STKFAC))) f->u.v.real_value=f->u.v.latch_value;
							f->u.v.prev_clock=operand;
							}
						}
						else
						{
						for(i=0;i<outports;i++)
							{
							f=(struct fac *)(*(op++)).pnt;
							f->u.v.prev_clock=operand;
							}
						}
					}
				}

			break;


	case T_LATCH_N:
			inports=(*(op++)).i;

			if(!latchphase)
				{
				f=(struct fac *)(*(op)).pnt;
				operand=f->u.v.real_value;   /* d pin is 1st pin */

				op+=inports;	/* skip input ports */

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					f->u.v.latch_value=operand;
					}
				}
				else
				{
				if((inports==1)||(ignore_clocks))
					{
					op+=inports;    /* skip input ports */

					outports=(*(op++)).i;
					for(i=0;i<outports;i++)
						{
						f=(struct fac *)(*(op++)).pnt;
						if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=f->u.v.latch_value;				
						}
					}
					else
					{
					f=(struct fac *)(*(op+1)).pnt;
					operand=f->u.v.real_value;   /* clock pin is 2nd pin */
					op+=inports;    /* skip input ports */

					outports=(*(op++)).i;

					if(operand==0)		/* 1->0 transition */
						{
						for(i=0;i<outports;i++)
							{
							f=(struct fac *)(*(op++)).pnt;
							if((f->u.v.prev_clock==1)&&(!(f->attributes&SYM_STKFAC))) f->u.v.real_value=f->u.v.latch_value;
							f->u.v.prev_clock=operand;
							}
						}
						else
						{
						for(i=0;i<outports;i++)
							{
							f=(struct fac *)(*(op++)).pnt;
							f->u.v.prev_clock=operand;
							}
						}
					}
				}

			break;

	case T_LATCH_SR: 
			inports=(*(op++)).i;

			if(!latchphase)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;   /* d pin is 1st pin */
				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;  /* s pin is 2nd pin */
				f=(struct fac *)(*(op++)).pnt;
				operand3=f->u.v.real_value;  /* r pin is 3rd pin */

				if(operand3==1) operand=0;
				else if(operand2==1) operand=1;	/* resolve in favor of reset */

				op++; /* skip clock */
				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					f->u.v.latch_value=operand;
					if(((operand2==1)||(operand3==1))&&(!(f->attributes&SYM_STKFAC))) f->u.v.real_value=operand;
					}
				}
				else
				{
				if(ignore_clocks)
					{
					op+=inports;    /* skip input ports */

					outports=(*(op++)).i;
					for(i=0;i<outports;i++)
						{
						f=(struct fac *)(*(op++)).pnt;
						if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=f->u.v.latch_value;				
						}
					}
					else
					{
					f=(struct fac *)(*(op+3)).pnt;
					operand=f->u.v.real_value;   /* clock pin is 4th pin */
					op+=inports;    /* skip input ports */

					outports=(*(op++)).i;

					if(operand==1)		/* 0->1 transition */
						{
						for(i=0;i<outports;i++)
							{
							f=(struct fac *)(*(op++)).pnt;
							if((f->u.v.prev_clock==0)&&(!(f->attributes&SYM_STKFAC))) f->u.v.real_value=f->u.v.latch_value;
							f->u.v.prev_clock=operand;
							}
						}
						else
						{
						for(i=0;i<outports;i++)
							{
							f=(struct fac *)(*(op++)).pnt;
							f->u.v.prev_clock=operand;
							}
						}
					}
				}

			break;

	case T_LATCH_SR_N:
			inports=(*(op++)).i;

			if(!latchphase)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;   /* d pin is 1st pin */
				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;  /* s pin is 2nd pin */
				f=(struct fac *)(*(op++)).pnt;
				operand3=f->u.v.real_value;  /* r pin is 3rd pin */

				if(operand3==1) operand=0;	/* pins are STILL positive active for negedge flops */
				else if(operand2==1) operand=1;	/* resolve in favor of reset */

				op++; /* skip clock */
				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					f->u.v.latch_value=operand;
					if(((operand2==1)||(operand3==1))&&(!(f->attributes&SYM_STKFAC))) f->u.v.real_value=operand;
					}
				}
				else
				{
				if(ignore_clocks)
					{
					op+=inports;    /* skip input ports */

					outports=(*(op++)).i;
					for(i=0;i<outports;i++)
						{
						f=(struct fac *)(*(op++)).pnt;
						if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=f->u.v.latch_value;				
						}
					}
					else
					{
					f=(struct fac *)(*(op+3)).pnt;
					operand=f->u.v.real_value;   /* clock pin is 4th pin */
					op+=inports;    /* skip input ports */

					outports=(*(op++)).i;

					if(operand==0)		/* 1->0 transition */
						{
						for(i=0;i<outports;i++)
							{
							f=(struct fac *)(*(op++)).pnt;
							if((f->u.v.prev_clock==1)&&(!(f->attributes&SYM_STKFAC))) f->u.v.real_value=f->u.v.latch_value;
							f->u.v.prev_clock=operand;
							}
						}
						else
						{
						for(i=0;i<outports;i++)
							{
							f=(struct fac *)(*(op++)).pnt;
							f->u.v.prev_clock=operand;
							}
						}
					}
				}

			break;


	case T_BUFIF1: 
			inports=(*(op++)).i;
			if((inports!=2)&&(inports!=3))
				{
				fprintf(stderr, CSIM"Malformed BUFIF1 with %d inputs encountered\n",inports);
				simerrors++;
				op+=inports;	/* skip input ports */
				op+=((*op).i+1);/* skip output ports */
				}
				else
				{
				if(inports==3) op++; /* skip extra $z if present (ver -pica) */
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;

				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;

				val=gate_bufif1[operand][operand2];

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}

			break;

	case T_BUFIF0: 
			inports=(*(op++)).i;
			if((inports!=2)&&(inports!=3))
				{
				fprintf(stderr, CSIM"Malformed BUFIF0 with %d inputs encountered\n",inports);
				simerrors++;
				op+=inports;	/* skip input ports */
				op+=((*op).i+1);/* skip output ports */
				}
				else
				{
				if(inports==3) op++; /* skip extra $z if present (ver -pica) */
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;

				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;

				val=gate_bufif0[operand][operand2];

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}

			break;

	case T_NOTIF1: 
			inports=(*(op++)).i;
			if((inports!=2)&&(inports!=3))
				{
				fprintf(stderr, CSIM"Malformed NOTIF1 with %d inputs encountered\n",inports);
				simerrors++;
				op+=inports;	/* skip input ports */
				op+=((*op).i+1);/* skip output ports */
				}
				else
				{
				if(inports==3) op++; /* skip extra $z if present (ver -pica) */
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;

				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;

				val=gate_notif1[operand][operand2];

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}

			break;

	case T_NOTIF0: 
			inports=(*(op++)).i;
			if((inports!=2)&&(inports!=3))
				{
				fprintf(stderr, CSIM"Malformed NOTIF0 with %d inputs encountered\n",inports);
				simerrors++;
				op+=inports;	/* skip input ports */
				op+=((*op).i+1);	/* skip output ports */
				}
				else
				{
				if(inports==3) op++; /* skip extra $z if present (ver -pica) */
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;

				f=(struct fac *)(*(op++)).pnt;
				operand2=f->u.v.real_value;

				val=gate_notif0[operand][operand2];

				outports=(*(op++)).i;
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}

			break;


	case T_SUPPLY0:
			inports=(*(op++)).i;
			val=0;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				if((operand==1)||(operand==2)) val=2;	/* blue smoke on 1 or X */
				}

			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				f->u.v.real_value=val;				
				}
			break;
			

	case T_SUPPLY1:
			inports=(*(op++)).i;
			val=1;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				if((operand==0)||(operand==2)) val=2;	/* blue smoke on 0 or X */
				}

			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}
			break;


	case T_WAND:
			inports=(*(op++)).i;
			val=3;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	val=0; break;
					case 1: if(val==3) val=1; break;
					case 3: break;
					case 2:
						if(val!=0) val=2;
						break;
					}
				}
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_WOR:
			inports=(*(op++)).i;
			val=3;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	if(val==3) val=0; break;
					case 1: val=1; break;
					case 3: break;
					case 2: 
						if(val!=1) val=2;
						break;
					}
				}
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_TRI:
			inports=(*(op++)).i;
			val=3;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	if((val==3)||(val==0)) val=0; else val=2; break;
					case 1: if((val==3)||(val==1)) val=1; else val=2; break;
					case 3: break;
					case 2: 
						val=2;
						break;
					}
				}
			outports=(*(op++)).i;
			if(val!=3)
				{
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}
				else
				{
				op+=outports;	/* keep old charge on net */
				}

			break;


	case T_TRIAND:
			inports=(*(op++)).i;
			val=3;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	val=0; break;
					case 1: if(val==3) val=1; break;
					case 3: break;
					case 2:
						if(val!=0) val=2;
						break;
					}
				}
			outports=(*(op++)).i;
			if(val!=3)
				{
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}
				else
				{
				op+=outports;	/* keep old charge on net */
				}

			break;


	case T_TRIOR:
			inports=(*(op++)).i;
			val=3;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	if(val==3) val=0; break;
					case 1: val=1; break;
					case 3: break;
					case 2: 
						if(val!=1) val=2;
						break;
					}
				}
			outports=(*(op++)).i;
			if(val!=3)
				{
				for(i=0;i<outports;i++)
					{
					f=(struct fac *)(*(op++)).pnt;
					if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
					}
				}
				else
				{
				op+=outports;	/* keep old charge on net */
				}

			break;


	case T_TRI0:
			inports=(*(op++)).i;
			val=3;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	if((val==3)||(val==0)) val=0; else val=2; break;
					case 1: if((val==3)||(val==1)) val=1; else val=2; break;
					case 3: break;
					case 2: 
						val=2;
						break;
					}
				}
			if(val==3) val=0;	/* pull toward 0 */
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;


	case T_TRI1:
			inports=(*(op++)).i;
			val=3;
			for(i=0;i<inports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				operand=f->u.v.real_value;
				
				switch(operand&3)
					{
					case 0:	if((val==3)||(val==0)) val=0; else val=2; break;
					case 1: if((val==3)||(val==1)) val=1; else val=2; break;
					case 3: break;
					case 2: 
						val=2;
						break;
					}
				}
			if(val==3) val=1;	/* pull toward 1 */
			outports=(*(op++)).i;
			for(i=0;i<outports;i++)
				{
				f=(struct fac *)(*(op++)).pnt;
				if(!(f->attributes&SYM_STKFAC)) f->u.v.real_value=val;				
				}

			break;




	case T_PMOS:	/* bi-di can't be done in cycle sim */
	case T_NMOS:	/* bi-di can't be done in cycle sim */

	default: 
		if(!simerror_reported_gates[(*(op-1)).i])
			{
			fprintf(stderr,CSIM"Gate type '%s' not supported\n", tokens[(*(op-1)).i]);
			simerror_reported_gates[(*(op-1)).i]=1;
			}

		op+=((*op).i+1);/* skip input ports */
		op+=((*op).i+1);/* skip output ports */

		simerrors++;
		break;

	}

if(debug_exec)
	{
	outports=(*(opdebug++)).i;
	fprintf(stderr," |");
	for(i=0;i<outports;i++)
		{
		f=(struct fac *)(*(opdebug++)).pnt;
		operand=f->u.v.real_value;
		fprintf(stderr, " %c","01xz"[operand]);
		}
	fprintf(stderr,"\n");
	}

}

}


int SimulateModel(int cycles)
{
int i;

if(put_before_get)
	{
	for(i=0;((i<cycles)&&(!simerrors));i++)
		{
		sim_cycle++;
	
		sim_opcodes(pre_op, pre_op+numpre_inst, 0);
		if(lat_op)
			{
			sim_opcodes(lat_op, lat_op+numlat_inst, 0);
			sim_opcodes(lat_op, lat_op+numlat_inst, 1);	/* propagate through latch */
			}
		}

	model_dirty=1;
	}
	else
	{
	for(i=0;((i<cycles)&&(!simerrors));i++)
		{
		sim_cycle++;
	
		if(pre_op) sim_opcodes(pre_op, pre_op+numpre_inst, 0);
		if(lat_op)
			{
			sim_opcodes(lat_op, lat_op+numlat_inst, 0);
			sim_opcodes(lat_op, lat_op+numlat_inst, 1);	/* propagate through latch */
			}
		if(post_op) sim_opcodes(post_op, post_op+numpost_inst, 0);
		}

	model_dirty=0;
	}	


return(simerrors);
}


int SimulateCones(void)
{
if(pre_op) sim_opcodes(pre_op, pre_op+numpre_inst, 0);
if(post_op) sim_opcodes(post_op, post_op+numpost_inst, 0);

return(simerrors);
}


int SimulateLatches(void)
{
if(lat_op)
	{
	sim_opcodes(lat_op, lat_op+numlat_inst, 0);
	sim_opcodes(lat_op, lat_op+numlat_inst, 1);	/* propagate through latch */
	}
if(post_op) sim_opcodes(post_op, post_op+numpost_inst, 0);

return(simerrors);
}


int IncrementCycleCount(int cycles)
{
sim_cycle+=cycles;
return(sim_cycle);
}


/*********************************************************************************************************************/

static int sigcmp(char *s1, char *s2)
{
unsigned char c1, c2;
int u1, u2;
         
for(;;)
        {
        c1=(unsigned char)*(s1++);
        c2=(unsigned char)*(s2++);
             
        if((!c1)&&(!c2)) return(0);
        if((c1<='9')&&(c2<='9')&&(c2>='0')&&(c1>='0'))
                {
                u1=(int)(c1&15);   
                u2=(int)(c2&15);
         
                while(((c2=(unsigned char)*s2)>='0')&&(c2<='9'))
                        {
                        u2*=10;
                        u2+=(unsigned int)(c2&15);
                        s2++;
                        }
    
                while(((c2=(unsigned char)*s1)>='0')&&(c2<='9'))
                        {
                        u1*=10;
                        u1+=(unsigned int)(c2&15);
                        s1++;
                        }
 
                if(u1==u2) continue;
                        else return((int)u1-(int)u2);
                }
                else
                {
                if(c1!=c2) return((int)c1-(int)c2);
                }
        }
}

static int compar_facs(const void *key, const void *v2)
{
struct fac *s2;
int rc;

s2=((struct fac *)v2);
rc=sigcmp((char *)key,s2->name);
return(rc);
}

static struct fac *bsearch_facs(char *ascii)
{
struct fac *rc;

rc=(struct fac *)bsearch(ascii, facs, numfacs, sizeof(struct fac), compar_facs);
if(rc) return(rc); else return(NULL);
}

int GetFacilityIndex(char *name, int which)
{
int len;
char *pnt;
struct fac *f;

if(!name) return(-1);
len=strlen(name);
if(!len) return(-1);

if(which>-1)
	{
	pnt=(char *)cyco_alloca(len+20);
	sprintf(pnt,"%s/%d",name,which);
	}
	else
	pnt=name;

f=bsearch_facs(pnt);
return((!f)?-1:(f-facs));	/* which element in the array */
}

/*********************************************************************************************************************/

int GetFac(int msb_index, int lsb_index)
{
int i, direction, iter;
unsigned int val=0;
struct fac *f, *fprime;
char operand;

if(model_dirty)
	{
	additional_cycles++;
	if(pre_op) sim_opcodes(pre_op, pre_op+numpre_inst, 0);
	model_dirty=0;
	}

if((msb_index<0)||(msb_index>=numfacs)||(lsb_index<0)||(lsb_index>=numfacs))
	{
	fprintf(stderr, CSIM"WARNING: GETFAC INDEX ERROR [%d:%d]\n", msb_index, lsb_index);
	simwarnings++;
	return(0);
	}

if(msb_index>lsb_index)
	{
	direction=-1;
	iter=msb_index-lsb_index+1;
	}
	else
	{
	direction=1;
	iter=lsb_index-msb_index+1;
	}
	
f=facs+msb_index;	
for(i=0;i<iter;i++)
	{
	if(!(f->attributes&SYM_ALIAS))
		{
		operand=(f->u.v.real_value)&1;
		}
		else
		{
		fprime=f->u.alias;
		operand=(fprime->u.v.real_value)&1;
		}
	val=(val<<1)|operand;
	f+=direction;
	}

return((int)val);
}

int GetFacName(char *name, int msb_index, int lsb_index)
{
return(GetFac(GetFacilityIndex(name, msb_index), GetFacilityIndex(name, lsb_index)));
}

/*********************************************************************************************************************/

char *GetFacX(char *buffer, int msb_index, int lsb_index)
{
int i, direction, iter;
struct fac *f, *fprime;
char operand;
char *pnt;

if(!buffer) return(NULL);
pnt=buffer;

if(model_dirty)
	{
	additional_cycles++;
	if(pre_op) sim_opcodes(pre_op, pre_op+numpre_inst, 0);
	model_dirty=0;
	}

if((msb_index<0)||(msb_index>=numfacs)||(lsb_index<0)||(lsb_index>=numfacs))
	{
	fprintf(stderr, CSIM"WARNING: GETFACX INDEX ERROR [%d:%d]\n", msb_index, lsb_index);
	simwarnings++;
	return(NULL);
	}

if(msb_index>lsb_index)
	{
	direction=-1;
	iter=msb_index-lsb_index+1;
	}
	else
	{
	direction=1;
	iter=lsb_index-msb_index+1;
	}
	
f=facs+msb_index;	
for(i=0;i<iter;i++)
	{
	if(!(f->attributes&SYM_ALIAS))
		{
		operand=(f->u.v.real_value)&3;
		}
		else
		{
		fprime=f->u.alias;
		operand=(fprime->u.v.real_value)&3;
		}
	*(pnt++)="01XZ"[operand];
	f+=direction;
	}

*pnt=0x00;
return(buffer);
}

char *GetFacNameX(char *name, char *buffer, int msb_index, int lsb_index)
{
char *pnt;
pnt=GetFacX(buffer, GetFacilityIndex(name, msb_index), GetFacilityIndex(name, lsb_index));
if(!pnt)
	{
	fprintf(stderr,CSIM"GETFACX FACILITY NOT FOUND: '%s[%d:%d]'\n",name,msb_index,lsb_index);
	}
return(pnt);
}

/*********************************************************************************************************************/

void StkFac(unsigned int val, int msb_index, int lsb_index)
{
int i, direction, iter;
struct fac *f, *fprime;
char operand;

model_dirty=getfac_mode;

if((msb_index<0)||(msb_index>=numfacs)||(lsb_index<0)||(lsb_index>=numfacs))
	{
	fprintf(stderr, CSIM"WARNING: STKFAC INDEX ERROR [%d:%d]\n", msb_index, lsb_index);
	simwarnings++;
	return;
	}

if(lsb_index>msb_index)
	{
	direction=-1;
	iter=lsb_index-msb_index+1;
	}
	else
	{
	direction=1;
	iter=msb_index-lsb_index+1;
	}
	
f=facs+lsb_index;	
for(i=0;i<iter;i++)
	{
	if(!(f->attributes&SYM_ALIAS))
		{
		if(!(f->attributes&SYM_CONST))
			{
			operand=(val&1);
			val>>=1;
		
			f->u.v.real_value=operand;
			f->attributes|=SYM_STKFAC;
			}
			else
			{
			fprintf(stderr, CSIM"WARNING: STKFAC ATTEMPT TO REDEFINE CONST '%s'\n", f->name);
			simwarnings++;
			}
		}
		else
		{
		fprime=f->u.alias;
		if(!(fprime->attributes&SYM_CONST))
			{
			operand=(val&1);
			val>>=1;
		
			fprime->u.v.real_value=operand;
			fprime->attributes|=SYM_STKFAC;
			}
			else
			{
			fprintf(stderr, CSIM"WARNING: STKFAC ATTEMPT TO REDEFINE CONST THROUGH '%s'\n", f->name);
			simwarnings++;
			}
		}

	
	f+=direction;
	}
}

void StkFacName(int val, char *name, int msb_index, int lsb_index)
{
StkFac((unsigned int)val, GetFacilityIndex(name, msb_index), GetFacilityIndex(name, lsb_index));
}

/*********************************************************************************************************************/

void StkFacX(char *buffer, int msb_index, int lsb_index)
{
int i, direction, iter;
struct fac *f, *fprime;
char operand;

model_dirty=getfac_mode;

if((msb_index<0)||(msb_index>=numfacs)||(lsb_index<0)||(lsb_index>=numfacs))
	{
	fprintf(stderr, CSIM"WARNING: STKFACX INDEX ERROR [%d:%d]\n", msb_index, lsb_index);
	simwarnings++;
	return;
	}

if(msb_index>lsb_index)
	{
	direction=-1;
	iter=msb_index-lsb_index+1;
	}
	else
	{
	direction=1;
	iter=lsb_index-msb_index+1;
	}
	
f=facs+msb_index;	
for(i=0;i<iter;i++)
	{
	if(!(f->attributes&SYM_ALIAS))
		{
		if(!(f->attributes&SYM_CONST))
			{
			switch(*(buffer++))
				{
				case 0x00:
					return;			/* ran out of runway */
				case '0': operand=0; break;
				case '1': operand=1; break;
				case 'Z':
				case 'z': operand=3; break;
				default:  operand=2; break;
				}
		
			f->u.v.real_value=operand;
			f->attributes|=SYM_STKFAC;
			}
			else
			{
			fprintf(stderr, CSIM"WARNING: STKFAC ATTEMPT TO REDEFINE CONST '%s'\n", f->name);
			simwarnings++;
			}
		}
		else
		{
		fprime=f->u.alias;
		if(!(fprime->attributes&SYM_CONST))
			{
			switch(*(buffer++))
				{
				case 0x00:
					return;			/* ran out of runway */
				case '0': operand=0; break;
				case '1': operand=1; break;
				case 'Z':
				case 'z': operand=3; break;
				default:  operand=2; break;
				}
		
			fprime->u.v.real_value=operand;
			fprime->attributes|=SYM_STKFAC;
			}
			else
			{
			fprintf(stderr, CSIM"WARNING: STKFAC ATTEMPT TO REDEFINE CONST THROUGH '%s'\n", f->name);
			simwarnings++;
			}
		}
	
	f+=direction;
	}
}

void StkFacNameX(char *name, char *buffer, int msb_index, int lsb_index)
{
StkFacX(buffer, GetFacilityIndex(name, msb_index), GetFacilityIndex(name, lsb_index));
}

/*********************************************************************************************************************/

void Unstick(int msb_index, int lsb_index)
{
int i, direction, iter;
struct fac *f, *fprime;

model_dirty=getfac_mode;

if((msb_index<0)||(msb_index>=numfacs)||(lsb_index<0)||(lsb_index>=numfacs))
	{
	fprintf(stderr, CSIM"WARNING: UNSTICK INDEX ERROR [%d:%d]\n", msb_index, lsb_index);
	simwarnings++;
	}

if(lsb_index>msb_index)
	{
	direction=-1;
	iter=lsb_index-msb_index+1;
	}
	else
	{
	direction=1;
	iter=msb_index-lsb_index+1;
	}
	
f=facs+lsb_index;	
for(i=0;i<iter;i++)
	{
	if(!(f->attributes&SYM_ALIAS))
		{
		f->attributes&=(~SYM_STKFAC);
		}
		else
		{
		fprime=f->u.alias;
		fprime->attributes&=(~SYM_STKFAC);
		}
	f+=direction;
	}
}

void UnstickName(char *name, int msb_index, int lsb_index)
{
Unstick(GetFacilityIndex(name, msb_index), GetFacilityIndex(name, lsb_index));
}

/*********************************************************************************************************************/

char *GetFacilityName(int index)
{
return(((index>=0)&&(index<numfacs))?facs[index].name:NULL);
}

int GetFacilityMax(void)
{
return(numfacs);
}

/*********************************************************************************************************************/

void SetSimClockMode(int mode)
{
ignore_clocks=mode;
}

void SetGetFacOrder(int mode)
{
put_before_get=mode;
}

void SetSimDebug(int mode)
{
debug_exec=mode;
}

void SetGetFacMode(int mode)
{
getfac_mode=mode;
}

/*********************************************************************************************************************/

int ImportModel(char *nam)
{
if(load_model(nam))
	{
	fprintf(stderr,CSIM"Running model '%s'...\n",nam);
	return(1);
	}
	else
	{
	fprintf(stderr,CSIM"Could not find model '%s', exiting.\n",nam);
	return(0);
	}
}

void InitializeModelValue(char init)
{
switch(init)
	{
	case '0':	init_value=0; break;
	case '1':	init_value=1; break;
	case 'Z':
	case 'z':	init_value=3; break;
	default:	init_value=2; break;
	}
}

int UnloadModel(void)
{
int i;

if(handle)
	{
	fprintf(stderr, CSIM"Freeing up model resources\n");
	if(pre_op) { free_2(pre_op); pre_op=NULL; }
	if(lat_op) { free_2(lat_op); lat_op=NULL; }
	if(post_op) { free_2(post_op); post_op=NULL; }
	if(facs) { free_2(facs); facs=NULL; }
	if(vcd_lxt) { free_2(vcd_lxt); vcd_lxt=NULL; }
	if(vcd)	{ free_2(vcd); vcd=NULL; }
	if(lxt)	{ lt_close(lxt); lxt=NULL; }
	fprintf(stderr, CSIM"%d error%s, %d warning%s encountered\n",simerrors,(simerrors==1)?"":"s",simwarnings,(simwarnings==1)?"":"s");
	fprintf(stderr, CSIM"Simulation terminated at cycle: %d (+%d).\n",sim_cycle,additional_cycles);
	
	handle=NULL;
	lineno=1;
	numfacs=0; numpre_inst=0; numlat_inst=0; numpost_inst=0;
	num_gates=0;
	simerrors=0; simwarnings=0;
	for(i=0;i<TOKENS;i++)
		{
		num_gates_each_type[i]=0;
		simerror_reported_gates[i]=0;
		}
	debug_exec=0;
	sim_cycle=0; additional_cycles=0;
	model_dirty=0;
	ignore_clocks=0;
	getfac_mode=0;
	init_value=0;

	vcd_tail_stale=0;
	remove_zeros_on_nonvectors=1;
	lxt_remove_zeros_on_nonvectors=1;

	return(0);
	}
	else
	{
	fprintf(stderr, CSIM"No model loaded!\n");
	return(255);
	}
}

/*********************************************************************************************************************/

static FILE *vcdfd=NULL;

static void vcddumpvarname(struct fac *f, char *id, char *s2)
{
char *s;
char ch;
char isnum=1;
char isnum2=1;
int i, len, len2;
char *base=NULL, *fixup;
char *base2=NULL, *fixup2;
int alias;

s=f->name;
alias=f->attributes&SYM_ALIAS;

if(s)
	{
	len=strlen(s);
	base=(char *)cyco_alloca(len+2);
	strcpy(base,f->name);

	fixup=base;
	for(i=0;i<len;i++)
		{
		ch=*fixup;
		if(ch=='/') { *fixup='.'; }
		fixup++;
		}

	fixup--;
	for(i=len-1;i>=0;i--)
		{
		ch=*fixup;
		if(ch=='.')
			{
			if(isnum)
				{
				*fixup='[';
				*(base+len)=']';
				*(base+len+1)=0x00;
				break;
				}
				else
				{
				break;
				}
			}
			else
		if((ch<'0')&&(ch>'9')) {isnum=0;};
		fixup--;
		}	
	}

if(remove_zeros_on_nonvectors)
	{
	if(s2)
		{
		len2=strlen(s2);
		base2=(char *)cyco_alloca(len2+2);
		strcpy(base2,s2);
	
		fixup2=base2;
		for(i=0;i<len2;i++)
			{
			ch=*fixup2;
			if(ch=='/') { *fixup2='.'; }
			fixup2++;
			}
	
		fixup2--;
		for(i=len2-1;i>=0;i--)
			{
			ch=*fixup2;
			if(ch=='.')
				{
				if(isnum2)
					{
					*fixup2='[';
					*(base2+len2)=']';
					*(base2+len2+1)=0x00;
					break;
					}
					else
					{
					break;
					}
				}
				else
			if((ch<'0')&&(ch>'9')) {isnum2=0;};
			fixup2--;
			}	
	
		if(!strcmp(fixup,"[0]"))
			{
			if((fixup-base)==(fixup2-base2))
				{
				if(strncmp(s, s2, (fixup2-base2)))
					{
					*fixup=0;
					}
				}
				else
				{
				*fixup=0;
				}
			}

		}
		else
		{
		if(!strcmp(fixup,"[0]"))
			{
			*fixup=0;
			}
		}
	}
	
fprintf(vcdfd,"$var "); 
if((f->attributes&SYM_REG)&&(!alias))
	{
	fprintf(vcdfd, "reg ");
	}
	else
	{
	fprintf(vcdfd, "wire ");
	}

if (base) fprintf(vcdfd, "1 %s %s $end\n",id, base);
}


FILE *OpenVCD(char *name)
{
static char vcdid[8]={'!',0,0,0,0,0,0,0};
int i, index;

if(name!=NULL)
	{
	vcdfd=fopen(name,"wb");
	}

if(vcdfd==NULL)
	{
	vcdfd=stdout;	/* sorry, it's going somewhere! */
	}


if(vcdfd)
	{
	struct vcd *v;
	struct fac *f, *faclast;
        time_t walltime;
   
        time(&walltime);
        fprintf(vcdfd, "$date\n");
        fprintf(vcdfd, "\t%s",asctime(localtime(&walltime)));
        fprintf(vcdfd, "$end\n");

	fprintf(vcdfd, "$version\n\t"CSIM_VERSION"\n$end\n"
		"$timescale\n\t1ns\n$end\n");

	vcd=v=calloc_2(numfacs, sizeof(struct vcd));
	f=facs;
	for(index = 0; index < numfacs ; index++)
                {
		v->prevval=0xff;	/* will never match vcd compare, so the dump will happen 1st cyc */

		if(!(f->attributes&SYM_ALIAS))
			{
			strcpy(v->identifier, vcdid);

			for(i=0;i<8;i++)	/* increment vcd id for next fac */
				{
				vcdid[i]++;
				if(vcdid[i]!=127) break;
				vcdid[i]='!';
				if(vcdid[i+1]==0x00)
					{
					vcdid[i+1]='!';
					break;
					}
				}
			}
		v++;
		f++;
                }

	if(vcd_bigendian)
		{
		v=vcd;
		f=facs;
		faclast=facs+numfacs-1;
		}
		else
		{
		v=vcd+numfacs-1;
		faclast=f=facs+numfacs-1;
		}

	for(index = 0; index < numfacs ; index++)
                {
		if(f->attributes&SYM_ALIAS)
			{
			strcpy(v->identifier, vcd[f->u.alias-facs].identifier);
			}

		if(skip_vcd_tmp_sigs)
			{
			if(strchr(f->name,'!'))
				{
				f->attributes|=SYM_TMPSIG;
				}
				else
				{
				vcddumpvarname(f, v->identifier, (f!=faclast)?(f+1)->name:NULL);
				}
			}
			else
			{
			vcddumpvarname(f, v->identifier, (f!=faclast)?(f+1)->name:NULL);
			}

		if(vcd_bigendian)
			{
			v++;
			f++;
			}
			else
			{
			v--;
			f--;
			}
                }
	fprintf(vcdfd,"$enddefinitions $end\n");
	}

return(vcdfd);
}


void WriteVCDFrame(FILE *vcdfd, int cycle)
{
int i;
char emitnum=1;
struct fac *f;
struct vcd *v;
char nowval;

f=facs;
v=vcd;
if(vcdfd)
for(i=0;i<numfacs;i++)
	{
	if((!(f->attributes&SYM_TMPSIG))&&(!(f->attributes&SYM_ALIAS)))
		{
		nowval=f->u.v.real_value;
		if(v->prevval!=nowval)
			{
			v->prevval=nowval;

			if(emitnum)
				{
				fprintf(vcdfd, "#%d\n", cycle); emitnum=0;
				}

			fprintf(vcdfd, "%c%s\n", "01xz"[nowval&3], v->identifier);			
			}
		}

	f++;
	v++;
	}

vcd_tail_stale=emitnum;
vcd_stale_time=cycle;
}


void CloseVCD(FILE *vcdfd)
{
/* printf("TAIL STALE: %d %d\n", vcd_tail_stale, vcd_stale_time); */
if((vcdfd)&&(vcd_tail_stale))
	{
	fprintf(vcdfd, "#%d\n", vcd_stale_time); vcd_tail_stale=0;
	}
if((vcdfd)&&(vcdfd!=stdout)) fclose(vcdfd);
vcdfd=NULL;
}


void SetVCDtmpsigMode(int tmpsigMode)
{
skip_vcd_tmp_sigs = tmpsigMode;
}

void SetVCDendian(int endianMode)
{
vcd_bigendian=endianMode;
}

void SetVCDremoveNonvectorZeroBits(int zeroMode)
{
remove_zeros_on_nonvectors=zeroMode;
}

/*********************************************************************************************************************/

/*********************************************************************************************************************/

static void lxtdumpvarname(struct fac *f, char *s2)
{
char *s;
char ch;
char isnum=1;
char isnum2=1;
int i, len, len2;
char *base=NULL, *fixup;
char *base2=NULL, *fixup2;
int alias;
int bitnum = -1;

s=f->name;
alias=f->attributes&SYM_ALIAS;

if(s)
	{
	len=strlen(s);
	base=(char *)cyco_alloca(len+2);
	strcpy(base,f->name);

	fixup=base;
	for(i=0;i<len;i++)
		{
		ch=*fixup;
		if(ch=='/') { *fixup='.'; }
		fixup++;
		}

	fixup--;
	for(i=len-1;i>=0;i--)
		{
		ch=*fixup;
		if(ch=='.')
			{
			if(isnum)
				{
				*fixup='[';
				*(base+len)=']';
				*(base+len+1)=0x00;
				sscanf(fixup+1, "%d", &bitnum);
				break;
				}
				else
				{
				break;
				}
			}
			else
		if((ch<'0')&&(ch>'9')) {isnum=0;};
		fixup--;
		}	
	}

if(lxt_remove_zeros_on_nonvectors)
	{
	if(s2)
		{
		len2=strlen(s2);
		base2=(char *)cyco_alloca(len2+2);
		strcpy(base2,s2);
	
		fixup2=base2;
		for(i=0;i<len2;i++)
			{
			ch=*fixup2;
			if(ch=='/') { *fixup2='.'; }
			fixup2++;
			}
	
		fixup2--;
		for(i=len2-1;i>=0;i--)
			{
			ch=*fixup2;
			if(ch=='.')
				{
				if(isnum2)
					{
					*fixup2='[';
					*(base2+len2)=']';
					*(base2+len2+1)=0x00;
					break;
					}
					else
					{
					break;
					}
				}
				else
			if((ch<'0')&&(ch>'9')) {isnum2=0;};
			fixup2--;
			}	
	
		if(!strcmp(fixup,"[0]"))
			{
			if((fixup-base)==(fixup2-base2))
				{
				if(strncmp(s, s2, (fixup2-base2)))
					{
					*fixup=0;
					bitnum=-1;
					}
				}
				else
				{
				*fixup=0;
				bitnum=-1;
				}
			}

		}
		else
		{
		if(!strcmp(fixup,"[0]"))
			{
			*fixup=0;
			bitnum=-1;
			}
		}
	}
	
if(base)
	{
	if(!(f->attributes&SYM_ALIAS))
		{
		f->lxtsymbol = lt_symbol_add(lxt, base, 0, bitnum, bitnum,  LT_SYM_F_BITS);
		}
		else
		{
		if(f->u.alias->lxtsymbol)
			{
			f->lxtsymbol = lt_symbol_add(lxt, base, 0, bitnum, bitnum,  LT_SYM_F_BITS);
			f->lxtsymbol->flags = LT_SYM_F_ALIAS;
			f->lxtsymbol->aliased_to = f->u.alias->lxtsymbol;
			}
		}
	}
}


void *OpenLXT(char *name)
{
int index;
int j;

lxt = lt_init(name);

if(lxt)
	{
	struct vcd *v;
	struct fac *f, *faclast;

	lt_set_clock_compress(lxt);
	lt_set_chg_compress(lxt);
	lt_symbol_bracket_stripping(lxt, 1);
	vcd_lxt=v=calloc_2(numfacs, sizeof(struct vcd));
	for(index = 0; index < numfacs ; index++)
                {
		v->prevval=0xff;	/* will never match vcd compare, so the dump will happen 1st cyc */
		v++;
                }

	faclast=facs+numfacs-1;

	for(j=0;j<2;j++)
		{
		f=facs;
		for(index = 0; index < numfacs ; index++)
	                {
			if( ((!(f->attributes&SYM_ALIAS))&&(j==0)) || ((f->attributes&SYM_ALIAS)&&(j==1)) )
			if(skip_lxt_tmp_sigs)
				{
				if(strchr(f->name,'!'))
					{
					f->attributes|=SYM_TMPSIG;
					}
					else
					{
					lxtdumpvarname(f, (f!=faclast)?(f+1)->name:NULL);
					}
				}
				else
				{
				lxtdumpvarname(f, (f!=faclast)?(f+1)->name:NULL);
				}
	
			f++;
	                }
		}
	}

return((void *)lxt);
}


void WriteLXTFrame(int cycle)
{
int i;
char emitnum=1;
struct fac *f;
struct vcd *v;
char nowval;
char value[2];

f=facs;
v=vcd_lxt;
if(lxt)
	{
	lt_set_time(lxt, cycle);
	value[1]=0;
	for(i=0;i<numfacs;i++)
		{
		if((!(f->attributes&SYM_TMPSIG))&&(!(f->attributes&SYM_ALIAS)))
			{
			nowval=f->u.v.real_value;
			if(v->prevval!=nowval)
				{
				v->prevval=nowval;
				value[0] = "01xz"[nowval&3];	
				lt_emit_value_bit_string(lxt, f->lxtsymbol, 0, value);
				}
			}
		f++;
		v++;
		}
	}
}


void CloseLXT(void)
{
if(lxt) 
	{
	lt_close(lxt);
	lxt=NULL;
	}
}


void SetLXTtmpsigMode(int tmpsigMode)
{
skip_lxt_tmp_sigs = tmpsigMode;
}

void SetLXTremoveNonvectorZeroBits(int zeroMode)
{
lxt_remove_zeros_on_nonvectors=zeroMode;
}

/*********************************************************************************************************************/

