#ifndef CSIM_H
#define CSIM_H

#define CSIM_VERSION "CSIM Version 1.3.34"

int ImportModel(char *nam);		/* load model into memory */
void InitializeModelValue(char init);	/* set default node value (for sanity, latches are ALWAYS ZERO) */
int UnloadModel(void);			/* unload model from memory */

int SimulateModel(int cycles);		/* simulate for "cycles" number of cycles */
int SimulateCones(void);		/* simulate only combinatorial logic cones */
int SimulateLatches(void);		/* simulate only latches */
int IncrementCycleCount(int cycles);	/* add cycles to internal cycle count */

void SetSimClockMode(int mode);		/* 0=normal edge; 1=ignore clocks, latch every call to Simulate_Model() */
void SetSimDebug(int mode);		/* 0=none, 1=print messages */
void SetGetFacMode(int mode);		/* 0=normal operation, 1=re-eval model if a Get follows a Stick */
void SetGetFacOrder(int mode);		/* 0=Get/Stick/Clock, 1=Stick/Get/Clock */

int GetFacilityMax(void);		/* returns max # of facs in model */
int GetFacilityIndex(char *name, int which); /* get fac index corresponding name + bitpos */
char *GetFacilityName(int index);	/* get name corresponding with an index value */

int GetFac(int msb_index, int lsb_index);	/* get facility value (right aligned, one bit per pos=0/1 bits only) */
int GetFacName(char *name, int msb_index, int lsb_index); /* same as getfac, but takes a name string */

char *GetFacX(char *buffer, int msb_index, int lsb_index); /* get fac value as a string--returns 01XZ values */
char *GetFacNameX(char *name, char *buffer, int msb_index, int lsb_index); /* same as getfacx, but takes a name string */

void StkFac(unsigned int val, int msb_index, int lsb_index); /* sticks a facility to a right aligned value */
void StkFacName(int val, char *name, int msb_index, int lsb_index); /* same as stkfac, but takes a name string */

void StkFacX(char *buffer, int msb_index, int lsb_index); /* sticks a facility using a string filled with 01XxZz */
void StkFacNameX(char *name, char *buffer, int msb_index, int lsb_index); /* same as stkfacx, but takes a name string */

void Unstick(int msb_index, int lsb_index); /* unstick a facility */
void UnstickName(char *name, int msb_index, int lsb_index); /* unstick with a name string */

FILE *OpenVCD(char *name);				/* open a vcd file (NULL=stdout) */
void WriteVCDFrame(FILE *vcdfd, int cycle);             /* write a frame to the vcd      */
void CloseVCD(FILE *vcdfd);				/* close the vcd + dump maxtime  */
void SetVCDtmpsigMode(int tmpsigMode);			/* 1 = only dump user-defined sigs, 0 = dump all sigs */
void SetVCDendian(int endianMode);			/* 1 = big endian [0:x], 0 = little endian [x:0]      */
void SetVCDremoveNonvectorZeroBits(int zeroMode);	/* 1 = remove [0] in VCD if there's no [1] for that signame */

void *OpenLXT(char *name);				/* open an lxt file (NULL=failure) */
void WriteLXTFrame(int cycle);				/* write a frame to the lxt        */
void CloseLXT(void);					/* close the lxt                   */
void SetLXTtmpsigMode(int tmpsigMode);			/* 1 = only dump user-defined sigs, 0 = dump all sigs */
void SetLXTremoveNonvectorZeroBits(int zeroMode);	/* 1 = remove [0] in LXT if there's no [1] for that signame */

#endif

