/* 
 * Copyright (c) Tony Bybell 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include "symbol.h"

int exec_quad(struct quad *q)
{
int value=0;

while(q)
	{
	switch(q->oper)
		{
		case OP_EQ:
			q->res->value = q->op1->value;
			break;
		case OP_NEG:
			q->res->value = -(q->op1->value);
			break;
		case OP_ADD:
			q->res->value = q->op1->value + q->op2->value;
			break;
		case OP_SUB:
			q->res->value = q->op1->value - q->op2->value;
			break;
		case OP_MUL:
			q->res->value = q->op1->value * q->op2->value;
			break;
		case OP_DIV:
			q->res->value = q->op1->value / q->op2->value;
			break;
		case OP_MOD:
			q->res->value = q->op1->value % q->op2->value;
			break;
		case OP_LSH:
			q->res->value = q->op1->value << q->op2->value;
			break;
		case OP_RSH:
			q->res->value = q->op1->value >> q->op2->value;
			break;
		case OP_AND:
			q->res->value = q->op1->value & q->op2->value;
			break;
		case OP_XOR:
			q->res->value = q->op1->value ^ q->op2->value;
			break;
		case OP_OR:
			q->res->value = q->op1->value | q->op2->value;
			break;
		case OP_NOT:
			q->res->value = ~(q->op1->value);
			break;
		case OP_NNOT:
			q->res->value = !(q->op1->value);
			break;
		case OP_LT:
			q->res->value = q->op1->value < q->op2->value;
			break;
		case OP_GT:
			q->res->value = q->op1->value > q->op2->value;
			break;
		case OP_LET:
			q->res->value = q->op1->value <= q->op2->value;
			break;
		case OP_GET:
			q->res->value = q->op1->value >= q->op2->value;
			break;
		case OP_EQV:
			q->res->value = q->op1->value == q->op2->value;
			break;
		case OP_NEQ:
			q->res->value = q->op1->value != q->op2->value;
			break;
		case OP_AAND:
			q->res->value = q->op1->value && q->op2->value;
			break;
		case OP_OOR:
			q->res->value = q->op1->value || q->op2->value;
			break;
		default:
			fprintf(stderr, "GEN Internal error: quad type %d encountered.\n",q->oper);
			exit(FATAL_ERROR);
		}
	value=q->res->value;
	q=q->next;
	}
return(value);
}

void exec_stmt(struct statement *s)
{
while(s)
	{
	switch(s->type)
		{
		case ST_TEXT:
			printf("%s",s->u.text);
			break;
		case ST_MACRO:
			if(!s->u.mac->op1->string)
				{
				printf("%d",exec_quad(s->u.mac)); 
				}
				else
				{
				printf("%s",s->u.mac->op1->string); 
				}
			break;
		case ST_EVAL:
			exec_quad(s->u.eval);
			break;
		case ST_GEN:
			exec_gen(s->u.gen);
			break;
		case ST_GST:
			exec_gst(s->u.gst);
			break;
		case ST_IF:
			exec_if(s->u.ifstmt);
			break;
		case ST_WHILE:
			exec_while(s->u.whilestmt);
			break;
		case ST_FOR:
			exec_for(s->u.forstmt);
			break;

		default:
			fprintf(stderr, "GEN Internal error: statement type %d encountered!\n",s->type);
			exit(FATAL_ERROR);
		}
	s=s->next;
	}
}

void exec_gen(struct generate *g)
{
int c1, c2, i;

c1=exec_quad(g->from);
c2=exec_quad(g->to);

if(c1<c2)
	{
	for(i=c1;i<=c2;)
		{
		g->sym->value=i;
		exec_stmt(g->next);
		i=(g->sym->value)+1;
		}
	}
	else
	{
	for(i=c1;i>=c2;)
		{
		g->sym->value=i;
		exec_stmt(g->next);
		i=(g->sym->value)-1;
		}
	}
}

void exec_gst(struct generate_string *g)
{
struct generate_string *g2;
struct gen_str *iter, *iter2;
int i, depth=0;

iter=g->head;

while(iter)
	{
	g->sym->string=iter->str;

	g2=g->gnext;	/* skip first entry */
	while(g2)	/* adimittedly this is an m*(n**2) hack but it works.. */
		{
		iter2=g2->head;
		for(i=0;i<depth;i++)
			{
			if(iter2) iter2=iter2->next; else break;
			}			
		if(iter2) g2->sym->string=iter2->str;		

		g2=g2->gnext;
		}

	exec_stmt(g->next);
	depth++;
	iter=iter->next;
	}

while(g)
	{
	g->sym->string=NULL;
	g=g->gnext;
	}
}


void exec_if(struct ifstmt *ifs)
{
if(exec_quad(ifs->cond))
	{
	exec_stmt(ifs->next);
	}
	else
	{
	exec_stmt(ifs->elsestmt);
	}
}

void exec_while(struct whilestmt *whs)
{
while(exec_quad(whs->cond))
	{
	exec_stmt(whs->next);
	}
}

void exec_for(struct forstmt *f)
{
for(exec_quad(f->start);exec_quad(f->cond);exec_quad(f->inc))
	{
	exec_stmt(f->next);
	}
}

