%{

/* 
 * Copyright (c) Tony Bybell 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include "y.tab.h"
#include "symbol.h"
int yylineno=1;
int nesting=0;
int printlineno=0;
static int plne=-1;	/* previous line number */
%}
%s GEN
%s TEX
%s TEX2
%s MAC
%%

[\t\r ]*\#mstart[\t\r ]*\;[^\n]*\n { BEGIN TEX; yylineno++; }
<TEX>[\t\r ]*\#mend[\t\r ]*\;[^\n]* { BEGIN 0; }

<TEX2>[^\#^\`^\n]+ { 
		char *t;
		int l=strlen(yytext);

		if((printlineno)&&(plne!=yylineno))
			{
			t=(char *)malloc(l+1+16);
			sprintf(t, "<<[%d]>>%s", yylineno, yytext);
			plne=yylineno;
			}
			else
			{
			strcpy(t=(char *)malloc(l+1),yytext);
			}

		yylval.pnt=t;
		return TEXT; }
<TEX2>[\n]+ 	{ 
		char *t;
		int l;

		strcpy(t=(char *)malloc((l=strlen(yytext))+1),yytext);
		yylval.pnt=t;
		yylineno+=l;
		return TEXT; }
<TEX2>[\t\r ]*\#endif[\t\r ]*\;[^\n]*\n { DEBUG("EndIf\n"); yylineno++; return ENDIF; }
<TEX2>[\t\r ]*\#else[\t\r ]*\;[^\n]*\n { DEBUG("Else\n"); yylineno++; return ELSE; }
<TEX2>[\t\r ]*\#if	{ DEBUG("If\n"); BEGIN GEN; return IF; }
<TEX2>[\t\r ]*\#while	{ DEBUG("While\n"); BEGIN GEN; return WHILE; }
<TEX2>[\t\r ]*\#wend[\t\r ]*\;[^\n]*\n { DEBUG("Wend\n"); yylineno++; return WEND; }
<TEX2>[\t\r ]*\#for	{ DEBUG("For\n"); BEGIN GEN; return FOR; }
<TEX2>[\t\r ]*\#next[\t\r ]*\;[^\n]*\n { DEBUG("Next\n"); yylineno++; return NEXT; }
<TEX2>[\t\r ]*\#eval	{ DEBUG("Eval\n");BEGIN GEN; return EVAL; }
<TEX2>[\t\r ]*\#generate { DEBUG("Generate\n");BEGIN GEN; nesting++; return GENERATE; }
<TEX2>[\t\r ]*#genend	{ DEBUG("END2\n"); BEGIN GEN; nesting--; return END; }
<TEX2>\`	{ DEBUG("MACRO\n"); BEGIN MAC; return MACRO; }

<TEX>[^\#^\n]+	{ 
		char *t;
		int l=strlen(yytext);

		if((printlineno)&&(plne!=yylineno))
			{
			t=(char *)malloc(l+1+16);
			sprintf(t, "<<[%d]>>%s", yylineno, yytext);
			plne=yylineno;
			}
			else
			{
			strcpy(t=(char *)malloc(l+1),yytext);
			}

		yylval.pnt=t;
		return TEXT; }
<TEX>[\n]+ 	{ 
		char *t;
		int l;

		strcpy(t=(char *)malloc((l=strlen(yytext))+1),yytext);
		yylval.pnt=t;
		yylineno+=l;
		return TEXT; }

<TEX>[\t\r ]*\#endif[\t\r ]*\;[^\n]*\n { DEBUG("EndIf\n"); yylineno++; return ENDIF; }
<TEX>[\t\r ]*\#else[\t\r ]*\;[^\n]*\n { DEBUG("Else\n"); yylineno++; return ELSE; }
<TEX>[\t\r ]*\#if	{ DEBUG("If\n"); BEGIN GEN; return IF; }
<TEX>[\t\r ]*\#while	{ DEBUG("While\n"); BEGIN GEN; return WHILE; }
<TEX>[\t\r ]*\#wend[\t\r ]*\;[^\n]*\n { DEBUG("Wend\n"); yylineno++; return WEND; }
<TEX>[\t\r ]*\#for	{ DEBUG("For\n"); BEGIN GEN; return FOR; }
<TEX>[\t\r ]*\#next[\t\r ]*\;[^\n]*\n { DEBUG("Next\n"); yylineno++; return NEXT; }
<TEX>[\t\r ]*\#eval	{ DEBUG("Eval\n");BEGIN GEN; return EVAL; }
<TEX>[\t\r ]*\#generate	{ DEBUG("Generate\n");BEGIN GEN; nesting++; return GENERATE; } 

<GEN>begin	{ DEBUG("BEGIN\n"); return BEG; }
<GEN>to		{ return TO; }

<GEN>\;[^\n]*\n	{  DEBUG("Nesting level: %d\n",nesting); yylineno++;
		   if(nesting) {BEGIN TEX2;} else {BEGIN TEX;}; return SEMI; }
<GEN>[A-Za-z_]+[A-Za-z_0-9]* { 
                  yylval.pnt=symfind(yytext); 
		  DEBUG("LABEL\n"); 
		  return LABEL; }
<GEN>[0-9]+	{ yylval.i=atoi(yytext); return NUMBER; }
<GEN>$[0-9A-Fa-f]+ { sscanf(yytext+1,"%x",&(yylval.i)); return HEX; }
<GEN>\"[^\"]*\" {
		int len=strlen(yytext+1);
		char *m=(char *)malloc(len);
		if(len>1) strncpy(m, yytext+1, len-1);
		*(m+len-1)=0;
		yylval.pnt=m;
		return STRING;
		}
<GEN>"<<"    { return LSH; }
<GEN>">>"    { return RSH; }
<GEN>"+"     { return ADD; }
<GEN>"-"     { return SUB; }
<GEN>"*"     { return MPY; }
<GEN>"/"     { return DIV; }
<GEN>"%"     { return MOD; }
<GEN>"&"     { return AND; }
<GEN>"|"     { return OR; }
<GEN>"^"     { return XOR; }
<GEN>"("     { return LP; }
<GEN>")"     { return RP; }
<GEN>"<"     { return LT; }
<GEN>">"     { return GT; }
<GEN>"<="    { return LET; }
<GEN>">="    { return GET; }
<GEN>"=="    { return EQV; }
<GEN>"!="    { return NEQ; }
<GEN>"&&"    { return AAND; }
<GEN>"||"    { return OOR; }	
<GEN>"="     { return EQ; }
<GEN>"~"     { return NOT; }
<GEN>"!"     { return NNOT; }
<GEN>","     { return COMMA; }

<GEN>[ \t\r]+	;
<GEN>\n		{ yylineno++; }
<GEN>.		return yytext[0];

<MAC>[A-Za-z_]+[A-Za-z_0-9]* { 
		  yylval.pnt=symfind(yytext);
		  DEBUG("MACLABEL\n"); 
		  return LABEL; }
<MAC>[0-9]+     { DEBUG("MACNUM\n"); yylval.i=atoi(yytext); return NUMBER; }
<MAC>\`		{ DEBUG("REVERT\n"); BEGIN TEX2; return MACROEND; }

<MAC>$[0-9A-Fa-f]+ { sscanf(yytext+1,"%x",&(yylval.i)); return HEX; }
<MAC>"<<"    { return LSH; }
<MAC>">>"    { return RSH; }
<MAC>"+"     { return ADD; }
<MAC>"-"     { return SUB; }
<MAC>"*"     { return MPY; }
<MAC>"/"     { return DIV; }
<MAC>"%"     { return MOD; }
<MAC>"&"     { return AND; }
<MAC>"|"     { return OR; }
<MAC>"^"     { return XOR; }
<MAC>"("     { return LP; }
<MAC>")"     { return RP; }
<MAC>"<"     { return LT; }
<MAC>">"     { return GT; }
<MAC>"<="    { return LET; }
<MAC>">="    { return GET; }
<MAC>"=="    { return EQV; }
<MAC>"!="    { return NEQ; }
<MAC>"&&"    { return AAND; }
<MAC>"||"    { return OOR; }	
<MAC>"="     { return EQ; }
<MAC>"~"     { return NOT; }
<MAC>"!"     { return NNOT; }

<MAC>[ \t\r]+	;
<MAC>\n		{ yylineno++; }

.		return yytext[0]; 
[ \t\r]+	;
\n		{ yylineno++; }


%%
