#include <sys/time.h>
#include <fcntl.h>
#include "symbol.h"

int didimport=0;
int isnl=0, prevnl=0;	/* for consecutive newline stripping */


struct vpp_filename *libchain=NULL, *libextchain=NULL;
char *lineroot=NULL, *linepos=NULL;
char *libname=NULL;

struct antext *anroot=NULL, *anpnt=NULL;  /* for module text */

/*
 * add token to the analyzed verilog buffer..
 */
void addtoken(void)
{
char c;

if((isnl)&&(prevnl)) {isnl=0; return; }

prevnl=isnl;
isnl=0;

strcpy(linepos,yytext);
linepos+=strlen(yytext);

if(!parameter_module_in_effect)
	{
	if(!anroot)
		{
		c=*yytext;
		if(c=='\n') return;
		anroot=anpnt=(struct antext *)calloc(1,sizeof(struct antext));
		}
		else
		{
		anpnt->next=(struct antext *)calloc(1,sizeof(struct antext));
		anpnt=anpnt->next;
		}

	c=*yytext;
	if((c==' ')||(c=='\t')||(c=='\r'))
		{
		anpnt->text=(char *)calloc(1,2);
		*(anpnt->text)=' ';
		anpnt->ssize=1;
		}
		else
		{
		anpnt->text=(char *)calloc(1,yyleng+1);
		memcpy(anpnt->text,yytext,yyleng);
		anpnt->ssize=yyleng;
		}
	}
}

/*
 * flush analyzed verilog to the library with the ver header
 * prepended to it.
 */
void flushtolib(void)
{
struct antext *an;
char *fname;
FILE *lib;
struct timeval tp;

currentmod->re_analyze_text = anroot;

if((!libname)||(parameter_module_in_effect))
	{
	anroot=NULL;
	return;  /* do not analyze to library file */
	}

gettimeofday(&tp,NULL);
fname=calloc(1,strlen(libname)+strlen(currentmod->name)+2+1);
sprintf(fname, "%s%s.v", libname, currentmod->name);

if((lib=fopen(fname,"w")))
	{
	an=anroot;	/* don't free it up like was done originally since it might be needed for param subst */
	while(an)
		{
		fwrite((void *)an->text,an->ssize,1,lib);
		an=an->next;
		}
	anroot=NULL;
	fclose(lib);
	}
	else
	{
	mesg("Could not open output file for '%s', Exiting.\n",fname);
	exit(255);
	}
free(fname);
}

void getmodfromlib(struct module *type)
{
struct vpp_filename *l, *e;
FILE *lib;
char *str, *ext;
int slen;
char *filename;

if(type->missing) return;       /* errored previously */

str=type->name;

if(!(l=libchain)) 
	{
	type->missing=~0;
	warn("| Import '%s' (failed: no library search path)\n",str);
	return; /* there's no library path to search! */
	}

slen=strlen(str);

while(l)
{
e=libextchain;
do
	{
	ext = e?e->name:"";
	filename=(char *)malloc(slen+1+strlen(l->name)+strlen(ext)+1);
	sprintf(filename, "%s/%s%s", l->name, str, ext);

	if((lib=fopen(filename,"rb")))
		{
	        vpp_update_yyfilename(filename);
		goto import;	/* found module! */
		}
		else
		if((l->next)||((e)&&(e->next)))
			{
			continue;
			}
		else
	                {
			type->missing=~0;
			warn("| Import '%s' (failed)\n",str);
			didimport++;
			free(filename);
			return;
	                }
	} while((e)&&(e=e->next));
l=l->next;
}

import: yylineno=1;
vpp_main(lib, filename);
yyrestart(lib);
yyparse();
fclose(lib);

if(yyerrors) 
        {
        mesg("** Library module '%s' is mangled.  Exiting.\n",filename);
	free(filename);
        exit(255);
        }
	else
	{
	warn("| Import '%s' (ok)\n",filename);
	free(filename);
	didimport++;
	}
}


void import(struct module *cu) /* recursively from modroot */
{
struct component *comp;

if((!cu)||(cu->visited)) return; /* flatten will inform on recursion errors */
cu->visited = ~0;

comp=cu->comp;
while(comp) 
        {
	if(!comp->type->defined)
		{
		getmodfromlib(comp->type);
		}
        import(comp->type);
        comp=comp->next;
        }

cu->visited = 0;
}


void reparse_param_modules(struct module *modroot)
{
struct param_comp_chain *pcc;
int doln = (param_comp_chain!=NULL);
int imported_at_least_one;

do
{
imported_at_least_one=0;
pcc = param_comp_chain;
while(pcc)
	{
	struct module *m = simple_modfind(modroot, pcc->fullname);

	if(m)
		{
		pcc->successfully_imported=~0;
		}
		else
		{
		if((m = simple_modfind(modroot, pcc->basename)))
			{
			FILE *thandle;
			struct antext *an = m->re_analyze_text;

			pcc->successfully_imported=~0;
			m->isexpanded = ~0;

			thandle=tmpfile(); 
			if(!thandle)
				{
			        mesg("** Could not open parameter expansion module tempfile.  Exiting.\n");
			        exit(255);
				}

			while(an)
				{
				fwrite((void *)an->text,an->ssize,1,thandle);
				an=an->next;
				}

			fseek(thandle, 0, SEEK_SET);
			parameter_module_in_effect = pcc;
			yylineno=1;
		        vpp_update_yyfilename(pcc->fullname);
			vpp_main(thandle, pcc->fullname);
			yyrestart(thandle);
			yyparse();
			fclose(thandle);

			if(yyerrors) 
			        {
			        mesg("** Parameter expansion module '%s' (via '%s') is mangled.  Exiting.\n",pcc->fullname, pcc->basename);
			        exit(255);
			        }
				else
				{
				warn("| Expanded '%s' into '%s' (ok)\n",pcc->basename, pcc->fullname);
				imported_at_least_one = ~0;
				}
			}
		}	
	pcc=pcc->next;
	}
} while(imported_at_least_one); /* in order to get transitive closure */

if(doln)
	{
	parameter_module_in_effect = NULL;
	warn("\n");
	}
}
