/* 
 * Copyright (c) Tony Bybell, Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/*
 * main.c
 * 19/08/97ajb
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "symbol.h"
#include "flatten.h"
#include "vpp_common.h"

#define PLURAL(x) (x),(((x)!=1)?"s":"")

extern int yydebug;

int yyerrors=0, resynced=0, num_includes=0, preproc_only=0;
struct filename *filename=NULL;


void yyerror(char *err)
{
char *buf, *ch, *ch2, v;
int l,col=0;

ch2=buf=(char *)calloc(1,l=linepos-lineroot+1);
ch=lineroot;

do
{
v=*(ch++);
if(!((v==' ')||(v=='\t')||(v=='\r')||(v==0x00))) v=' ';
if(v=='\t') { col=(col+8)&(~0x07); } else col++;
*(ch2++)=v;
} while(v);

*(ch2-2-(strlen(yytext)-1))='^';

if(!strcmp("*/",yytext))
	{
	mesg("** Comment end detected outside of a comment in '%s' on line %d.%d:\n",yyfilename, yylineno,col-1);
	}
	else
	{
	mesg("** %s near token '%s' in '%s' on line %d.%d:\n",err,yytext,yyfilename, yylineno,col-1);
	}

warn("%s\n%s\n",lineroot,buf);

free(buf);
}

int yywrap(void)
{
return(1);
}


static void insert_args(char ***argv, int *argc, int i);

int main(int argc, char **argv)
{
struct module *flattenfrom=NULL;
char *flattenname=NULL;
struct filename *lastfilename=NULL;
int stdincompile=0;
struct module *currentmodcache;
int supressstats=0;
int outputvhdl=0;
int i;

for(i=1;i<argc;i++)
	{
	int arglen = strlen(argv[i]);

	if((!strcmp(argv[i],"-h"))||(!strcmp(argv[i],"-help")))
		{
                warn( "Usage:\n"
                      "------\n"
                      "%s [input filename(s)] [options]\n\n"
                      "-\t\taccept input file from stdin\n"
                      "-a libname\talso analyze to library \"lib/libname\"\n"
                      "-c\t\tenable VCD style compressed ids\n"
                      "-dff\t\tenable \"latch\"/\"latch_n\" keywords for cycle sim latches\n"
                      "-h[elp]\t\tprints this screen\n"
                      "-f name\t\tforce flatten from module \"name\" else autodetect\n"
		      "+define+x=y\tequivalent to `define X Y in source\n"
		      "+incdir+dirname\tadd dirname to include search path\n"
		      "+libext+ext\tadd ext to filename when searching for files\n"
		      "-license\tprint VPP license to stdout\n"
                      "-n\t\tdo not output an ivf file to disk\n"
		      "-pica\t\tgenerate special wire gates (needed for vsim *only*)\n"
                      "-s\t\tsupress parse statistics table\n"
		      "-snr\t\tskip stats on modules not instantiated\n"
                      "-v\t\toutput vhdl translation to stdio\n"
		      "-y dirname\tadd directory to source input path\n"
		      "-Dx=y\t\tequivalent to `define X Y in source\n"
                      "-E\t\tSend preprocessed verilog to stdout\n"
                      "-F file\t\tread command line args from file\n"
                      "\n",argv[0]);
		exit(0);
		}
	else
	if(!strcmp(argv[i],"-c"))
		{
		usecompressed_ids=~0;
		}
	else
	if(!strcmp(argv[i],"-dff"))
		{
		uselatchkw=~0;
		}
	else
	if(!strcmp(argv[i],"-snr"))
		{
		skipnorefs=~0;
		}
	else
	if(!strcmp(argv[i],"-n"))
		{
		noivf=~0;
		}
	else
	if(!strcmp(argv[i],"-E"))
		{
		preproc_only=~0;
		dont_add_comments=~0;
		}
	else
	if(!strcmp(argv[i],"-pica"))
		{
		mode_pica_vsim_heritage=~0;
		}
	else
	if(!strcmp(argv[i],"-f"))
		{
		i++;
		if(i<argc)
			{	
			flattenname=argv[i];
			}
			else
			{
			mesg("** Missing module name after -f option!\n");
			exit(1);
			}	
		}
	else
	if(!strcmp(argv[i],"-F"))
		{
		i++;
		if(i<argc)
                        {
                        /* insert content of file into the argv[] array: */
                        if (++num_includes > 50)
                           {
                           mesg("** Too many -F options encountered (%d),\n"
                                "   possibly infinite recursion detected\n",
                                num_includes);
                           exit(1);
                           }
			insert_args(&argv, &argc, i);
			}	
			else
			{
			mesg("** Missing file name after -F option!\n");
			exit(1);
			}	
		}
	else
	if(!strcmp(argv[i],"-a"))
		{
		i++;
		if(i<argc)
			{	
			libname=(char *)calloc(1,4+strlen(argv[i])+1+1);
			strcpy(libname,"lib/");
			mkdir(libname,0755);
			strcat(libname,argv[i]);
			if(mkdir(libname,0755)==-1)
				if(errno!=EEXIST)
					{
					mesg("** Could not make '%s' directory for library!\n",libname);
					exit(1);
					}
			strcat(libname,"/");
			}
			else
			{
			mesg("** Missing library name after -a option!\n");
			exit(1);
			}	
		}
	else
	if(!strcmp(argv[i],"-y"))
		{
		i++;
		if(i<argc)
			{	
			char *lname=(char *)malloc(strlen(argv[i])+1);
			strcpy(lname, argv[i]);

			if(!libchain)
                		{     
                		struct vpp_filename *l;
                		l=(struct vpp_filename *)calloc(1,sizeof(struct vpp_filename));
                		l->name=lname;
                		libchain=l;
                		}
                		else
                		{
                		struct vpp_filename *l=libchain;
                		struct vpp_filename *l2;

				while(l->next) l=l->next;

                		l2=(struct vpp_filename *)calloc(1,sizeof(struct vpp_filename));
                		l2->name=lname;  
                		l->next=l2; 
                		}   
			}
			else
			{
			mesg("** Missing directory name after -y option!\n");
			exit(1);
			}	
		}
	else
	if(!strcmp(argv[i],"-s"))
		{
		supressstats=~0;
		}
	else
	if(!strcmp(argv[i],"-v"))
		{
		outputvhdl=~0;
		}
	else
	if(!strcmp(argv[i],"-license"))
		{
		print_license();
		exit(0);
		}
	else
	if((arglen>=8)&&(!strncmp(argv[i],"+incdir+",8)))
		{
		if(arglen==8)
			{
			warn("** Missing path for +incdir+ in command line argument %d, ignoring.\n", i);
			}
			else
			{
			char *lname=(char *)malloc(arglen-8+1);
			char *tok;
			strcpy(lname, argv[i]+8);

			tok=strtok(lname,"+");
			while(tok)
				{
				int toklen=strlen(tok);
                                if(!toklen)
                                        {
                                        /* strtok seems to never hit this */
                                        warn("** Missing path for +incdir+ (consecutive ++) in command line argument %d, ignoring.\n", i);
                                        }
                                        else
					{				
					if(!incchain)
			               		{     
			               		struct vpp_filename *l;
			               		l=(struct vpp_filename *)calloc(1,sizeof(struct vpp_filename));
			               		strcpy(l->name=malloc(toklen+1), tok);
			               		incchain=l;
			               		}
			               		else
			               		{
			               		struct vpp_filename *l=incchain;
			               		struct vpp_filename *l2;
			
						while(l->next) l=l->next;
	
			               		l2=(struct vpp_filename *)calloc(1,sizeof(struct vpp_filename));
			               		strcpy(l2->name=malloc(toklen+1), tok);
			               		l->next=l2; 
			               		}   
					}
				tok=strtok(NULL, "+");
				}
			free(lname);			
			}
		}
	else
	if((arglen>=8)&&(!strncmp(argv[i],"+define+",8)))
		{
		if(arglen==8)
			{
			warn("** Missing value for +define+ in command line argument %d, ignoring.\n", i);
			}
			else
			{
			char *lname=(char *)malloc(arglen-8+1);
			char *tok;
			strcpy(lname, argv[i]+8);

			tok=strtok(lname,"+");
			while(tok)
				{
				int toklen=strlen(tok);
                                if(!toklen)
                                        {
                                        /* strtok seems to never hit this */
                                        warn("** Missing value for +define+ (consecutive ++) in command line argument %d, ignoring.\n", i);
                                        }
                                        else
					{				
		                        char *dname=(char *)malloc(toklen+8+1);
        		                char *pnt;
        		                sprintf(dname, "`define %s", tok);
        		                pnt = dname+8;
        		                while(*pnt) 
        		                        {
        		                        if(*pnt=='=')
                		                        {
                		                        *pnt=' ';
                		                        break;
                	                        	}
               		                 	pnt++;
               	                 		}
		                        store_define(dname);
		                        mprintf("// vpp_file <ArgumentList>\n// vpp_line %d\n", i);
		                        mprintf(dname);
		                        mprintf("\n");
		                        free(dname);
					}
				tok=strtok(NULL, "+");
				}
			free(lname);			
			}
		}
	else
	if((arglen>=8)&&(!strncmp(argv[i],"+libext+",8)))
		{
		if(arglen==8)
			{
			warn("** Missing extension for +libext+ in command line argument %d, ignoring.\n", i);
			}
			else
			{
			char *lname=(char *)malloc(arglen-8+1);
			char *tok;
			strcpy(lname, argv[i]+8);

			tok=strtok(lname,"+");
			while(tok)
				{
                                int toklen=strlen(tok);
                                if(!toklen)
                                        {
                                        /* strtok seems to never hit this */
                                        warn("** Missing path for +libext+ (consecutive ++) in command line argument %d, ignoring.\n", i);
                                        }
                                        else
                                        {
					if(!libextchain)
			               		{     
			               		struct vpp_filename *l;
			               		l=(struct vpp_filename *)calloc(1,sizeof(struct vpp_filename));
			               		strcpy(l->name=malloc(toklen+1), tok);
			               		libextchain=l;
			               		}
			               		else
			               		{
			               		struct vpp_filename *l=libextchain;
			               		struct vpp_filename *l2;
			
						while(l->next) l=l->next;
			
			               		l2=(struct vpp_filename *)calloc(1,sizeof(struct vpp_filename));
			               		strcpy(l2->name=malloc(toklen+1), tok);  
			               		l->next=l2; 
			               		}   
					}
				tok=strtok(NULL, "+");
				}
			free(lname);
			}
		}
	else
	if((arglen>=2)&&(!strncmp(argv[i],"-D",2)))
		{
		if(arglen==2)
			{
			warn("** Missing define for -D in command line argument %d, ignoring.\n", i);
			}
			else
			{
			char *dname=(char *)malloc(arglen-2+8+1);
			char *pnt;
			sprintf(dname, "`define %s", argv[i]+2);
			pnt = dname+8;
			while(*pnt)
				{
				if(*pnt=='=') 
					{
					*pnt=' ';
					break;
					}
				pnt++;
				}
			store_define(dname);
                        mprintf("// vpp_file <ArgumentList>\n// vpp_line %d\n", i);
                        mprintf(dname);
                        mprintf("\n");
			free(dname);
			}
		}
	else
	if(argv[i][0]=='-')
		{
		if(strlen(argv[i])>1)
			{
			mesg("** Unknown option \"%s\".\n",argv[i]);
			exit(1);
			}
			else
			{
			if(!filename)
				{
				stdincompile=~0;
				}
				else
				{
				mesg("** Both \"%s\" and stdin selected as verilog sources.\n",filename->name);
				exit(1);
				}
			}
		}
	else
	if(stdincompile)
		{
		mesg("** Both \"%s\" and stdin selected as verilog sources.\n",argv[i]);
		exit(1);
		}
	else
		{
                struct filename *fn = calloc(1, sizeof(struct filename));
                fn->name = argv[i];
                if (lastfilename)
                   {
                   lastfilename->next = fn;
                   lastfilename = fn;
                   }
                else
                   filename = lastfilename = fn;
		}
	}


if(!outputvhdl)
warn("\n"VERNAME"\n");

vpp_update_yyfilename("<stdin>");

if(!filename && !stdincompile)
	{
	mesg("No files to process (use -h for help), exiting.\n\n");
	exit(0);	
	}

#define PP_BUF_SZ 10000
if (preproc_only)
   {
   char buf[PP_BUF_SZ];
   int num_read;
   while ((num_read = preproc(buf, PP_BUF_SZ)) > 0)
      fwrite(buf, sizeof(char), num_read, stdout);
   exit(0);
   }

lineroot=linepos=(char *)malloc(8192); /* max line length */
*lineroot=0x00; /* null terminate the text input string */

orbase=orstack=(struct bitvec **)calloc(1024,sizeof(struct bitvec *));
norbase=norstack=(struct bitvec **)calloc(1024,sizeof(struct bitvec *));
andbase=andstack=(struct bitvec **)calloc(1024,sizeof(struct bitvec *));
nandbase=nandstack=(struct bitvec **)calloc(1024,sizeof(struct bitvec *));
xorbase=xorstack=(struct bitvec **)calloc(1024,sizeof(struct bitvec *));
xnorbase=xnorstack=(struct bitvec **)calloc(1024,sizeof(struct bitvec *));

#ifdef YYDEBUG
yydebug = 1;
#endif

yyparse();

if((yyerrors)||(resynced))
	{
	warn("Source program has errors.\n");
	exit(255);
	}

DEBUG("Finished Parsing!\n\n");

/*
 * vhdl expansion is mutually exclusive with flattening, etc...
 */ 
if(outputvhdl)
	{
	vhdl(modroot);
	exit(0);
	}

currentmodcache=currentmod;

if(portmismatches)
	{
	mesg("** %d named port mismatches encountered.  Exiting.\n",portmismatches);
	exit(255);
	}


if(flattenname)
	{
	findmatchingmodule(modroot,flattenname);
	if(!(flattenfrom=matchmodule))
		{
		mesg("** Module \"%s\" not found!\n",flattenname);
		exit(255);
		}
	}
	else
	{
	if(!noivf)
		{
		findtopmodule(modroot);
		if(!(flattenfrom=topmodule))
			{
			mesg("** Top level module not found!  Try -f option.\n");
			exit(255);
			}
		}
		else
		{
		warn("IVF generation was suppressed.  Exiting.\n\n");
		exit(0);
		}
	}	

import(flattenfrom);
warn("\n");
reparse_param_modules(modroot);
if(!supressstats) moddump(modroot);
        
wandfix(modroot);
currentmod=currentmodcache;


warn("Flattening from module : %s\n",flattenfrom->name);
flatmodule=(struct flatmodule *)calloc(1,sizeof(struct flatmodule));
flatten2entry(flattenfrom,flattenfrom->name,"");

if(portmismatches)
	{
	mesg("** %d port mismatches encountered.  Exiting.\n",portmismatches);
	exit(255);
	}

warn("%d symbol%s, %d gate%s, %d wire%s, %d reg%s in flattened module : %s.\n",
     PLURAL(flatmodule->numsymbols), PLURAL(flatmodule->numgates),
     PLURAL(flatmodule->numwires), PLURAL(flatmodule->numregs),  
     flattenfrom->name);

if(noivf)
	{
	warn("\nProcessed \"%s\" successfully.\nIVF generation was suppressed.  Exiting.\n\n",flattenfrom->name);
	}
else
if(ivffile)
	{
	warn("\nWrote file \"%s.ivf\" successfully.  Exiting.\n\n",flattenfrom->name);
	}
	else
	{
	warn("Error with \"%s.ivf\" file!  Terminated early.\n",flattenfrom->name);
	}

exit(0);
}


#define MaxTokSz 1000

/* insert the content of the file named in (*argv)[i] into the argv array.
 * Also update argc to reflect the new argv size.
 */
static void insert_args(char ***argv, int *argc, int i)
{
   const char *filename = (*argv)[i];
   FILE *fp = fopen(filename,"r");
   struct filename *arglist = 0, *argend = 0;
   int c, inside_comment = 0, k = 0, argcnt = 0, remaining, new_argc;
   char s[MaxTokSz + 1], **new_argv;
   struct filename *argp;
   if (fp == 0)
   {
      mesg("** Cannot open argument file \"%s\" for reading\n", filename);
      exit(1);
   }
   while ((c = getc(fp)) != EOF)
   {
      /* first check the comment status */
      if      ((inside_comment == 0) && (c == '/'))  inside_comment = 1;
      else if ((inside_comment == 1) && (c == '*'))  inside_comment = 2;
      else if ((inside_comment == 1) && (c == '/'))  inside_comment = 3;
      else if ( inside_comment == 1)                 inside_comment = 0;
      else if ((inside_comment == 2) && (c == '*'))  inside_comment = 4;
      else if ((inside_comment == 3) && (c == '\n')) inside_comment = 5;
      else if ((inside_comment == 4) && (c == '/'))  inside_comment = 5;
      else if ((inside_comment == 4) && (c == '*'))  inside_comment = 4;
      else if ( inside_comment == 4)                 inside_comment = 2;
      if (((k > 0) && isspace(c)) || ((k > 1) && (inside_comment >= 2)))
      {  /* found end-of-token. Add it to the arglist: */
         argp = calloc(1, sizeof(struct filename));
         if (inside_comment >= 2)
            k -= 1;             /* must backup one position */
         argp->name = calloc(k+1, 1);
         strncpy((char*)argp->name, s, k);
         ((char*)(argp->name))[k] = 0;  /* null termination */
         if (argend == 0)
            arglist = argp;
         else
            argend->next = argp;
         argend = argp;
         argcnt += 1;
      k = 0;
      }
      else if ((inside_comment < 2) && !isspace(c))
         s[k++] = c;            /* copy char to s */
      else
         k = 0;
      if (inside_comment == 5) inside_comment = 0;
   }
   fclose(fp);
   if (inside_comment >= 2)
   {
      mesg("** Unterminated comment found while scanning \"%s\"\n", filename);
      exit(1);
   }
   if (k > 0)
   {
      mesg("** Newline missing at end of file \"%s\"\n", filename);
      exit(1);
   }
   new_argc = *argc + argcnt;
   new_argv = calloc(new_argc, sizeof(char *));
   for (k = 0; k <= i; k += 1)
      new_argv[k] = (*argv)[k];
   argp = arglist;
   for (k = i + 1; k <= i + argcnt; k += 1)
   {
      struct filename *remove = argp;
      new_argv[k] = (char*)argp->name;
      argp = argp->next;
      free(remove);
   }
   remaining = *argc - i - 1;
   for (k = i + argcnt + 1; k <= i + argcnt + remaining; k += 1)
      new_argv[k] = (*argv)[k - argcnt];
   *argv = new_argv;
   *argc = new_argc;
}
