/* 
 * Copyright (c) Tony Bybell, Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/*
 * symbol.h
 * 19/08/97ajb
 */

#ifndef SYMBOL_H
#define SYMBOL_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "vpp_common.h"

#define VERNAME "Ver: Structural Verilog Compiler v1.3.36 (w)1999-2002 BSI"

#define MAX2(a,b) (a > b ? a : b)
#define MAX3(a,b,c) (MAX2(MAX2(a,b),c))
#define ABS(a) (((a) < 0)?-(a):(a))

#define TRUE  1
#define FALSE 0
#define SYMPRIME 211

#ifdef DEBUG_PARSE
#define DEBUG   printf
#endif

#ifndef DEBUG_PARSE
#define DEBUG   blackhole_parse
void    DEBUG(char *, ...);
#endif

#ifdef DEBUG_SYM
#define DEBUG2   printf
#endif

#ifndef DEBUG_SYM
#define DEBUG2   blackhole_sym
void    DEBUG2(char *, ...);
#endif

#ifdef DEBUG_WAND
#define DEBUGW   printf
#endif

#ifndef DEBUG_WAND
#define DEBUGW   blackhole_sym
void    DEBUGW(char *, ...);
#endif

struct param_comp_chain
{
struct param_comp_chain *next;

char *basename;
char *fullname;
struct pnumlist *p;
int successfully_imported;
};


struct antext
{
struct antext *next;
char *text;
int ssize; /* length of text */
};

struct param
{
struct param *next;

struct param *chain;	/* points forward in parameter chain */
int which;
char *name;
int value;
int width;
int base;
char *str;
};

struct gatename 	/* so we can handle arrays of gates */
{
char *name;
int lo, hi;
};

struct gpchain
{
struct gpchain *next;
struct symslot *newsymbol; /* what the old gp->port / wire->port2 becomes in fixes */

struct symslot **fix;	   /* symslot to fix */
};

struct symslot
{
struct symslot *next;
char *name;
char id[8];		/* VCD style hashed id */
int bitnum;		/* encoded in name, but for convenience if needed */
			/* is -1 for module port headers... */
int value;		/* 0, 1, X (2), Z (3) */
int inaccess, outaccess;
int lorange, hirange;	/* only modified if bitnum==-1 */
int dir;		/* same as ^^^.  dir=1 [3:0] fmt, dir=-1 [0:3] */

int portdefined;	/* set when input or output stmt found */

int signaltype;		/* set to equal WIRE, WOR, etc... */
int delayr, delayf, delayz; /* wire "delay" */
int str0, str1;		/* "strengths" */
int flags;

struct gpchain *wchain;	/* gateport chain for wires.. */

struct gate *vhdlgate;	/* pointer to what to expand from */	
};

/* values for struct symslot.flags */
#define SF_VECTORED      0x01   /* set this if it's a "vectored" symslot */
#define SF_REGISTER      0x02   /* symbol is declared `reg' */
#define SF_NON_BLOCKING  0x04   /* reg assignment is non-blocking */
#define SF_BLOCKING      0x08   /* reg assignment is blocking */
#define SF_UNCONNECTED   0x10   /* symbol is unconnected */

struct bitvec
{			/* used for concatenation statements */
char *name;
int length;		
struct symslot *bits[1];
};


struct metavec
{
int length;
struct bitvec *vec[1];
};


struct reg
{
struct reg *next;
char not_a_latch;       /* TRUE if assigned in a combinational ALWAYS block */
char already_assigned;  /* TRUE if an ALWAYS block already assigned to reg */
struct symslot *port; 	/* rise + fall times are zero for registers */
};


struct wire
{
struct wire *next;
int delayr, delayf, delayz;	/* rise and fall times */
struct symslot *port1;
struct symslot *port2;
};


struct gateport
{
struct gateport *next;
struct symslot *port;
};

struct gate
{
struct gate *next;
int delayr, delayf, delayz;	/* rise and fall and hi impedance times */
int str0, str1;			/* strengths */
int type;

struct gateport *port;
struct gateport *outport; /* since it's at end of chain--makes quick to find */
int flags;
};

/* values for struct gate.flags */
#define GF_IN_USE     0x01

struct pnumlist
{
struct pnumlist *next;
int value;
char valid;
};

struct component
{
struct component *next;
struct module *type;
char *name;
struct metavec *port;
struct pnumlist *pnums;

int linenumber;		/* for debugging mismatched ports... */

int arraystart;		/* only for array of instances.. */
int arrayfin;
int isarray;		/* set when an array of instances */
};

struct wirechain
{			/* for processing wires/wands on endmodule reduce op */
struct wirechain *next;
struct symslot *sym;
};

struct module
{
struct module *left;
struct module *right;
char *name;

struct antext *re_analyze_text;	/* use this to re-analyze a module for parameter substitution */

struct gateport *moduleports;	/* in header */
struct component *comp;		/* components in this module */
struct wire *wires;		/* assign'd wires in this module */
struct reg *reg;		/* register bits in this module */

struct symslot *symbol[SYMPRIME];
struct param   *params[SYMPRIME];

struct gate *gates;

struct wirechain *wiresigs;	/* chain of all the wire/wand sigs in module */ 

int missing;		/* set when getmodfromlib fails */
int defined;		/* incremented when module is defined */
int referencecount;	/* incremented when module is instantiated */
int comps;		/* incremented when components instantiated */
int numsignals;		/* incremented when new signal name encountered */
int numgates;		/* incremented when gate instantiated */
int numports;		/* incremented upon port encountered in header */
int numinputports;
int numoutputports;
int numwires;		/* incremented on assigns which are simple = */
int numregs;		/* incremented on register declarations */
int numparams;		/* says how many params exist in params table */

int visited;		/* for flatten() to test module recursion */
int isexpanded;		/* for info in moddump */
};

struct filename
{
   const char *name;
   struct filename *next;
};

void yyerror(char *);
int yyparse(void);
void yyrestart(FILE *);

void vhdl(struct module *);
void idpnt(struct symslot *b);

struct param *paramfind(char *);
struct param *paramadd(char *);

int hash(char *);
struct module *modadd(struct module **,char *);
struct module *modfind(struct module *, char *, struct module **, int *);
struct module *simple_modfind(struct module *,char *);
void moddump(struct module *);

struct symslot *symfind(char *);
struct symslot *addsym(char *, int);
struct symslot *addsym2(char *, int);
void addgate(void);
void addport(struct symslot *);
struct symslot *createtemp(void);
void ioportwatch(char *);
struct symslot *new_unconnected(void);

void saveports(); /* for nested assigns inside */
void retrports(); /* of gate primitives..      */

struct wire *addwire(struct symslot *, struct symslot *);
struct reg *addreg(struct symslot *);
struct component *addcomponent(char *, struct module *, struct metavec *);


void wandfix(struct module *);	/* this fnt group was formerly in ver.y.. */
void wandfix2(struct module *);
int sortport(struct component *);
void supplytest(struct symslot *);
void vecheck(char *);   
struct symslot *wdelay(struct symslot *); 
void wirevecadd(char *);   
void expandgate(struct gatename *, struct metavec *); 
struct bitvec *unaryreduce(struct bitvec *, int, char *);   
void freevex(struct bitvec **);                                          
int portmatch(struct bitvec *, struct bitvec *); 
void swaprange(void);				/* ...and was moved to symbol.c */

void addtoken(void);
void flushtolib(void);
void getmodfromlib(struct module *);
void import(struct module *);

void mesg(const char *, ...);
void warn(const char *, ...);

extern char *yytext;
extern int yylineno, yyerrors, yyleng, resynced;
extern char *yyfilename;

void vpp_update_yylineno(const char *str);
void vpp_update_yyfilename(const char *str);

extern int portmismatches;
extern int commentdepth;
extern char *lineroot, *linepos;
extern int isnl, prevnl;                /* for consecutive newline stripping */

extern struct antext *anroot, *anpnt;	/* analyzed linked list of text */

extern int didimport;
extern struct vpp_filename *libchain;
extern struct vpp_filename *libextchain;
extern struct vpp_filename *incchain;

extern char *libname;			/* specified by -a option of ver */

extern int uselatchkw;			/* must explicitly enable explicit latchnames */
extern int skipnorefs;			/* must explicitly enable this! */
extern int psub;

extern struct module *modroot;
extern struct module *currentmod;
extern int isgate;
extern int gtype;
extern int signaltype;
extern struct gateport *gptr;
extern int gdelayr, gdelayf, gdelayz;
extern int wdelayr, wdelayf, wdelayz;
extern int str0, str1;     /* strength (left,right) */
extern int assigninprogress;

extern int wlo, whi, wdir; /* wire lo, hi, dir */
extern int wvectored;

extern int ismodulehdr;
extern int ismodulehdr, isinputstmt, isoutputstmt, isregstmt, iswirestmt;
extern int iscomponent;
extern int lorange, hirange, rangedir;
extern int componentline;

extern struct symslot *astemp;
extern int tempnum;

extern char *lineroot, *linepos; /* error reporting in lex */

extern struct bitvec **orbase, **orstack;
extern struct bitvec **norbase, **norstack;
extern struct bitvec **andbase, **andstack;
extern struct bitvec **nandbase, **nandstack;
extern struct bitvec **xorbase, **xorstack;
extern struct bitvec **xnorbase, **xnorstack;

extern struct param_comp_chain *param_comp_chain;
extern struct param_comp_chain *parameter_module_in_effect;
void reparse_param_modules(struct module *);
int param_subst(struct param *p);

int vpp_main(FILE *f, const char *name);
int mgetbuf(char *buf, int maxsiz);

#endif

