/* 
 * Copyright (c) Tony Bybell, Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

%{

/*
 * vtokens.l
 * 19/08/97ajb
 */

#include <stdarg.h>             /* variable number of args used in mesg() */
#include "alloca.h"
#include "symbol.h"
#include "behave.h"
#include "y.tab.h"

int preproc(char *buf, int max_size);	/* from preproc.c */

int uselatchkw=0;
int commentdepth=0;
char *yyfilename = NULL;
int yylineno=1;


void vpp_update_yyfilename(const char *str)
{
if(!yyfilename)
	{
	yyfilename=malloc(strlen(str)+1);
	strcpy(yyfilename, str);
	}
	else
	{
	free(yyfilename);
	yyfilename=malloc(strlen(str)+1);
	strcpy(yyfilename, str);
	}
}

void vpp_update_yylineno(const char *str)
{
yylineno = atoi(str) - 1;	/* because \n follows this vpp directive! */
}


int psub=FALSE;	/* disable parameter substitution for LHS of parameter defs */
extern int param_width;

char *numstring=NULL;
char *fixnumber(char *s)	/* gets rid of _ and whitespace in/preceeding numbers.. */
{
char *d, *o;
char ch;

o=d=s;

while((ch=*(s++)))
	{
	if((ch=='_')||(ch==' ')||(ch=='\t')) continue;
	*(d++)=ch;
	}
*d=0x00;
return(o);
}

#define YY_INPUT(buf, result, max_size) \
   result = preproc(buf, max_size);

%}

%x CMNT

%%

"/*"		{ commentdepth++; addtoken(); BEGIN CMNT; }
<CMNT>"/*"	{ addtoken(); /* commentdepth++; <=- VERILOG DOES NOT ALLOW NESTED COMMENTS */ }
<CMNT>"*/" 	{ addtoken(); commentdepth--; if (!commentdepth) { BEGIN INITIAL; } }
<CMNT>.		{ addtoken(); }
<CMNT>\n	{ isnl=~0; addtoken(); linepos=lineroot; *lineroot=0x00; yylineno++; }

module		{ addtoken(); return MODULE; }
endmodule 	{ addtoken(); flushtolib(); return ENDMODULE; }

parameter	{ addtoken(); return PARAMETER; }

input		{ addtoken(); return INPUT; }
output		{ addtoken(); return OUTPUT; }
reg		{ addtoken(); return REG; }
assign		{ addtoken(); return ASSIGN; }

(latch|latch_n|latch_sr|latch_sr_n) {
                  char *idcache;
		  struct param *p;
		  addtoken();

		  if(uselatchkw)	/* must explicitly turn on */
			{		/* latch as a keyword now! */
                        int sz = strlen(yytext);
                        switch (sz)
                        {
			case 5:  return LATCH;
                        case 7:  return LATCH_N;
			case 8:  return LATCH_SR;
                        case 10: return LATCH_SR_N;
                        default: abort();
                        }
			}

		  if(psub)	/* parameter substitution stuff */
		  if((p=paramfind(yytext)))
			{
			if(parameter_module_in_effect)
				{
				yylval.i=param_subst(p);
				}
				else
				{
				yylval.i=p->value;
				}
			return NUMBER;
			}

		  idcache=calloc(1,strlen(yytext)+1);
		  strcpy(idcache,yytext);
		  yylval.pnt=idcache;

		  return IDENTIFIER; } 

nmos		{ addtoken(); return NMOS; }
rnmos		{ addtoken(); return NMOS; }
pmos		{ addtoken(); return PMOS; }
rpmos		{ addtoken(); return PMOS; }

and		{ addtoken(); return AND; }
or		{ addtoken(); return OR; }

not		{ addtoken(); return NOT; }
notif0		{ addtoken(); return NOTIF0; }
notif1		{ addtoken(); return NOTIF1; }

buf		{ addtoken(); return BUF; }
bufif0		{ addtoken(); return BUFIF0; }
bufif1		{ addtoken(); return BUFIF1; }

xor		{ addtoken(); return XOR; }
nand		{ addtoken(); return NAND; }
nor		{ addtoken(); return NOR; }
xnor		{ addtoken(); return XNOR; }
const0		{ addtoken(); return CONST0; }
const1		{ addtoken(); return CONST1; }
constx		{ addtoken(); return CONSTX; }
constz		{ addtoken(); return CONSTZ; }

pulldown	{ addtoken(); return CONST0; }
pullup		{ addtoken(); return CONST1; }

wire		{ addtoken(); yylval.i=WIRE; return WIRE; }
wor		{ addtoken(); yylval.i=WOR; return WOR; }
wand		{ addtoken(); yylval.i=WAND; return WAND; }
tri		{ addtoken(); yylval.i=TRI; return TRI; }
trior		{ addtoken(); yylval.i=TRIOR; return TRIOR; }
triand		{ addtoken(); yylval.i=TRIAND; return TRIAND; }
tri1		{ addtoken(); yylval.i=TRI1; return TRI1; }
tri0		{ addtoken(); yylval.i=TRI0; return TRI0; }

supply0		{ addtoken(); yylval.i=SUPPLY0; return SUPPLY0; }
supply1		{ addtoken(); yylval.i=SUPPLY1; return SUPPLY1; }

highz0		{ addtoken(); yylval.i=HIGHZ0; return HIGHZ0; }
highz1		{ addtoken(); yylval.i=HIGHZ1; return HIGHZ1; }
strong0		{ addtoken(); yylval.i=STRONG0; return STRONG0; }
strong1		{ addtoken(); yylval.i=STRONG1; return STRONG1; }
pull0		{ addtoken(); yylval.i=PULL0; return PULL0; }
pull1		{ addtoken(); yylval.i=PULL1; return PULL1; }
weak0		{ addtoken(); yylval.i=WEAK0; return WEAK0; }
weak1		{ addtoken(); yylval.i=WEAK1; return WEAK1; }

vectored	{ addtoken(); return VECTORED; }
scalared	{ addtoken(); return SCALARED; }

always          { addtoken(); return ALWAYS; }
begin           { addtoken(); return BEGIN_TOKEN; }
end             { addtoken(); return END; }
if              { addtoken(); return IF; }
else            { addtoken(); return ELSE; }
case            { addtoken(); return CASE; }
casex           { addtoken(); return CASEX; }
casez           { addtoken(); return CASEZ; }
endcase         { addtoken(); return ENDCASE; }
default         { addtoken(); return DEFAULT; }
posedge         { addtoken(); return POSEDGE; }
negedge         { addtoken(); return NEGEDGE; }

"@"             { addtoken(); return EVENT; }
"{"		{ addtoken(); return LB; }
"}"		{ addtoken(); return RB; }

"="		{ addtoken(); return EQ; }
"("		{ addtoken(); return LP; }
")"		{ addtoken(); return RP; }

"<<"    { addtoken(); return LSH; }
">>"    { addtoken(); return RSH; }
"+"     { addtoken(); return ADD; }
"-"     { addtoken(); return SUB; }
"*"     { addtoken(); return MPY; }
"/"     { addtoken(); return DIV; }
"%"     { addtoken(); return MOD; }
"^"     { addtoken(); return XOR; }
"<"     { addtoken(); return LT; }
">"     { addtoken(); return GT; }
"<="    { addtoken(); return LET; }
">="    { addtoken(); return GET; }
"=="    { addtoken(); return EQV; }
"!="    { addtoken(); return NEQ; }
"&&"    { addtoken(); return AAND; }
"||"    { addtoken(); return OOR; }	
"!"     { addtoken(); return NNOT; }

"~"		{ addtoken(); return BITNOT; }
"&"		{ addtoken(); return BITAND; }
"|"		{ addtoken(); return BITOR; }

"~&"		{ addtoken(); return BITNAND; }
"~|"		{ addtoken(); return BITNOR; }
"~^"		{ addtoken(); return BITXNOR; }

"?"             { addtoken(); return MUX; }

"["		{ addtoken(); return LBRACK; }
"]"		{ addtoken(); return RBRACK; }
","		{ addtoken(); return COMMA; }
":"		{ addtoken(); return COLON; }
";"		{ addtoken(); return SEMI; }
"#"		{ addtoken(); return POUND; }
[A-Za-z_]+[A-Za-z_0-9\$]* { 
		  char *idcache;
		  struct param *p;
		  addtoken();

		  if(psub)	/* parameter substitution stuff */
		  if((p=paramfind(yytext)))
			{
			if (p->str == 0)
			     {
			     if(parameter_module_in_effect)
				{
				yylval.i=param_subst(p);
				}
				else
				{
				yylval.i=p->value;
				}
			     return NUMBER;
			     }
			     else
			     {
			     yylval.pnt=p->str;
                             param_width = p->width;
                             switch (p->base)
                                  {
                                  case 2:  return BBASE;
                                  case 8:  return OBASE;
                                  case 10: return DBASE;
                                  default: return HBASE;
                                  }
			     }
			}

		  idcache=calloc(1,strlen(yytext)+1);
		  strcpy(idcache,yytext);
		  yylval.pnt=idcache;

		  return IDENTIFIER; } 

\\[^ \t\r\n]+	  { 
		  char *idcache, *idc2, *pnt;
		  char ch;
		  struct param *p;
		  int elen=0;

		  addtoken();

		  pnt=yytext+1;
		  while((ch=*(pnt++)))
			{
			if((ch=='!')||(ch=='/')||(ch=='[')||(ch==']')||(ch=='\\'))
				elen+=2;
				else elen++;
			}

		  idcache=idc2=calloc(1,elen+1); 
		  pnt=yytext+1;
		  while((ch=*(pnt++)))
			{
			if((ch=='!')||(ch=='/')||(ch=='[')||(ch==']')||(ch=='\\'))
				{
				*(idc2++)='\\';
				}
			*(idc2++)=ch;
			}

		  if(psub)	/* parameter substitution stuff */
		  if((p=paramfind(idcache)))
			{
			if(parameter_module_in_effect)
				{
				yylval.i=param_subst(p);
				}
				else
				{
				yylval.i=p->value;
				}
			free(idcache);	/* unneeded here */
			return NUMBER;
			}

		  yylval.pnt=idcache;
		  return IDENTIFIER; } 

[0-9]+		{ addtoken(); yylval.i=atoi(yytext); return NUMBER; }

\'[dD][ \t]*[0-9][0-9_]* { 
		addtoken(); 
		numstring=calloc(1,strlen(yytext)-1);   /* skip 'd leading descriptor */
		strcpy(numstring,yytext+2);
		yylval.pnt=fixnumber(numstring);
                param_width = -1;
		return DBASE;
		}

\'[bB][ \t]*[xXzZ?01][xXzZ?01_]* {
		addtoken(); 
		numstring=calloc(1,strlen(yytext)-1);	/* skip 'b leading descriptor */
		strcpy(numstring,yytext+2);
		yylval.pnt=fixnumber(numstring);
                param_width = -1;
		return BBASE;
		}

\'[oO][ \t]*[xXzZ0-7][xXzZ0-7_]* {
		addtoken(); 
		numstring=calloc(1,strlen(yytext)-1);	/* skip 'o leading descriptor */
		strcpy(numstring,yytext+2);
		yylval.pnt=fixnumber(numstring);
                param_width = -1;
		return OBASE;
		}

\'[hH][ \t]*[xXzZ0-9A-Fa-f][xXzZ0-9A-Fa-f_]* {
		addtoken(); 
		numstring=calloc(1,strlen(yytext)-1);	/* skip 'h leading descriptor */
		strcpy(numstring,yytext+2);
		yylval.pnt=fixnumber(numstring);
                param_width = -1;
		return HBASE;
		}

\.		{ addtoken(); return DOT; }

"*/"		{ addtoken(); return UNTERMCOMMENT; }

[ \t\r]+	{ addtoken(); }
\n		{ isnl=~0; addtoken(); linepos=lineroot; *lineroot=0x00; yylineno++; }

\`[^\n]*	{
                char *tok=ver_alloca(yyleng+1);
		strcpy(tok, yytext);

                tok = strtok(tok," \t");
                if (tok != 0)
			{
			warn("** Skipping directive '%s' in file '%s' line %d.\n", tok, yyfilename, yylineno);
			}
		}

\/\/[^\n]*      {
                char *tok=ver_alloca(yyleng+1);
		strcpy(tok, yytext);

                tok = strtok(tok,"/ \t");
                if (tok != 0)
		if (!strcmp("vpp_file", tok))
			{
			tok = strtok(NULL, "");
			if(tok) vpp_update_yyfilename(tok);
			}
		else
		if (!strcmp("vpp_line", tok))
			{
			tok = strtok(NULL, "");
			if(tok) vpp_update_yylineno(tok);
			}
		else
		   {
		   addtoken();	/* we don't want line changing primitives messing with libraries or expanded components */
                   if ((strcasecmp(tok,"pragma") == 0)
                       || (strcasecmp(tok,"synopsys") == 0)
                       || (strcasecmp(tok,"ambit") == 0))
                      if (!pragma_expected(FALSE))
                         warn("-- WARNING: unexpected pragma in line: %d\n",
                              yylineno);
                      else
                      {
                         if ((tok = strtok(0, "/ \t")) == 0)
                            warn("-- WARNING: pragma not followed by a keyword"
                                 " in line %1d\n", yylineno);
                         yylval.pnt = strdup(tok ? tok : "missing_pragma");
                         return PRAGMA;
                      }
		   }
                }
.		{ addtoken(); return yytext[0]; }
%%
