%p 6000
%e 2000
%{

/* Copyright (C) 1996 Himanshu M. Thaker

This file is part of vpp.

Vpp is distributed in the hope that it will be useful,
but without any warranty.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.

Everyone is granted permission to copy, modify and redistribute
vpp, but only under the conditions described in the
document "vpp copying permission notice".   An exact copy
of the document is supposed to have been given to you along with
vpp so that you can know how you may redistribute it all.
It should be in a file named COPYING.  Among other things, the
copyright notice and this notice must be preserved on all copies.  */

/*
 * $Header: /home/cvspsrv/cvsroot/eda/vbpp/vpp.l,v 1.2 2000/07/29 10:40:31 jching Exp $
 * $Log: vpp.l,v $
 * Revision 1.2  2000/07/29 10:40:31  jching
 * Changed get_define_value API to allow error detection.
 *
 * Revision 1.1.1.1  2000/02/21 04:05:37  ahvezda
 * Import of vbpp
 *
 * Revision 1.8  1995/04/23  17:18:12  hemi
 * All features except string handling.
 */

#include "vpp_common.h"
#include "vpp_yacc_stuff.h"
#include "vpp_proto.h"
#include "vpp.tab.h"
#include <math.h>

int vpp_queue_comment=0;
char *vpp_queue_comment_data=NULL;

int vppcommentdepth=0;
int vppcommentline=0;
static int vppbeginmode=INITIAL;
static int eatcachemode=INITIAL;
#define VPPBEGIN(x) do{vppbeginmode=(x);BEGIN((x));}while(0)

struct vpp_filename *incchain=NULL;
static FILE *incdir_fopen(char *nam)
{
FILE *fp=NULL;
struct vpp_filename *l;  
int len;

if(incchain)
        {
        len = strlen(nam);
        l=incchain;
        while(l)
                {
                char *path = (char *)malloc(strlen(l->name) + 1 + len + 1);
                sprintf(path, "%s/%s", l->name, nam);
                fp=fopen(path, "r");
                free(path);
                if(fp) break;
                l=l->next;
                }
        }

if(!fp) fp=fopen(nam, "r");

return(fp);
}


extern char ttid[]; /* From vpp.y */

typedef struct names
   {
        char *macro_name;
        char *macro;
	int is_number;
	double value;
        struct names *next;
   } names;

void store_it(char *str, int subst, int expand_define);
void store_define(char *def_text);
static int vhash(char *text);
int delete_entry(names *table[], char *macro_name, int hashval);
void add_macro(names *table[], char *macro_name, int hashval, char *text);
void print_variable(char *name, int expand_define);
void store_lines();
void do_comment_count(int force_print, int count);
void do_post_command(char *txt, int post_type);
int is_keyword(char *name);
void undef_macro(char *def_text);

gen_ll_mgr *code_pointer = NULL;
char line[MAXNAME];
names *Defines_Table[BASE];
int nl_count = 1;
int expand_defines = FALSE;

names *lookup_define(names *table[], char *macro_name, int hashval);
#define MAX_INCLUDE_DEPTH 64
YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
int include_nl_count[MAX_INCLUDE_DEPTH];
char *include_fname[MAX_INCLUDE_DEPTH];
int last_mode = INITIAL;
int include_stack_ptr = 0;

#define YY_USER_ACTION store_lines();
char old_line[MAXNAME];
char current_line[MAXNAME];
#define COMMENT_EVERY 100

char *current_file=NULL;
int dont_add_comments = FALSE;

#define PP 0
#define MM 1
#define PE 2
#define ME 3

%}

%x MPP
%x COPY
%x INCL
%x SKIP_TO_EOL
%x EAT
%x EATT

WHITE           [ \t]
NEWLINE		[\n]
DIGIT10         [0-9_]
NMBR10          {DIGIT10}+
NAME            [a-zA-Z_][a-zA-Z0-9_\$]*
MEM             \[[^;\n\]]*\]

%%

<INITIAL,MPP>"/\*"  {/*eat_up_comment(FALSE);*/ 
			eatcachemode=vppbeginmode; 
			mprintf("%s",wwtext); 
			vppcommentdepth=1; 
			vppcommentline=nl_count; 
			VPPBEGIN(EAT); 
			}
<COPY>"/\*"         {/*eat_up_comment(TRUE);*/ 
			eatcachemode=vppbeginmode; 
			store_it(wwtext,FALSE,FALSE); 
			vppcommentdepth=1; 
			vppcommentline=nl_count; 
			VPPBEGIN(EATT); 
			}
                    
<EAT>"*/"          { mprintf("%s",wwtext); if(!(--vppcommentdepth)){VPPBEGIN(eatcachemode);}}
<EAT>"/*"          { mprintf("%s",wwtext); /* vppcommentdepth++; <=- VERILOG DOES NOT ALLOW NESTED COMMENTS */ }
<EAT>\n            { mprintf("%s",wwtext); nl_count++; }
<EAT>[^*\n]+       { mprintf("%s",wwtext); }
<EAT>.             { mprintf("%s",wwtext); }

<EATT>"*/"         { store_it(wwtext,FALSE,FALSE); if(!(--vppcommentdepth)){VPPBEGIN(eatcachemode);}}
<EATT>"/*"         { store_it(wwtext,FALSE,FALSE); /* vppcommentdepth++; <=- VERILOG DOES NOT ALLOW NESTED COMMENTS */ }
<EATT>\n           { store_it(wwtext,FALSE,FALSE); nl_count++; }
<EATT>[^*\n]+      { store_it(wwtext,FALSE,FALSE); }
<EATT>.            { store_it(wwtext,FALSE,FALSE); }

		    
<INITIAL,MPP>"//".*$	    {  if(!vpp_queue_comment)
                                        {
                                        mprintf("%s",wwtext);
                                        }  
                                        else
                                        {
                                        if(vpp_queue_comment_data) free(vpp_queue_comment_data);
                                        vpp_queue_comment_data = strdup(wwtext);
                                        }
                             }

<COPY>"//".*$                {store_it(wwtext,FALSE,FALSE);}

<COPY>`accelerate		{store_it(wwtext,FALSE,FALSE);}
<COPY>`noaccelerate		{store_it(wwtext,FALSE,FALSE);}
<COPY>`autoexpand_vectornets	{store_it(wwtext,FALSE,FALSE);}
<COPY>`celldefine		{store_it(wwtext,FALSE,FALSE);}
<COPY>`endcelldefine		{store_it(wwtext,FALSE,FALSE);}
<COPY>`default_decay_time	{store_it(wwtext,FALSE,FALSE);}
<COPY>`default_nettype		{store_it(wwtext,FALSE,FALSE);}
<COPY>`default_rswitch_strength	{store_it(wwtext,FALSE,FALSE);}
<COPY>`default_switch_strength	{store_it(wwtext,FALSE,FALSE);}
<COPY>`default_trireg_strength	{store_it(wwtext,FALSE,FALSE);}
<COPY>`delay_mode_distributed	{store_it(wwtext,FALSE,FALSE);}
<COPY>`delay_mode_path		{store_it(wwtext,FALSE,FALSE);}
<COPY>`delay_mode_unit		{store_it(wwtext,FALSE,FALSE);}
<COPY>`delay_mode_zero		{store_it(wwtext,FALSE,FALSE);}
<COPY>`expand_vectornets	{store_it(wwtext,FALSE,FALSE);}
<COPY>`noexpand_vectornets	{store_it(wwtext,FALSE,FALSE);}
<COPY>`protect			{store_it(wwtext,FALSE,FALSE);}
<COPY>`endprotect		{store_it(wwtext,FALSE,FALSE);}
<COPY>`pre_16a_paths		{store_it(wwtext,FALSE,FALSE);}
<COPY>`end_pre_16a_paths	{store_it(wwtext,FALSE,FALSE);}
<COPY>`protected		{store_it(wwtext,FALSE,FALSE);}
<COPY>`unprotected		{store_it(wwtext,FALSE,FALSE);}
<COPY>`remove_gatenames		{store_it(wwtext,FALSE,FALSE);}
<COPY>`noremove_gatenames	{store_it(wwtext,FALSE,FALSE);}
<COPY>`resetall			{store_it(wwtext,FALSE,FALSE);}
<COPY>`rs_technology		{store_it(wwtext,FALSE,FALSE);}
<COPY>"`switch default"		{store_it(wwtext,FALSE,FALSE);}
<COPY>"`switch resisitive"	{store_it(wwtext,FALSE,FALSE);}
<COPY>"`switch XL"		{store_it(wwtext,FALSE,FALSE);}
<COPY>`timescale		{store_it(wwtext,FALSE,FALSE);}
<COPY>`unconnected_drive	{store_it(wwtext,FALSE,FALSE);}
<COPY>`nounconnected_drive	{store_it(wwtext,FALSE,FALSE);}
<COPY>`uselib			{store_it(wwtext,FALSE,FALSE);}

<INITIAL>`accelerate		{mprintf("%s",wwtext);}
<INITIAL>`noaccelerate		{mprintf("%s",wwtext);}
<INITIAL>`autoexpand_vectornets	{mprintf("%s",wwtext);}
<INITIAL>`celldefine		{mprintf("%s",wwtext);}
<INITIAL>`endcelldefine		{mprintf("%s",wwtext);}
<INITIAL>`default_decay_time	{mprintf("%s",wwtext);}
<INITIAL>`default_nettype	{mprintf("%s",wwtext);}
<INITIAL>`default_rswitch_strength	{mprintf("%s",wwtext);}
<INITIAL>`default_switch_strength	{mprintf("%s",wwtext);}
<INITIAL>`default_trireg_strength	{mprintf("%s",wwtext);}
<INITIAL>`delay_mode_distributed	{mprintf("%s",wwtext);}
<INITIAL>`delay_mode_path	{mprintf("%s",wwtext);}
<INITIAL>`delay_mode_unit	{mprintf("%s",wwtext);}
<INITIAL>`delay_mode_zero	{mprintf("%s",wwtext);}
<INITIAL>`expand_vectornets	{mprintf("%s",wwtext);}
<INITIAL>`noexpand_vectornets	{mprintf("%s",wwtext);}
<INITIAL>`protect		{mprintf("%s",wwtext);}
<INITIAL>`endprotect		{mprintf("%s",wwtext);}
<INITIAL>`pre_16a_paths		{mprintf("%s",wwtext);}
<INITIAL>`end_pre_16a_paths	{mprintf("%s",wwtext);}
<INITIAL>`protected		{mprintf("%s",wwtext);}
<INITIAL>`unprotected		{mprintf("%s",wwtext);}
<INITIAL>`remove_gatenames	{mprintf("%s",wwtext);}
<INITIAL>`noremove_gatenames	{mprintf("%s",wwtext);}
<INITIAL>`resetall		{mprintf("%s",wwtext);}
<INITIAL>`rs_technology		{mprintf("%s",wwtext);}
<INITIAL>"`switch default"	{mprintf("%s",wwtext);}
<INITIAL>"`switch resisitive"	{mprintf("%s",wwtext);}
<INITIAL>"`switch XL"		{mprintf("%s",wwtext);}
<INITIAL>`timescale		{mprintf("%s",wwtext);}
<INITIAL>`unconnected_drive	{mprintf("%s",wwtext);}
<INITIAL>`nounconnected_drive	{mprintf("%s",wwtext);}
<INITIAL>`uselib		{mprintf("%s",wwtext);}

<*>^{WHITE}*"`ifdef"	{return TTIFDEF;}
<*>^{WHITE}*"`ifndef"	{return TTIFNDEF;}
<*>^{WHITE}*"`if"	{return TTIF;}
<*>^{WHITE}*"`let"	{return TTLET;}
<*>^{WHITE}*"`for"	{return TTFOR;}
<*>^{WHITE}*"`while"	{return TTWHILE;}
<*>^{WHITE}*"`switch"	{return TTSWITCH;}
<*>^{WHITE}*"`case"	{return TTCASE;}
<*>^{WHITE}*"`default"	{return TTDEFAULT;}
<*>^{WHITE}*`breaksw[^\n]* {return TTBREAKSW;}
<*>^{WHITE}*`else[^\n]* {return TTELSE;}
<*>^{WHITE}*`endif[^\n]* {return TTENDIF;}
<*>^{WHITE}*`endfor[^\n]* {return TTENDFOR;}
<*>^{WHITE}*`endwhile[^\n]* {return TTENDWHILE;}
<*>^{WHITE}*`endswitch[^\n]* {return TTENDSWITCH;}

<INITIAL>`define[^\n]* {mprintf("%s",wwtext);store_define(wwtext);}
<COPY>`define[^\n]* {store_it(wwtext,FALSE, FALSE);store_define(wwtext);}

<INITIAL>`undef[^\n]* {undef_macro(wwtext);mprintf("%s",wwtext);}
<COPY>`undef[^\n]* {undef_macro(wwtext);store_it(wwtext,FALSE, FALSE);}

<INITIAL,COPY>`include	   {VPPBEGIN(INCL);}

<INITIAL,COPY,MPP>{NAME}{WHITE}*{MEM}?{WHITE}*"++" {do_post_command(wwtext,PP);}
<INITIAL,COPY,MPP>{NAME}{WHITE}*{MEM}?{WHITE}*"--" {do_post_command(wwtext,MM);}
<INITIAL,COPY,MPP>{NAME}{WHITE}*{MEM}?{WHITE}*"+=" {do_post_command(wwtext,PE);}
<INITIAL,COPY,MPP>{NAME}{WHITE}*{MEM}?{WHITE}*"-=" {do_post_command(wwtext,ME);}

<INCL>[ \t"]*	   {}
<INCL>[^ \"\t\n]+  {
		    FILE *fp;

		    if (include_stack_ptr >= MAX_INCLUDE_DEPTH) {
		      fprintf(stderr, "ERROR : lex : includes nested too deep\n");
		      exit (1);
		    }
		    fp = incdir_fopen(wwtext);
		    if (!fp) {
		      fprintf(stderr, "ERROR : lex : could not open file '%s'\n",wwtext);
		      VPPBEGIN(SKIP_TO_EOL);
		    }
		    else {
		      include_nl_count[include_stack_ptr] = nl_count;
		      include_fname[include_stack_ptr] = NMALLOC(strlen(current_file)+1, char);
                      strcpy(include_fname[include_stack_ptr], current_file);
                      free(current_file);
                      current_file = NMALLOC(strlen(wwtext)+1,char);
                      strcpy(current_file, wwtext);

		      include_stack[include_stack_ptr++] = YY_CURRENT_BUFFER;

		      nl_count = 1;
		      wwin = fp;
		      ww_switch_to_buffer(ww_create_buffer(wwin, YY_BUF_SIZE));
                      do_comment_count(TRUE,nl_count);
		      VPPBEGIN(INITIAL);
		    }
		}
<SKIP_TO_EOL>.*$	{VPPBEGIN(INITIAL);}

<<EOF>>		{
		  if (!include_stack_ptr) yyterminate();
		  else {
		    include_stack_ptr--;
		    ww_delete_buffer(YY_CURRENT_BUFFER);
		    ww_switch_to_buffer(include_stack[include_stack_ptr]);
		    nl_count = include_nl_count[include_stack_ptr];
                    free(current_file);
                    current_file = include_fname[include_stack_ptr];
                    do_comment_count(TRUE, nl_count);
		    VPPBEGIN(SKIP_TO_EOL);
		  }
		}

<MPP>"LOG2"	{return TTLOG2;}
<MPP>"ROUND"	{return TTROUND;}
<MPP>"CEIL"	{return TTCEIL;}
<MPP>"FLOOR"	{return TTFLOOR;}
<MPP>"MAX"	{return TTMAX;}
<MPP>"MIN"	{return TTMIN;}
<MPP>"ABS"	{return TTABS;}
<MPP>"STOI"	{return TTSTOI;}
<MPP>"ITOS"	{return TTITOS;}
<MPP>"SUBSTR"	{return TTSUBSTR;}
<MPP>"SYSTEM"	{return TTSYSTEM;}
<MPP>"EVEN"	{return TTEVEN;}
<MPP>"ODD"	{return TTODD;}

<INITIAL>`{NAME}  {print_variable(wwtext+1, expand_defines);}
<COPY>`{NAME}  {store_it(wwtext,TRUE,FALSE);}

<INITIAL>{WHITE}+ {strcpy(ttid, wwtext); return TTWHITE;}
<INITIAL,MPP>{NAME}	{strcpy(ttid, wwtext); return TTNAME;}

<INITIAL>\\::  {mprintf("::");}
<COPY>\\::  {store_it("::",FALSE,FALSE);}

<*>::    {}

<INITIAL,MPP>{NMBR10}	{strcpy(ttid, wwtext); return TTNUM;}
<INITIAL,MPP>{DIGIT10}*\.{DIGIT10}+ {strcpy(ttid, wwtext);return TTREALNUM;}
<INITIAL,MPP>{DIGIT10}+\.{DIGIT10}* {strcpy(ttid, wwtext);return TTREALNUM;}

<MPP>\"[^\"\n]*[\"]?  {strcpy(ttid,wwtext); return TTQS;}
<MPP>{WHITE}+ 	{}
<MPP>{NEWLINE} 	{nl_count++; return '\n';}
<MPP>"`"	{}
<MPP>":"	{return ':';}
<MPP>"("	{return '(';}
<MPP>")"	{return ')';}
<MPP>"+"	{return '+';}
<MPP>"-"	{return '-';}
<MPP>"*"	{return '*';}
<MPP>"/"	{return '/';}
<MPP>"%"	{return '%';}
<MPP>"**"	{return TT_POWER;}
<MPP>"=="	{return TTL_EQ;}
<MPP>"!="	{return TTL_NEQ;}
<MPP>"&&"	{return TTL_AND;}
<MPP>"||"	{return TTL_OR;}
<MPP>"<"	{return '<';}
<MPP>"<="	{return TTL_LEQ;}
<MPP>">"	{return '>';}
<MPP>">="	{return TTL_GEQ;}
<MPP>"&"	{return '&';}
<MPP>"|"	{return '|';}
<MPP>"^"	{return '^';}
<MPP>"<<"	{return TTL_SHIFTL;}
<MPP>">>"	{return TTL_SHIFTR;}
<MPP>"!"	{return '!';}
<MPP>"="	{return '=';}
<MPP>";"	{return ';';}
<MPP>"~"	{return '~';}
<MPP>","	{return ',';}
<MPP>.		{return TT_UNKNOWN;}

<COPY>{NEWLINE}	{nl_count++; store_it(wwtext,FALSE,FALSE);}
<COPY>.		{store_it(wwtext,FALSE,FALSE);}
<INITIAL>{NEWLINE} {nl_count++; mprintf("\n"); do_comment_count(FALSE,nl_count);}
<INITIAL>.	{mprintf("%c",wwtext[0]);}

%%

void
store_it(char *str, int subst, int expand_define)
{
    char nn[MAXSTR];
    int len;
    names *np;
    var_values *vvp;

    if ((strcmp(str,"\n") != 0) && (!subst)) {
	sprintf(line, "%s%s", line,str);
    }
    else {
        if ((is_keyword(str+1) && subst)) mprintf("%s",str);
	else if (!subst) sprintf(line, "%s\n",line);
	if (code_pointer == NULL) {
	    code_pointer = MALLOC(gen_ll_mgr);
	    code_pointer->head = NULL;
	    code_pointer->tail = NULL;
	}
	if (code_pointer->head == NULL) {
	    code_pointer->head = MALLOC(gen_ll);
	    code_pointer->head->next = NULL;
	    code_pointer->head->prev = NULL;
	    code_pointer->tail = code_pointer->head;
	}
	else {
	    code_pointer->tail->next = MALLOC(gen_ll);
	    code_pointer->tail->next->prev = code_pointer->tail;
	    code_pointer->tail = code_pointer->tail->next;
	}
	
	code_pointer->tail->ll_type = CODE_POINTER;
	code_pointer->tail->item = (ll_generic *)NMALLOC(strlen(line)+1,char);
	strcpy(((char *)code_pointer->tail->item), line);
	code_pointer->tail->line_count = nl_count;
	
	code_pointer->tail->next = NULL;
	sprintf(line,"%s", "");

	if (subst) {
	    code_pointer->tail->next = MALLOC(gen_ll);
	    code_pointer->tail->next->prev = code_pointer->tail;
	    code_pointer->tail = code_pointer->tail->next;

	    len = strlen(str);
	    /*
	     * First check to see if it a `define.  If it is,
	     * then we DON'T convert it.  Otherwise, we do convert it.
	     */
	    if ((np = lookup_define(Defines_Table, str+1,vhash(str+1)))) {
		if (expand_define) {
		    code_pointer->tail->ll_type = CODE_POINTER;
		    code_pointer->tail->item =
		      (ll_generic *)NMALLOC(strlen(np->macro)+1,char);
		    strcpy(((char *)code_pointer->tail->item),np->macro);
		}
		else {
		    code_pointer->tail->ll_type = CODE_POINTER;
		    code_pointer->tail->item =
		      (ll_generic *)NMALLOC(len+1,char);
		    strcpy(((char*)code_pointer->tail->item), str);
		}
	    }
	    else {
	        if ((vvp = get_variable_ref(STRING,str+1)) == NULL) {
		  sprintf(nn,"ERROR : variable `%s has not been assigned prior to use",str+1);
		  wwerror(nn);
		}
		code_pointer->tail->ll_type = CODE_POINTER2;
		code_pointer->tail->item =
		  (ll_generic *)NMALLOC(len,char); /* note no +1 */
		strcpy(((char *)code_pointer->tail->item), str+1);
	    }
	    code_pointer->tail->line_count = nl_count;
	    code_pointer->tail->next = NULL;
	}
    }
}

/**************************************************************************
 *
 * Function Name :
 * Parameters :
 * Return Value :
 * Purpose :
 *    This function assumes that wwtext contains %name%, where name is
 * what is to be looked up and substituted with the appropriate integer
 * value, that has either been defined before with "`let", or inside
 * a looping construct.
 *
 **************************************************************************/
void
lookup()
{
    char name[MAXSTR];

    strcpy(name,wwtext+3);
    print_variable(name, TRUE);
}


/**************************************************************************
 *
 * Function Name : int store_define()
 * Parameters :
 *      char *def_text - text of the define statement
 * Purpose :
 *    To add a "`define" into the table.  Note that if a previous entry
 * already exists, it is first removed, and then the new entry superceeds
 * the old one.
 *
 **************************************************************************/
void
store_define(char *def_text)
{
   char *txt_ptr;
   char macro_name[MAXSTR];
   int hashval;

   txt_ptr = (char *)(((int)strstr(def_text, "define")) + ((int )strlen("define")));
   sscanf(txt_ptr, "%s", macro_name);
   txt_ptr = (char *)(((int)strstr(txt_ptr, macro_name)) + ((int)strlen(macro_name)));

   hashval = vhash(macro_name);

   /*
    * First, see if the name has previously been stored.
    */
   if (lookup_define(Defines_Table, macro_name, hashval) != NULL) {
/*
       fprintf(stderr, "WARNING : lex : redefining '%s', using '%s'\n",macro_name,txt_ptr);
*/
       delete_entry(Defines_Table, macro_name, hashval);
   }
   /*
    * Now, store the name and macro into the table.
    */
   add_macro(Defines_Table, macro_name, hashval, txt_ptr);
}


/**************************************************************************
 *
 * Function Name : int vhash()
 * Parameters :
 *      char *text - text you wish to get hash value of
 * Purpose :
 *      This routine sums up the ascii value of the text, mod a BASE value,
 * and returns this value as the hash value.
 *
 **************************************************************************/
int
vhash(char *text)
{
  int sum;
  int i;

  sum = 0;
  for(i=0; i<strlen(text); i++) sum = sum + text[i];
  return (sum % BASE);
}

/**************************************************************************
 *
 * Function Name : names *lookup_define()
 * Parameters :
 *      names *table[] - table where names are to be stored in
 *      char *macro_name - name to lookup
 *      int hashval - computed hash value;
 * Return Value :
 *      names * -  NULL if not found, address of a 'names' if found
 * Purpose :
 *    See if the name given is defined in the table.  To make for quick
 * searches, the entries are hashed into a linked list table.
 *
 **************************************************************************/
names *
lookup_define(names *table[], char *macro_name, int hashval)
{
  names *np;

  for (np = table[hashval]; np != NULL; np=np->next) {
      if (strcmp(macro_name,np->macro_name) == 0) return(np);
  }
  return(NULL);
}

/**************************************************************************
 *
 * Function Name : int delete_entry()
 * Parameters :
 *      names *table[] - table where names are to be stored in
 *      char *macro_name - name to delete
 *      int hashval - computed hash value;
 * Return Value :
 *      none
 * Purpose :
 *    To delete an entry in a hash'd database.  We may wish to delete
 * a macro for example if we come across a duplicate identical macro def,
 * then the newest one replaces the current one.  Note that this routine
 * simply deletes the entry from the table.  It is up to the user to add
 * a new name, if desired.
 *
 **************************************************************************/
int
delete_entry(names *table[], char *macro_name, int hashval)
{
  names *np;
  names *last_np;

  for (last_np = np = table[hashval]; np != NULL; np=np->next) {
      if (strcmp(macro_name,np->macro_name) == 0) {
        if (last_np == table[hashval]) table[hashval] = np->next;
        else last_np->next = np->next;
        free(np->macro_name);
        free(np->macro);
        free(np);
        return(1);
      }
      last_np = np;
  }
  fprintf(stderr,"WARNING : lex : delete_entry : unable to find entry to delete!\n");
  return(0);
}

/**************************************************************************
 *
 * Function Name : int add_macro()
 * Parameters :
 *      names *table[] - table where names are to be stored in
 *      char *macro_name - name to store macro as
 *      int hashval - computed hash value;
 *      char *text - pointer to an actual copy of macro corresponding to
 *                   macro_name
 * Return Value :
 *      none
 * Purpose :
 *    To add a macro definition into a hash'd database.
 *
 **************************************************************************/
void
add_macro(names *table[], char *macro_name, int hashval, char *text)
{
    names *table_ptr;
    int i;

    table_ptr = table[hashval];
    /*
     * First, check for // and [unescaped] \/\* because comments are legal in a define
     * statement, but should not be included into the expansion
     */
    for (i=0; i<strlen(text); i++) {
	if (text[i] == '/') {
	    if (i < (strlen(text)-1)) {
		if ((text[i+1] == '/') || (text[i+1] == '*')) {
		  text[i] = '\0';
		  break;
		}
	    }
	}
    }
    /*
     * If we are at the head of the index, insert there.
     */
    if (table_ptr == NULL) {
        table[hashval] = table_ptr = MALLOC(names);
    }
    /*
     * search to the end of the linked list and insert there
     */
    else {
        while (table_ptr->next != NULL) table_ptr = table_ptr->next;
        table_ptr->next = MALLOC(names);
        table_ptr = table_ptr->next;
    }
    table_ptr->next = NULL;
    table_ptr->macro_name = NMALLOC(strlen(macro_name)+1,char);
    strcpy(table_ptr->macro_name, macro_name);
    table_ptr->macro = NMALLOC(strlen(text)+1,char);
    strcpy(table_ptr->macro, text);
    table_ptr->is_number = ascii2double(table_ptr->macro,&(table_ptr->value));
}

void
print_double(double value)
{
    int tmp;
 
    if (is_an_integer(value)) {
	tmp = (int) round_int(value);
	mprintf("%d",tmp);
    }
    else {
	mprintf("%f",value);
    }
}

/**************************************************************************
 *
 * Function Name :
 * Parameters :
 * Return Value :
 * Purpose :
 *
 **************************************************************************/
void
print_variable(char *name, int expand_define)
{
    var_values *vvp;
    char strg[MAXSTR];
    double dval;
    int i;

    if (lookup_define(Defines_Table, name, vhash(name))) {
	if (expand_define) {
	    mprintf("%s",lookup_define(Defines_Table, name, vhash(name))->macro);
	    /*
	    print_double(lookup_define(Defines_Table, name, vhash(name))->value);
	    */
	}
	else mprintf("`%s",name);
    }
    else if (is_keyword(name)) mprintf("`%s",name);
    else {
	vvp = get_variable_ref(STRING,name);
	if (vvp == NULL) {
	    sprintf(strg,"ERROR : '%s' not defined",name);
	    wwerror(strg);
	}
	else {
	    if ((vvp->strg)[0] == '"') {
		for (i=1;i<strlen(vvp->strg);i++) {
		    if ((vvp->strg)[i] == '"') break;
		    else mprintf("%c",(vvp->strg)[i]);
		}
	    }
	    else {
		ascii2double(vvp->strg,&dval);
		print_double(dval);
	    }
	}
    }
}

/**************************************************************************
 *
 * Function Name :
 * Parameters :
 *    none
 * Return Value :
 * Purpose :
 *
 **************************************************************************/
void
store_lines()
{
    if (strcmp(wwtext,"\n") == 0) {
	strcpy(old_line,current_line);
	strcpy(current_line, "");
    }
    else strcat(current_line, wwtext);
}

/**************************************************************************
 *
 * Function Name :
 * Parameters :
 * Return Value :
 * Purpose :
 *
 **************************************************************************/
static char *previous_file = NULL;

void
do_comment_count(int force_print, int count)
{

    if ((((nl_count % COMMENT_EVERY) == 0) || (force_print)) && !dont_add_comments) 
	{
	if(!previous_file)
		{
		previous_file = malloc(strlen(current_file)+1);
		strcpy(previous_file, current_file);
		mprintf("// vpp_file %s\n",previous_file);		
		}
	else
	if(strcmp(previous_file, current_file))
		{
		free(previous_file);
		previous_file = malloc(strlen(current_file)+1);
		strcpy(previous_file, current_file);
		mprintf("// vpp_file %s\n",previous_file);		
		}
	
	mprintf("// vpp_line %d\n", count);
    	}
}

/**************************************************************************
 *
 * Function Name :
 * Parameters :
 * Return Value :
 * Purpose :
 *    Given a text followed by either ++ or --, we expand it to 
 * "name = name+1" or "name = name-1".  Note that we unput into the
 * parse stream, so that parsing occurs on the expanded text.
 *
 **************************************************************************/
void
do_post_command(char *txt, int post_type)
{
    int i;
    int len;
    char tmp[MAXSTR];

    strcpy(tmp, txt);

    len = strlen(tmp);

    if ((post_type == PP) || (post_type == MM)) {
	unput('1');
    }
    if ((post_type == MM) || (post_type == ME)) {
	unput('-');
    }
    else {
	unput('+');
    }
    for (i=(len-3);i>=0;i--){
	unput(tmp[i]);
    }
    unput('=');
    for (i=(len-3);i>=0;i--) {
	unput(tmp[i]);
    }
}


/**************************************************************************
 *
 * Function Name :
 * Parameters :
 * Return Value :
 * Purpose :
 *
 **************************************************************************/
int
get_define_value(char *name, double *dval)
{
    names *np;

    if ((np=lookup_define(Defines_Table,name,vhash(name)))) {
	*dval = np->value;
	return 0;
    }

    return -1;
}

/**************************************************************************
 *
 * Function Name :
 * Parameters :
 * Return Value :
 * Purpose :
 *
 **************************************************************************/
int
is_keyword(char *name)
{
  if (strcmp(name,"ifdef") == 0) return (TRUE);
  else if (strcmp(name,"ifndef") == 0) return (TRUE);
  else if (strcmp(name,"if") == 0) return (TRUE);
  else if (strcmp(name,"let") == 0) return (TRUE);
  else if (strcmp(name,"for") == 0) return (TRUE);
  else if (strcmp(name,"while") == 0) return (TRUE);
  else if (strcmp(name,"switch") == 0) return (TRUE);
  else if (strcmp(name,"case") == 0) return (TRUE);
  else if (strcmp(name,"default") == 0) return (TRUE);
  else if (strcmp(name,"breaksw") == 0) return (TRUE);
  else if (strcmp(name,"else") == 0) return (TRUE);
  else if (strcmp(name,"endif") == 0) return (TRUE);
  else if (strcmp(name,"endwhile") == 0) return (TRUE);
  else if (strcmp(name,"endswitch") == 0) return (TRUE);
  else if (strcmp(name,"define") == 0) return (TRUE);

  return (FALSE);
}

#if 0
/**************************************************************************
 *
 * Function Name : int eat_up_comment(fp)
 * Parameters :
 *      FILE *fp;
 * Purpose :
 *      To parse out comments.
 *      (This has been replaced with production rules since it doesn't work
 *      correctly if you drop off the end of a file!)
 *
 **************************************************************************/
void
eat_up_comment(int to_stack)
{
    register int c;
    char tmp[MAXNAME];
    
    if (to_stack) store_it("/*",FALSE,FALSE);
    else mprintf("/*");
    for (;;) {
      while ((c = input()) != '*' && c != EOF) {
	tmp[0] = c; tmp[1]='\0';
	if (to_stack) store_it(tmp,FALSE,FALSE);
	else mprintf("%c",c);
	if (c == '\n') nl_count++; /* eat up text */
      }
      if (c == '*') {
	tmp[0] = c; tmp[1]='\0';
	if (to_stack) store_it(tmp,FALSE,FALSE);
	else mprintf("%c",c);
	while ((c = input()) == '*') {
	  tmp[0] = c; tmp[1]='\0';
	  if (to_stack) store_it(tmp,FALSE,FALSE);
	  else mprintf("%c",c);
	}
	tmp[0] = c; tmp[1]='\0';
	if (to_stack) store_it(tmp,FALSE,FALSE);
	else mprintf("%c",c);
	if (c == '/') {
	  break; /* found the end of comment */
	}
	if (c == EOF) {
	  mprintf("EOF in comment\n");
	  break;
	} /* of if we got EOF */
      } /* of if we got a '*' */
    } /* end of for loop */
 } /* end of eat_up_comment() */
#endif

/**************************************************************************
 *
 * Function Name : int undef_macro(def_text)
 * Parameters :
 *     char *def_text - text of the macro to undefine
 * Return Value :
 * Purpose :
 *     This function exists to undefine a macro, if it so exists.  Note
 * that no warning message is given if the macro does not exist.
 *
 **************************************************************************/
void
undef_macro(char *def_text)
{
   char *txt_ptr;
   char macro_name[MAXSTR];
   int hashval;

   txt_ptr = (char *) (((int)strstr(def_text, "undef")) + ((int)(char *)strlen("undef")));
   sscanf(txt_ptr, "%s", macro_name);

   hashval = vhash(macro_name);

   /*
    * First, see if the name has previously been stored.
    */
   if (lookup_define(Defines_Table, macro_name, hashval) != NULL) {
     delete_entry(Defines_Table, macro_name, hashval); 
  }
}

void
set_to_vpp_mode()
{
    last_mode = YY_START;
    VPPBEGIN(MPP);
}

void
set_to_initial_mode()
{
    last_mode = YY_START;
    VPPBEGIN(INITIAL);
}

void
set_to_copy()
{
    last_mode = YY_START;
    VPPBEGIN(COPY);
}

void
set_to_last_mode()
{
    VPPBEGIN(last_mode);
}

/**************************************************************************
 *
 * Function Name :
 * Parameters :
 * Return Value :
 * Purpose :
 *
 **************************************************************************/

