
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        presim.c
**
**   Functions:   adddelay (static), addin (static),
**                addout (static),
**                getnode (static), input (static), newnode (static),
**                newobject (static),
**                problist (static), push (static), space (static)
**                presim, prnode
**
**   Description: functions to read in .ivf file and setup network
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**		  Anthony J. Bybell(AJB) - Penn State University
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**   08/18/97	 AJB	  Added xnor primitive reference to fnt table.
**   08/30/97    AJB      Removed compiler warnings.
**   01/15/98    AJB      Added supply, tri, and wire primitive refs
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include <ctype.h>
#include "net.h"
#include "system_defs.h"
#include "ivf_format.h"

/* used to create linked lists */
typedef struct list *lptr;
typedef int    *lcontents;

void push(lcontents, lptr *);	/* to get rid of compiler warning */

extern          prim_and ();
extern          prim_or ();
extern          prim_nand ();
extern          prim_nor ();
extern          prim_xor ();
extern		prim_xnor();

extern		prim_supply1();
extern		prim_supply0();
extern		prim_tri1();
extern		prim_tri0();
extern		prim_wire();
extern		prim_wand();
extern		prim_wor();

extern		prim_nmos();
extern		prim_pmos();

extern		prim_buf();
extern		prim_bufif0();
extern		prim_bufif1();

extern          prim_not();
extern          prim_notif0();
extern          prim_notif1();

extern          prim_null ();
extern          prim_eq ();
extern          prim_ne ();
extern          prim_lt ();
extern          prim_le ();
extern          prim_ge ();
extern          prim_conditional ();
extern          prim_selected ();

struct functions funs[] = {
    GATE_AND_STR, prim_and,
    GATE_OR_STR, prim_or,
    GATE_NAND_STR, prim_nand,
    GATE_NOR_STR, prim_nor,
    GATE_XOR_STR, prim_xor,
    GATE_XNOR_STR,prim_xnor,

    GATE_SUPPLY1_STR,prim_supply1,
    GATE_SUPPLY0_STR,prim_supply0,
    GATE_TRI1_STR,prim_tri1,
    GATE_TRI0_STR,prim_tri0,
    GATE_WIRE_STR,prim_wire,
    GATE_WAND_STR,prim_wand,
    GATE_WOR_STR,prim_wor,

    GATE_NMOS_STR, prim_nmos,
    GATE_PMOS_STR, prim_pmos,

    GATE_BUF_STR,prim_buf,
    GATE_BUFIF0_STR,prim_bufif0,
    GATE_BUFIF1_STR,prim_bufif1,

    GATE_NOT_STR, prim_not,
    GATE_NOTIF0_STR, prim_notif0,
    GATE_NOTIF1_STR, prim_notif1,

    GATE_NULL_STR, prim_null,
    GATE_EQ_STR, prim_eq,
    GATE_NE_STR, prim_ne,
    GATE_LT_STR, prim_lt,
    GATE_LE_STR, prim_le,
    GATE_GT_STR, prim_ge,
    GATE_COND_STR, prim_conditional,
    GATE_SEL_STR, prim_selected,
    0, 0
};


struct list
{
    lptr            next;
    lcontents       this;	/* actually, points to lots of different
				 * things */
};

static int      lineno = 0;
static int      obindex = 0;	/* number of objects			 */

static char    *space ();
static nptr     getnode ();
static void     newnode ();
static void     newobject ();
static void     adddelay ();
static void     addin ();
static void     addout ();


/*
**   Function:     space (static)
**
**   File:         presim.c
**
**   Description:  routine to return free storage on an even byte
**                 boundary;  if we still have preallocated storage
**                 available, return that; if not, attempt to allocate
**                 new storage for the nodes
**
**   Arguments:    n - number of bytes to allocate
**
**   Return Value: pointer to allocated storage
*/
/*
**  THIS ROUTINE IS SUSPECT !!!!!
**    what happens if we can't satisfy the request, and allocate more
**    memory, we lose the memory block that wasn't big enough to satisfy
**    our original request!!!, also, note that return addresses from
**    this routine cannot be 'free'd !!!
*/
static char    *
space (n)
    register int    n;
{
    register char  *p;
    static char    *storage;	/* pointer to free storage */

    /* round the number of requested bytes up to an even longword  */
    n = (n + 3) & ~3;

    /* if we requested more than we currently have free, allocate more mem.  */
    if (n > nfree)
    {
	/*
	 * if we requested more than our standard block size, allocate just
	 * the amount requested
	 */
	if (n > NFREE)
	{
	    if ((p = (char *) malloc (n)) == NULL)
	    {
		verror ("No more room", "");
		exit (0);
	    }
	    return (p);
	}

	/*
	 * we have requested less than out standard block size, just allocate
	 * a standard block and go from there
	 */
	if ((storage = (char *) malloc (NFREE)) == NULL)
	{
	    verror ("No more room", "");
	    exit (0);
	}
	nfree = NFREE;
    }

    /* reduce our free block by the specified size  */
    nfree -= n;
    p = storage;
    storage += n;
    return (p);
}


/*
**   Function:     getnode (static)
**
**   File:         presim.c
**
**   Description:  routine to take a node name and return the address of
**                 the node in the hash table; if the node is not in the
**                 hash table, this is an error
**
**   Arguments:    'p' - the name of the node
**
**   Return Value: the address of the node in the hash table
*/
static          nptr
getnode (p)
    register char  *p;
{
    register nptr   n;

    /* if the node is already there, then just return its address  */
    if (n = find (p))
	return (n);

    /* else this is an error  */
    fprintf (stderr, "couldn't find node %s\n", p);
    exit (1);
}


/*
**   Function:     newnode (static)
**
**   File:         presim.c
**
**   Description:  routine to add a new node of name 'node_name', if
**                 the node does not already exist
**
**   Arguments:    node_name - name of node to add
**
**   Return Value: None.
*/
static void
newnode (node_name)
    char           *node_name;
{
    nptr            ntemp;
    char           *p = node_name;

    register nptr   n;
    register char  *q;
    register char  *f;

    /* if the node is already there, then just return its address  */
    if (n = find (node_name))
    {
	/* this is an error, the node already exists  */
	fprintf (stderr, "error, node %s already exists\n", node_name);
	exit (1);
    }

    /* otherwise, allocate new node from free storage */
    n = (nptr) space (sizeof (struct Node));
    nindex++;

    /* initialize node entries */
    n->nname = "0";
    n->nlink = NULL;
    n->events = NULL;
    n->in = NULL;
    n->out = NULL;
    n->hnext = NULL;
    n->cause = NULL;
    n->nvalue = 'x';
    n->nflags = 0;
    n->ctime = 0;
    n->delta_stamp = 0;

    /* if node name is just a number, store it as such */
    if ((f = numberp (p)) != 0)
    {
	/*
	 * 'p' points to a string which is the ASCII representation of a
	 * number
	 */
	n->nname = (char *) f;
    }
    else
    {
	/* 'p' is a string which names the node, copy the name as it appears  */
	n->nname = q = (char *) space (strlen (p) + 1);
	while (*q++ = *p++);

	p = n->nname;
	n->nflags |= NAMED;
    }

    n_insert (n);		/* insert node into hash table */
}


/*
**   Function:     newobject (static)
**
**   File:         presim.c
**
**   Called by:    input (presim.c)
**
**   Description:  routine to add a new object of name obj_name targv[2]
**                 and of object type targv[1]; all other object
**                 fields are initialized to default values;
**                 if the object type is a process, add it to the
**                 object list; set the current object to this object
**
**   Arguments:    gate_name - name of gate to add
**                 gate_type - type/function of gate
**
**   Return Value: None.
*/
static void
newobject (gate_name, gate_type)
    char           *gate_name, *gate_type;
{
    register obptr  ob;
    int             i;


    ob = (obptr) space (sizeof (struct Object));
    obindex++;
    ob->oname = (char *) space (strlen (gate_name) + 1);
    strcpy (ob->oname, gate_name);
    ob->otype = (char *) space (strlen (gate_type) + 1);
    strcpy (ob->otype, gate_type);
    ob->in = NULL;
    ob->out = NULL;
    ob->eval = "0";
    ob->delay = 0;
    ob->dly_type = 0;


    /* set pointer to current object						 */
    curobj = ob;

    curobj->eval = (char *) space (strlen (gate_type) + 1);
    strcpy (curobj->eval, gate_type);

    /* search for the function and copy its address into the object  */
    for (i = 0; funs[i].fun; i++)
    {
	/* if we found the function, then we are done  */
	if (!strcmp (curobj->eval, funs[i].name))
	{
	    curobj->fun = funs[i].fun;
	    break;
	}
    }

    /* if we didn't find the function, search the list of process statements  */
    if (funs[i].fun == NULL)
    {

	/* search for the function and copy its address into the object  */
	for (i = 0; proc_funs[i].fun; i++)
	{
	    /* if we found the function, then we are done  */
	    if (!strcmp (curobj->eval, proc_funs[i].name))
	    {
		curobj->fun = proc_funs[i].fun;
		break;
	    }
	}

	if (proc_funs[i].fun == NULL)
	{
	    fprintf (stderr, "couldn't find address of function %s\n", gate_type);
	    exit (1);
	}

	/* add the process to the seq_obs list	 */
	if (seq_obs != NULL)
	{
	    end_seq->next = (oblptr) space (sizeof (struct oblist));
	    end_seq = end_seq->next;
	}
	else
	{
	    end_seq = (oblptr) space (sizeof (struct oblist));
	    seq_obs = end_seq;
	}
	end_seq->this = ob;
	end_seq->next = (oblptr) 0;

    }


}



/*
**   Function:     adddelay (static)
**
**   File:         presim.c
**
**   Description:  routine to store the delay for the current gate
**
**   Arguments:    delay_time - amount of time for delay for this gate
**
**   Return Value: None.
*/
static void
adddelay (delay_time)
    int             delay_time;
{
    curobj->delay = delay_time;
}



/*
**   Function:     addin (static)
**
**   File:         presim.c
**
**   Description:  routine to add an input node to the  list of input nodes
**                 for the current object
**
**   Arguments:    node_name - name of input node to add
**
**   Return Value: None.
*/
static void
addin (node_name)
    char           *node_name;
{
    nptr            ntemp;

    ntemp = getnode (node_name);
    /*  always add nodes to the input list for the object  */
    push ((lcontents) ntemp, (lptr *) & curobj->in);

    /*  only make the node fire the object if this is not a process or it's
    **  on the sensitivity list
    */
    if (curobj->oflags != PROCESS_GATE)
    {
        push ((lcontents) curobj, (lptr *) & ntemp->in);
    }
}


/*
**   Function:     addout (static)
**
**   File:         presim.c
**
**   Description:  routine to add the node passed in to the list of
**                 output nodes for the current object
**
**   Arguments:    node_name - name of output node to add
**
**   Return Value: None.
*/
static void
addout (node_name)
    char           *node_name;
{
    nptr            ntemp;

    ntemp = getnode (node_name);
    push ((lcontents) curobj, (lptr *) & ntemp->out);
    push ((lcontents) ntemp, (lptr *) & curobj->out);
}


/*
**   Function:     problist (static)
**
**   File:         presim.c
**
**   Description:  routine to print the specified object list to the
**                 supplied file descriptor
**
**   Arguments:    f - file descriptor to print object list to
**                 obl - pointer to start of object list
**
**   Return Value: None.
*/
static void
problist (f, obl)
    FILE           *f;
    oblptr          obl;
{
    oblptr          l;

    if (obl == NULL)
    {
	fprintf (f, "\tNone\n");
    }
    else
    {
	for (l = obl; l != NULL; l = l->next)
	{
	    fprintf (f, "\t%s\n", l->this->oname);
	}
    }
}


/*
**   Function:     push
**
**   File:         presim.c
**
**   Description:  routine to add 'what' to the tail of 'list'
**                 this routine has no concept of what 'what' or 'list' is
**                 and only works because "a pointer is a pointer is a ..."
**
**   Arguments:    what - item to add to the list
**                 list - list to add item to
**
**   Return Value: None.
*/
void
push (what, list)
    lcontents       what;
    lptr           *list;
{
    register lptr   l, l2;

    /* allocate space for this header  */
    l = (lptr) space (sizeof (struct list));

    /* point the new header to 'what'  */
    l->this = what;
    l->next = NULL;

    /* find the end of the list */
    if (*list == NULL)
    {
	/* first item in the list  */
	*list = l;
    }
    else
    {
	/* scan down the list until we find the end */
	for (l2 = *list; l2->next != NULL; l2 = l2->next);

	/* add the last item in the list  */
	l2->next = l;
    }
}


/*
**   Function:     presim
**
**   File:         presim.c
**
**   Description:  routine to coordinate reading in the .ivf file and
**                 building the network of nodes and objects
**
**   Arguments:    db_file - the database file to open/read (must end in .ivf)
**
**   Return Value: None.
*/
void
presim (db_file)
    char           *db_file;
{
    FILE           *in;
    char           *fname;	/* current input file 	 */
    nptr mynode;

    /* initialize the hash table  */
    init_hash ();

    /* add the three predefined nodes  */

    /* add the zero node */
    newnode (SIG_ZERO);

    /* add the one node  */
    newnode (SIG_ONE);

    /* add the dummy node (we don't care what it's value is  */
    newnode (STR_DUMMY_SIGNAL);

newnode(SIG_X);
mynode=find(SIG_X); mynode->nvalue='x';

newnode(SIG_Z);	/* add x and z */
mynode=find(SIG_Z); mynode->nvalue='z';


    fname = db_file;
    lineno = 0;

    /* open the .ivf file   */
    in = fopen (fname, "r");

    if (in == NULL)
    {
	fprintf (stderr, "can not open %s for input\n", fname);
	exit (1);
    }


    fprintf (stderr, "reading database file %s ...\n", db_file);

    read_ivf (in);
    fclose (in);

    /* output count of number of items read in  */
    fprintf (stderr, "database setup complete: %d nodes,  %d objects\n",
	     nindex, obindex);
    fprintf (stderr, "\n");
}


/*
**   Function:     prnode
**
**   File:         presim.c
**
**   Description:  routine to print out the contents of the specified node
**
**   Arguments:    n - the node to output
**
**   Return Value: None.
*/
void
prnode (n, long_dump)
    register nptr   n;
    int             long_dump;
{
    unsigned int    i;

    if (long_dump == TRUE)
    {
	/* long dump format  */
	lprintf (outfile, "\n%s - %c\n", n->nname, n->nvalue);
	lprintf (outfile, "   flags: ");

	if (n->nflags != 0)
	{
	    for (i = 1; i <= 0x8000; i <<= 1)
	    {
		switch (n->nflags & i)
		{
		  case TRACED:
		    lprintf (outfile, "TRACED ");
		    break;

		  case NAMED:
		    lprintf (outfile, "NAMED ");
		    break;

		  case ALIAS:
		    lprintf (outfile, "ALIAS ");
		    break;

		  case INITED:
		    lprintf (outfile, "INITED ");
		    break;

		  case STOPONCHANGE:
		    lprintf (outfile, "STOPONCHANGE ");
		    break;

		  case USERDELAY:
		    lprintf (outfile, "USERDELAY ");
		    break;

		  case VISITED:
		    lprintf (outfile, "VISITED ");
		    break;

		  case FORCE_VALUE:
		    lprintf (outfile, "FORCE_VALUE ");
		    break;

		  case FORCE_ONCE:
		    lprintf (outfile, "FORCE_ONCE ");
		    break;

		  default:
		    break;
		}
	    }
	}
	else
	{
	    lprintf (outfile, "none");
	}
	lprintf (outfile, "\n");

	/* dump the time of last change  */
	lprintf (outfile, "time of last change:  %ld\n", n->ctime);
	
	lprintf (outfile, "cause of last change: ");
        if (n->cause != NULL)
        {
            lprintf (outfile, "%s\n", n->cause->nname);
        }
        else
        {
            lprintf (outfile, "forced\n");
        }

	/* dump the nodes we are effected by  */
	lprintf (outfile, "in:");
	problist (outfile, n->in);

	/* dump the nodes we effect  */
	lprintf (outfile, "out:");
	problist (outfile, n->out);
    }
    else
    {
	/* abbreviated dump  */
	lprintf (outfile, "%s - %c\n", n->nname, n->nvalue);
    }
}


static FILE    *in_file;
static char    *find_token ();
static char    *get_next_token ();
static void     ivf_gate_def ();
static void     ivf_sig_def ();
static void     ivf_proc_def ();

#define CHK_IVF_END(a)   \
    if ((a) == NULL) \
    { \
        fprintf (stderr, "premature end to .ivf file found\n"); \
        exit (1); \
    }


read_ivf (in)
    FILE           *in;
{
    char           *tok_ptr;

    in_file = in;

    /* search for the flattened view  */
    tok_ptr = find_token (MOD_VIEW_FLAT);

    if (tok_ptr == NULL)
    {
	fprintf (stderr, "error, flattened description not found in .ivf file\n");
	exit (1);
    }

    tok_ptr = get_next_token ();

    if (strcmp (tok_ptr, IVF_OPEN_STR))
    {
	/* we couldn't find our open brace  */
	fprintf (stderr, "error, open brace not found in .ivf file\n");
	exit (1);
    }

    /* loop through the flattened view  */
    for (;;)
    {
	tok_ptr = get_next_token ();
	CHK_IVF_END (tok_ptr);

	if (!strcmp (tok_ptr, FLAT_SIG))
	{
	    /* definition of signals  */
	    ivf_sig_def ();
	}
	else if (!strcmp (tok_ptr, FLAT_GATE))
	{
	    /* primitive gate definition  */
	    ivf_gate_def ();
	}
	else if (!strcmp (tok_ptr, FLAT_PROC))
	{
	    /* process definition  */
	    ivf_proc_def ();
	}
	else if (!strcmp (tok_ptr, IVF_CLOSE_STR))
	{
	    /* our closeing brace, finis  */
	    break;
	}
	else
	{
	    /* error  */
	    fprintf (stderr, "illegal token, %s, found in flattened .ivf description",
		     tok_ptr);
	    exit (1);
	}
    }
}

#define is_delimiter(a) (isspace((a)) || \
                         ((a) == IVF_OPEN_CHAR) || \
                         ((a) == IVF_CLOSE_CHAR))

static char    *
get_next_token ()
{
    static char     line[LSIZE];
    int             done = FALSE;
    char           *c_ptr;
    int 	    getc_rc;

    c_ptr = &(line[0]);

    /* skip any white space  */
    do
    {
	*c_ptr=(getc_rc=getc (in_file));
	if (getc_rc == EOF)
	{
	    done = TRUE;
	    break;
	}
    }
    while (isspace (*c_ptr));

    if (done != TRUE)
    {
	/* save characters until we encounter a delimiter  */
	while (!is_delimiter (*c_ptr))
	{
	    ++c_ptr;

	    *c_ptr=(getc_rc=getc (in_file));
	    if (getc_rc == EOF)
	    {
		done = TRUE;
		break;
	    }
	}

	/* advance the pointer if we don't have a white space delimiter  */
	if (!isspace (*c_ptr) && c_ptr == &(line[0]))
	{
	    c_ptr++;
	}
	else
	{
	    /* unget the delimiter  */
	    ungetc (*c_ptr, in_file);
	}

	/* nul-terminate the string  */
	*c_ptr = '\0';
    }

    if (done == TRUE)
    {
	/* we have reached EOF  */
	return (NULL);
    }
    else
    {
	return (&(line[0]));
    }
}


static char    *
find_token (token_str)
    char           *token_str;
{
    char           *tok_ptr;
    char           *ret_val;

    for (;;)
    {
	tok_ptr = get_next_token ();

	/* if we've run out of input or found the token, we are done  */
	if (tok_ptr == NULL)
	{
	    ret_val = NULL;
	    break;
	}
	else if (!strcmp (token_str, tok_ptr))
	{
	    ret_val = tok_ptr;
	    break;
	}
    }

    return (ret_val);
}

static void
ivf_sig_def ()
{
    char           *tok_ptr;

    /* verify we have our open brace  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    if (strcmp (tok_ptr, IVF_OPEN_STR))
    {
	fprintf (stderr, "found %s when expecting %s in .ivf file\n",
		 tok_ptr, IVF_OPEN_STR);
	exit (1);
    }

    for (;;)
    {
	tok_ptr = get_next_token ();
	CHK_IVF_END (tok_ptr);

	if (!strcmp (tok_ptr, IVF_CLOSE_STR))
	{
	    /* we found our terminating closing brace  */
	    break;
	}

	/* otherwise, add the specified signal to our string definitions  */
	newnode (tok_ptr);
    }
}

static void
ivf_gate_def ()
{
    int             delay_time;
    int             num_io;
    char           *tok_ptr;
    int             i;
    char            gate_name[200];

    /* acquire our gate name  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);
    strcpy (gate_name, tok_ptr);

    /* verify we have our open brace  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    if (strcmp (tok_ptr, IVF_OPEN_STR))
    {
	fprintf (stderr, "found %s when expecting %s in .ivf file\n",
		 tok_ptr, IVF_OPEN_STR);
	exit (1);
    }

    /* read the gate type  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    /* add the gate to the database  */
    newobject (gate_name, tok_ptr);

    /* read the delay time  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    delay_time = check_int_str (tok_ptr);
    if (delay_time != 0)
    {
	adddelay (delay_time);
    }

    /* read the delay flag  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);
    if (!strcmp (tok_ptr, STR_DELAY_TRANSPORT))
    {
	/* a transport delay  */
	curobj->dly_type = TRANSPORT_DELAY;
    }
    else
    {
	/* an inertial delay  */
	curobj->dly_type = INERTIAL_DELAY;
    }


    /* get the number on inputs  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    num_io = check_int_str (tok_ptr);

    /* get all the inputs  */
    for (i = 0; i < num_io; i++)
    {
	tok_ptr = get_next_token ();
	CHK_IVF_END (tok_ptr);

	/* add the input to the list of inputs for this node  */
	addin (tok_ptr);
    }

    /* get the number on outputs  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    num_io = check_int_str (tok_ptr);

    /* get all the outputs  */
    for (i = 0; i < num_io; i++)
    {
	tok_ptr = get_next_token ();
	CHK_IVF_END (tok_ptr);

	/* add the output to the list of outputs for this node  */
	addout (tok_ptr);
    }

    /* get the terminating brace  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    if (strcmp (tok_ptr, IVF_CLOSE_STR))
    {
	fprintf (stderr, "illegal token, %s, found expected %s\n",
		 tok_ptr, IVF_CLOSE_STR);
	exit (1);
    }

}

static void
ivf_proc_def ()
{
    int             num_io;
    char           *tok_ptr;
    int             i;
    char            proc_name[200];
    nptr            ntemp;
    
    /* acquire our process name  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);
    strcpy (proc_name, tok_ptr);

    /* verify we have our open brace  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    if (strcmp (tok_ptr, IVF_OPEN_STR))
    {
	fprintf (stderr, "found %s when expecting %s in .ivf file\n",
		 tok_ptr, IVF_OPEN_STR);
	exit (1);
    }

    /* read the long process name  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    /* add the process to the database  */
    newobject (proc_name, tok_ptr);

    /* get the number on items on the sensitivity list  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    num_io = check_int_str (tok_ptr);

    /* point nodes on sensitivity list to objects  */
    for (i = 0; i < num_io; i++)
    {
	tok_ptr = get_next_token ();
	CHK_IVF_END (tok_ptr);
        
        /*  only point nodes at object which are in the sensitivity list  */
        ntemp = getnode (tok_ptr);
        push ((lcontents) curobj, (lptr *) & ntemp->in);
        curobj->oflags = PROCESS_GATE;
    }

    /*
     * get the number of external signals, and add to both input & output
     * lists
     */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    num_io = check_int_str (tok_ptr);

    /* get the signals  */
    for (i = 0; i < num_io; i++)
    {
	tok_ptr = get_next_token ();
	CHK_IVF_END (tok_ptr);

	/* add the signal to both the input & the output list  */
	addout (tok_ptr);
	addin (tok_ptr);
    }

    /* get the number of variables  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    num_io = check_int_str (tok_ptr);

    /* get all the variables  */
    for (i = 0; i < num_io; i++)
    {
	tok_ptr = get_next_token ();
	CHK_IVF_END (tok_ptr);

	/* define the variable  */
	newnode (tok_ptr);

	/* add the variables to the list of inputs for this node only */
	addin (tok_ptr);
    }

    /* get the terminating brace  */
    tok_ptr = get_next_token ();
    CHK_IVF_END (tok_ptr);

    if (strcmp (tok_ptr, IVF_CLOSE_STR))
    {
	fprintf (stderr, "illegal token, %s, found expected %s\n",
		 tok_ptr, IVF_CLOSE_STR);
	exit (1);
    }
}


check_int_str (str)
    char           *str;
{
    char           *s2;

    /* verify the specified string is a valid integer string  */

    for (s2 = str; *s2 != '\0'; s2++)
    {
	if (!isdigit (*s2))
	{
	    fprintf (stderr, "illegal character, %c, found in digit string\n", *s2);
	    return (-1);
	}
    }

    return (atoi (str));
}
