
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        prims.c
**
**   Functions:   eval_obj, inited_node, prim_and,
**                prim_conditional, prim_eq,
**                prim_ge, prim_gt, prim_le, prim_lt, prim_nand,
**                prim_ne, prim_nor, prim_not,
**                prim_or, prim_selected, prim_xor, prim_xnor
**
**
**   Description:
**
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**		  Anthony J. Bybell(AJB) - Penn State University
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**   10/18/88    ARM      Modifiy selected and conditional for new 
**                        operand ordering (see vcomp/VERSION_HISTORY)
**                        for details on selected and conditional 
**                        operand ordering
**   08/18/97    AJB	  Added xnor primitive.
**   01/15/98    AJB      Added supply, tri, and wire primitives
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include "net.h"
#include "system_defs.h"

#define get_val(nl)  ((nl)->this->nvalue)

eval_obj (b)
    obptr           b;
{
    (*b->fun) (b);
}

inited_node (node)
    nptr            node;
{
    if ((node->nvalue == 'x') || (node->nvalue == 'z'))
    {
#if 0
        /*  mark the node as initialized and being forced  */
	node->nflags |= (INITED | FORCE_ONCE);

        /*  force this new value  */
        new_out (node, '0', NO_DELTA, INERTIAL_DELAY);
#endif
        
	return 1;
    }
    else
	return 0;
}

prim_and (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out = '1';

    /*  break on the first '0' for an AND  */
if(!(obj->in->next))	{ out = get_val(obj->in); }
else
    for (nl = obj->in; nl; nl = nl->next)
    {
	if (get_val(nl) == '0')
	{
	    out = '0';
	    break;
	}
	else
	if (get_val(nl) != '1')
	{
	    out = 'x';
	}
    }

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}




/*	vhdl nand gate can only have 2 inputs				*/
prim_nand (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out = '0';

if(!(obj->in->next))	{ out = get_val(obj->in); }
else
    for (nl = obj->in; nl; nl = nl->next)
    {
	if (get_val(nl) == '0')
	{
	    out = '1';
	    break;
	}
	else
	if (get_val(nl) != '1')
	{
	    out = 'x';
	}
    }
    
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


/*	vhdl nor gate can only have 2 inputs				*/
prim_nor (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out = '1';

if(!(obj->in->next))	{ out = get_val(obj->in); }
else
    for (nl = obj->in; nl; nl = nl->next)
    {
	if (get_val(nl) == '1')
	{
	    out = '0';
	    break;
	}
	else
	if (get_val(nl) != '0')
	{
	    out = 'x';
	}

    }
    
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


prim_or (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out = '0';

if(!(obj->in->next))	{ out = get_val(obj->in); }
else
    for (nl = obj->in; nl; nl = nl->next)
    {
	if (get_val(nl) == '1')
	{
	    out = '1';
	    break;
	}
	else
	if (get_val(nl) != '0')
	{
	    out = 'x';
	}
    }
    
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

prim_xor (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;

    out =get_val(obj->in);

    for (nl = obj->in->next; nl; nl = nl->next)
    {
	if ( (get_val(nl) == 'x') ||
		(get_val(nl) == 'z') )
		{
		out='x';
		break;
		}

	if (out == get_val(nl))
	    out = '0';
	else
	    out = '1';

    }
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


prim_xnor (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;

    out =get_val(obj->in);

    for (nl = obj->in->next; nl; nl = nl->next)
    {
	if ( (get_val(nl) == 'x') ||
		(get_val(nl) == 'z') )
		{
		out='x';
		break;
		}

	if (out == get_val(nl))
	    out = '1';
	else
	    out = '0';
    }
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


prim_not (obj)
obptr (obj);
{
    char            out;

    if (get_val(obj->in) == '0')
	out = '1';
    else 
    if (get_val(obj->in) == '1')
	out = '0';
    else
	out = 'x';

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

prim_buf (obj)
obptr (obj);
{
    char            out;

    if (get_val(obj->in) == '0')
	out = '0';
    else 
    if (get_val(obj->in) == '1')
	out = '1';
    else
	out = 'x';

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


prim_bufif0 (obj)
obptr (obj);
{
char out, a, e;

a=get_val(obj->in->next);
e=get_val(obj->in->next->next);

if(e=='1')
	{
	out='z';	
	}
else
if(e=='0')
	{
	if(a=='z') out='x'; else out=a;
	}
else
	{
	if((a=='0')||(a=='1')) out='z'; else out='x';
	}

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

prim_bufif1 (obj)
obptr (obj);
{
char out, a, e;

a=get_val(obj->in->next);
e=get_val(obj->in->next->next);

if(e=='0')
	{
	out='z';	
	}
else
if(e=='1')
	{
	if(a=='z') out='x'; else out=a;
	}
else
	{
	if((a=='0')||(a=='1')) out='z'; else out='x';
	}

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


prim_pmos (obj)
obptr (obj);
{
char out, a, e;

a=get_val(obj->in->next);
e=get_val(obj->in->next->next);

if(e=='1')
	{
	out='z';	
	}
else
if(e=='0')
	{
	out=a;
	}
else
	{
	if(a=='x') out='x'; else out='z';
	}

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


prim_nmos (obj)
obptr (obj);
{
char out, a, e;

a=get_val(obj->in->next);
e=get_val(obj->in->next->next);

if(e=='0')
	{
	out='z';	
	}
else
if(e=='1')
	{
	out=a;
	}
else
	{
	if(a=='x') out='x'; else out='z';
	}

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}



prim_notif0 (obj)
obptr (obj);
{
char out, a, e;

a=get_val(obj->in->next);
e=get_val(obj->in->next->next);

if(e=='1')
	{
	out='z';	
	}
else
if(e=='0')
	{
	switch(a)
		{
		case '0':  out='1'; break;
		case '1':  out='0'; break;
		default: out='x'; break;
		}
	}
else
	{
	if((a=='0')||(a=='1')) out='z'; else out='x';
	}

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

prim_notif1 (obj)
obptr (obj);
{
char out, a, e;

a=get_val(obj->in->next);
e=get_val(obj->in->next->next);

if(e=='0')
	{
	out='z';	
	}
else
if(e=='1')
	{
	switch(a)
		{
		case '0':  out='1'; break;
		case '1':  out='0'; break;
		default: out='x'; break;
		}
	}
else
	{
	if((a=='0')||(a=='1')) out='z'; else out='x';
	}

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


prim_null (obj)
obptr (obj);
{
    char            out;

    out = get_val(obj->in) ;

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

prim_supply1 (obj)
obptr (obj);
{
    char            out;

    out = '1' ;

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

prim_supply0 (obj)
obptr (obj);
{
    char            out;

    out = '0' ;

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}



prim_tri1 (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;

    out ='z';  /* skips $0 element with "->next" below! */

    for (nl = obj->in->next; nl; nl = nl->next)
    {
    if((get_val(nl)=='x'))
	{
	out='x';
	break;
	}
     else
     if((out=='z')&&(get_val(nl)!='z'))
	{
	out=get_val(nl);
	continue;
	}
     else
     if((out!='z')&&(get_val(nl)!=out)&&(get_val(nl)!='z'))
	{
	out='x';
	break;
	}
    }

    if(out=='z') out='1';

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

prim_tri0 (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;

    out ='z';  /* skips $0 element with "->next" below! */

    for (nl = obj->in->next; nl; nl = nl->next)
    {
    if((get_val(nl)=='x'))
	{
	out='x';
	break;
	}
     else
     if((out=='z')&&(get_val(nl)!='z'))
	{
	out=get_val(nl);
	continue;
	}
     else
     if((out!='z')&&(get_val(nl)!=out)&&(get_val(nl)!='z'))
	{
	out='x';
	break;
	}
    }

    if(out=='z') out='0';

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


/* same as tri0/1 but w/o the resistive pull.. */
prim_wire (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;

    out ='z';  /* skips $0 element with "->next" below! */

    for (nl = obj->in->next; nl; nl = nl->next)
    {
    if((get_val(nl)=='x'))
	{
	out='x';
	break;
	}
     else
     if((out=='z')&&(get_val(nl)!='z'))
	{
	out=get_val(nl);
	continue;
	}
     else
     if((out!='z')&&(get_val(nl)!=out)&&(get_val(nl)!='z'))
	{
	out='x';
	break;
	}
    }

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


/* similar to wire.. */
prim_wand (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;

    out ='z';  /* skips $0 element with "->next" below! */

    for (nl = obj->in->next; nl; nl = nl->next)
    {
    if((get_val(nl)=='0'))
        {
        out='0';
        break;
        }
    else
    if((get_val(nl)=='x'))
	{
	out='x';
        continue;
	}
     else
     if((out=='z')&&(get_val(nl)!='z'))
	{
	out=get_val(nl);
	continue;
	}
     else
     if((out!='z')&&(get_val(nl)!=out)&&(get_val(nl)!='z'))
	{
	out='x';
        continue;
	}
    }

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


/* similar to wand.. */
prim_wor (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;

    out ='z';  /* skips $0 element with "->next" below! */

    for (nl = obj->in->next; nl; nl = nl->next)
    {
    if((get_val(nl)=='1'))
        {
        out='1';
        break;
        }
    else
    if((get_val(nl)=='x'))
	{
	out='x';
        continue;
	}
     else
     if((out=='z')&&(get_val(nl)!='z'))
	{
	out=get_val(nl);
	continue;
	}
     else
     if((out!='z')&&(get_val(nl)!=out)&&(get_val(nl)!='z'))
	{
	out='x';
        continue;
	}
    }

    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}




prim_selected (obj)
    obptr           obj;

/*  note:  initializes all input nodes which are looked at for test	*/
/*
**
**   Given the following:
**      N - the length of the output vector, in bits
**      M - the length of the expression, in bits
**      K - the number of choices
**
**   the list of outputs looks like the following:
**       output 0   - bit 1 to write to
**       output 1   - bit 2 to write to
**         .
**         .
**         .
**       output N-1 - bit N to write to
**
**   note that this is a multi-output 'gate', the number of outputs, N,
**   is significant and is used in parsing the output list below
**
**   the list of inputs looks like the following:
**       input 0   - output waveform, bit 1
**       input 1   - output waveform, bit 2
**         .
**         .
**         .
**       input N-1 - output waveform, bit N
**
**       input N+0   - alternative waveform, bit 1
**       input N+1   - alternative waveform, bit 2
**         .
**         .
**         .
**       input 2*N-1 - alternative waveform, bit N
**
**       input 2*N+0   - expression, bit 1
**       input 2*N+1   - expression, bit 2
**         .
**         .
**         .
**       input 2*N+M-1 - expression, bit M
**
**       input 2*N+M   - DUMMY SIGNAL 
**              (place holder to determine expression size, M)
**
**       input 2*N+M+1 - choice 1, bit 1
**       input 2*N+M+2 - choice 1, bit 2
**         .
**         .
**         .
**       input 2*N+2*M - choice 1, bit M
**
**       input 2*N+2*M+1 - choice 2, bit 1
**       input 2*N+2*M+2 - choice 2, bit 2
**         .
**         .
**         .
**       input 2*N+3*M   - choice 2, bit M
**
**     etc.
**
**       input 2*N+K*M+1 - choice K, bit 1
**       input 2*N+K*M+2 - choice K, bit 2
**         .
**         .
**         .
**       input 2*N+K*M   - choice K, bit M
**
**
**  the algorithm is in any of the 'choices' matches the 'expression',
**  then assign the output waveform to the proper nodes; if not, assign
**  the alternative waveform.  if the alternative waveform is the
**  DUMMY_SIGNAL, don't output anything (i.e. don't change the value for
**  the 'gate's outputs)
**
*/

{
#if 0
    char            out;
    nptr            alt_wf;
    nlptr           nl;

    /*  get a pointer to the alt_wf  */
    nl = obj->in->next;
    alt_wf = nl->this;
    
    /*  scan through the choices  */
    for (nl = nl->next; nl != NULL; nl = nl->next)
    {
	if (get_val(nl) == '1')
        {
            /*  output the first waveform as the output  */
            out = get_val(obj->in);
	    break;
	}
    }

    if (nl == NULL)
    {
        /*  none of the choices matched, see if we have an alternative  */
        if (strcmp (alt_wf->nname, STR_DUMMY_SIGNAL))
        {
            /*  we have an alternative, output it  */
            out = (inited_node(alt_wf) ? '0' : alt_wf->nvalue);
        }
        else
        {
            /*  we have no alternative, don't touch the output  */
            return;
        }
    }

    /*  output the new value  */
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
#endif

    int N, M, match, i;
    nlptr nl, exp_ptr, e_ptr, ow;

    /*  count the number of outputs, N  */
    for (nl = obj->out, N=0; nl != NULL; nl = nl->next, N++);

    /*  skip the output and alternative waveforms  */
    for (nl = obj->in, i=0; i < 2*N ; nl = nl->next, i++);

    /*  keep a pointer to the start of the expression  */
    exp_ptr = nl;

    /*  count the number of bits in the expression, M  */
    for (M=0; strcmp(nl->this->nname, STR_DUMMY_SIGNAL) != 0; 
         nl=nl->next, M++);

    /*  skip past the dummy signal  */
    nl = nl->next;
    
    /*
     * scan through all choices, until we find a match or we
     * run out of choices
     */
    for (match = FALSE; nl != NULL && match == FALSE;)
    {
        for (i = 0, e_ptr = exp_ptr; i < M; 
             nl = nl->next, e_ptr = e_ptr->next, i++)
        {
            /*  check if this bit matches  */
            if (get_val(nl) == get_val(e_ptr))
            {
                /*  if we have no more bits left, we have a match  */
                if (i == (M-1))
                {
                    match = TRUE;
                    break;
                }
                else
                {
                    /*  go and check more bits  */
                    continue;
                }
            }
            else
            {
                /*  skip the rest of this choice  */
                for (; i < M; i++)
                {
                    nl=nl->next;
                }
                
                /*  we don't have a match, break out of this loop  */
                break;
            }
        }
    }
    
    if (match == TRUE)
    {
        /*  we had a match, output the expression  */
        for (i=0, nl = obj->in, ow = obj->out; i < N; 
             ow = ow->next, nl = nl->next, i++)
        {
            /*  output the result  */
            new_out (ow->this, get_val(nl), obj->delay, obj->dly_type);
        }
    }
    else
    {
        /*  no match, check if we have an alternative waveform  */

        /*  skip the expression  */
        for (nl = obj->in, i=0; i < N ; nl = nl->next, i++);

        /*  see if the alternative is a DUMMY_SIGNAL  */
        if (strcmp(nl->this->nname, STR_DUMMY_SIGNAL) != 0)
        {
            /*  we have an alternative waveform, output it  */
            for (i=0, ow = obj->out; i < N; 
                 ow = ow->next, nl = nl->next, i++)
            {
                /*  output the result  */
                new_out (ow->this, get_val(nl), obj->delay, obj->dly_type);
            }
        }
    }
}


/*
**  for conditionals, we have three inputs
**      input 1 - the output waveform
**      input 2 - the expression
**      input 3 - the else clause
*/
prim_conditional (obj)
    obptr           obj;
{
/*  note:  initializes all input nodes which are looked at for test	*/
    nlptr           nl;
    char            out;

    nl = obj->in->next;

    /*  initialize the expression (if necessary) and check it  */
    if (get_val(nl) == '1')
    {
        /*  output the first operand  */
        out = get_val(obj->in);
    }
    else
    {
        /*  output the third (next) operand  */
        nl = nl->next;
        out = get_val(nl);
    }

    /*  output the result  */
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


prim_eq (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;

    out = '1';

    for (nl = obj->in; nl != NULL; nl = nl->next->next)
    {
	if (get_val(nl) != get_val(nl->next))
	{
	    /*  we are done on first failure  */
	    out = '0';
	    break;
	}
    }
    
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

prim_ne (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;

    out = '0';

    for (nl = obj->in; nl != NULL; nl = nl->next->next)
    {
	/*  we are done on first failure  */
	if (get_val(nl) != get_val(nl->next))
	{
	    out = '1';
	    break;
	}
    }
    
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

/*
**  THE FOLLOWING 'GATES' ARE CURRENTLY UNUSED
*/

prim_ge (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;
    char            a1, a2;

    inited_node (obj->in->this);
    a1 = obj->in->this->nvalue;

    for (nl = obj->in->next; nl; nl = nl->next)
    {
	inited_node (nl->this);
	a2 = nl->this->nvalue;
	out = a1 >= a2;
    }
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}

prim_gt (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;
    char            a1, a2;

    inited_node (obj->in->this);
    a1 = obj->in->this->nvalue;

    for (nl = obj->in->next; nl; nl = nl->next)
    {
	inited_node (nl->this);
	a2 = nl->this->nvalue;
	out = a1 > a2;
    }
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}



prim_le (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;
    char            a1, a2;

    inited_node (obj->in->this);
    a1 = obj->in->this->nvalue;

    for (nl = obj->in->next; nl; nl = nl->next)
    {
	inited_node (nl->this);
	a2 = nl->this->nvalue;
	out = a1 <= a2;
    }
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}


prim_lt (obj)
    obptr           obj;
{
    nlptr           nl;
    char            out;
    char            a1, a2;

    inited_node (obj->in->this);
    a1 = obj->in->this->nvalue;

    for (nl = obj->in->next; nl; nl = nl->next)
    {
	inited_node (nl->this);
	a2 = nl->this->nvalue;
	out = a1 < a2;
    }
    new_out (obj->out->this, out, obj->delay, obj->dly_type);
}
